/*
 * Copyright(c) 2007 Intel Corporation. All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Maintained at www.Open-FCoE.org
 */

/*
 * FCOE protocol file
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/spinlock.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if_ether.h>
#include <linux/kthread.h>
#include <scsi/scsi_tcq.h>
#include <scsi/scsicam.h>
#include <scsi/scsi_transport.h>
#include <scsi/scsi_transport_fc.h>
#include <net/rtnetlink.h>

#include "sa_assert.h"
#include "fc_types.h"
#include "fc_frame.h"
#include "fc_print.h"
#include "fc_encaps.h"
#define	FCOE_T11_AUG07		/* use new FCoE version */
#include "fc_fcoe.h"
#include "fc_fcoe_old.h"	/* also include old definitions */
#include "fcdev.h"
#include "fcoeioctl.h"
#include "ofc_dbg.h"
#include "fcoe_def.h"
#include "crc32_le.h"

/*
 * Function Prototyes
 */
#define FCOE_MAX_QUEUE_DEPTH  256
int fcoe_percpu_receive_thread(void *arg);

extern int debug_fcoe;

static int fcoe_check_wait_queue(struct fcdev *);
static void fcoe_insert_wait_queue_head(struct fcdev *, struct sk_buff *);
static void fcoe_insert_wait_queue(struct fcdev *, struct sk_buff *);
static void fcoe_recv_flogi(struct fcoe_softc *, struct fc_frame *, u64);

/*
 * this is the fcoe receive function
 * called by NET_RX_SOFTIRQ
 * this function will receive the packet and
 * build fc frame and pass it up
 */
int fcoe_rcv(struct sk_buff *skb, struct net_device *dev,
	     struct packet_type *ptype, struct net_device *olddev)
{
	struct fcdev *fdev;
	struct fcoe_rcv_info *fr;
	struct fcoe_softc *fc;
	struct fcoe_dev_stats *stats;
	u_int8_t *data;
	struct fc_frame_header *fh;
	ushort oxid;
	int cpu_idx;
	struct fcoe_percpu_s *fps;
	struct fcoe_info *fci = &fcoei;

	fc = container_of(ptype, struct fcoe_softc, fcoe_packet_type);
	fdev = fc->fd;
	if (unlikely(fdev == NULL)) {
		OFC_DBG("cannot find hba structure");
		goto err2;
	}

	if (unlikely(debug_fcoe)) {
		OFC_DBG("skb_info: len:%d data_len:%d head:%p data:%p tail:%p "
		       "end:%p sum:%d dev:%s", skb->len, skb->data_len,
			skb->head, skb->data, skb_tail_pointer(skb),
			skb_end_pointer(skb), skb->csum,
		       skb->dev ? skb->dev->name : "<NULL>");

	}

	/* check for FCOE packet type */
	if (unlikely(eth_hdr(skb)->h_proto != htons(ETH_P_FCOE))) {
		OFC_DBG("wrong FC type frame");
		goto err;
	}
	data = skb->data;
	data += fc->fcoe_hlen;
	fh = (struct fc_frame_header *)data;
	oxid = net16_get(&fh->fh_ox_id);

	fr = fcoe_dev_from_skb(skb);
	fr->fd = fdev;
	fr->ptype = ptype;
	cpu_idx = 0;
#ifdef CONFIG_SMP
	/*
	 * if there is no thread for a given idx then use cpu 0
	 * assuming that cpu 0 is always there
	 * Need to support for not having cpu 0 -XXXX
	 */
	cpu_idx = oxid & openfc_cpu_mask;
	if (fci->fcoe_percpu[cpu_idx] == NULL)
		cpu_idx = 0;
#endif
	fps = fci->fcoe_percpu[cpu_idx];

	spin_lock_bh(&fps->fcoe_rx_list.lock);
	__skb_queue_tail(&fps->fcoe_rx_list, skb);
	if (fps->fcoe_rx_list.qlen == 1)
		wake_up_process(fps->thread);

	spin_unlock_bh(&fps->fcoe_rx_list.lock);

	return 0;
err:
#ifdef CONFIG_SMP
	stats = fdev->dev_stats[smp_processor_id()];
#else
	stats = fdev->dev_stats[0];
#endif
	stats->ErrorFrames++;

err2:
	kfree_skb(skb);
	return -1;
}

static inline int fcoe_start_io(struct sk_buff *skb)
{
	int rc;

	skb_get(skb);
	rc = dev_queue_xmit(skb);
	if (rc != 0)
		return rc;
	kfree_skb(skb);
	return 0;
}

/*
 * Free a frame that was allocated by fcoe_alloc_frame().
 * The frame will be inside an sk_buff.
 */
static void fcoe_frame_free(struct fc_frame *fp)
{
	kfree_skb((struct sk_buff *)fp->fr_free_priv);
}

/*
 * Zero the number of frags in skb,
 * Memory should be freed by the aplication which has requested the i/o
 */
static void fcoe_skb_destroy(struct sk_buff *skb)
{
	struct fcoe_info *fci = &fcoei;
	struct fc_frame *fp = (struct fc_frame *)skb->head;
	struct fcdev *fdev = (struct fcdev *)fp->fr_dev;
	struct fcoe_softc *fc = NULL;

	ASSERT(fp->fr_free_priv == skb);

	if (fdev)
		fc = (struct fcoe_softc *)fdev->drv_priv;

	if (unlikely(fp->fr_destructor))
		(fp->fr_destructor) (fp->fr_arg);

	skb_shinfo(skb)->nr_frags = 0;
	atomic_dec(&fci->cleanup_needed);
	if (fc && fc->fcoe_pending_queue.qlen)
		fcoe_check_wait_queue(fdev);
}

/*
 * Allocate a frame intended to be sent via fcoe_xmit.
 * Get an sk_buff for the frame and set the length.
 */
struct fc_frame *fcoe_frame_alloc(size_t len)
{
	struct fc_frame *fp;
	struct sk_buff *skb;

	ASSERT((len % sizeof(uint32_t)) == 0);
	len += sizeof(struct fc_frame_header);
	skb = dev_alloc_skb(len + sizeof(*fp) +
			    FC_FRAME_HEADROOM + FC_FRAME_TAILROOM);
	if (!skb)
		return NULL;
	fp = (struct fc_frame *)skb->head;
	fc_frame_init(fp);
	fp->fr_free = fcoe_frame_free;
	fp->fr_free_priv = skb;
	skb_reserve(skb, sizeof(*fp) + FC_FRAME_HEADROOM);
	skb_put(skb, len);
	fp->fr_hdr = (struct fc_frame_header *)skb->data;
	fp->fr_len = len;
	fp->fr_destructor = NULL;
	return fp;
}

/*
 * this is the frame xmit routine
 */
int fcoe_xmit(struct fcdev *fc_dev, struct fc_frame *fp)
{
	int indx;
	int wlen, rc = 0;
	u_int32_t crc;
	struct ethhdr *eh;
	struct fcoe_info *fci = &fcoei;
	struct fcoe_crc_eof *cp;
	struct sk_buff *skb;
	struct fcoe_dev_stats *stats;
	struct fc_frame_header *fh;
	u_int hlen;		/* header length implies the version */
	u_int tlen;		/* trailer length */
	int flogi_in_progress = 0;
	struct fcoe_softc *fc;
	struct scatterlist *sg;
	void *data;
	uint8_t sof, eof;
	int flg = 0;

	if (unlikely(debug_fcoe))
		fc_print_frame_hdr("fcoe_xmit", fp);

	ASSERT((fp->fr_len % sizeof(uint32_t)) == 0);

	fc = (struct fcoe_softc *)fc_dev->drv_priv;
	/*
	 * if it is a flogi then we need to learn gw-addr
	 * and my own fcid
	 */
	fh = fc_frame_header_get(fp);
	if (unlikely(fh->fh_r_ctl == FC_RCTL_ELS_REQ)) {
		if (fc_frame_payload_op(fp) == ELS_FLOGI) {
			fc->flogi_oxid = net16_get(&fh->fh_ox_id);
			fc->address_mode = FCOE_FCOUI_ADDR_MODE;
			fc->flogi_progress = 1;
			flogi_in_progress = 1;
		} else if (fc->flogi_progress && net24_get(&fh->fh_s_id) != 0) {
			/*
			 * Here we must've gotten an SID by accepting an FLOGI
			 * from a point-to-point connection.  Switch to using
			 * the source mac based on the SID.  The destination
			 * MAC in this case would have been set by receving the
			 * FLOGI.
			 */
			fc_fcoe_set_mac(fc->data_src_addr, &fh->fh_s_id);
			fc->flogi_progress = 0;
		}
	}

	ASSERT(fp->fr_free == fcoe_frame_free);	/* allocated by us */
	skb = (struct sk_buff *)fp->fr_free_priv;
	ASSERT(skb->data == (unsigned char *)fp->fr_hdr);

	sof = fp->fr_sof;
	eof = fp->fr_eof;

	crc = ~0;
	crc = crc32_sb8_64_bit(crc, skb->data, skb_headlen(skb));
	for (indx = 0, sg = fp->fr_sg; indx < fp->fr_sg_len; indx++, sg++) {
		u_long off = sg->offset;
		u_long len = sg->length;

		skb_fill_page_desc(skb, indx, sg_page(sg), off, len);
		skb->len += len;
		skb->data_len += len;
		while (len > 0) {
			u_long clen;

			clen = min(len, PAGE_SIZE - (off & ~PAGE_MASK));
			data = kmap_atomic(sg_page(sg) + (off >> PAGE_SHIFT),
					   KM_SKB_DATA_SOFTIRQ);
			crc = crc32_sb8_64_bit(crc, data + (off & ~PAGE_MASK),
					       clen);
			kunmap_atomic(data, KM_SKB_DATA_SOFTIRQ);
			off += clen;
			len -= clen;
		}
	}

	/*
	 * Get header and trailer lengths.
	 * This is temporary code until we get rid of the old protocol.
	 * Both versions have essentially the same trailer layout but T11
	 * has padding afterwards.
	 */
	hlen = fc->fcoe_hlen;
	tlen = sizeof(struct fcoe_crc_eof);
	if (hlen == sizeof(struct fcoe_hdr_old))
		tlen = sizeof(struct fcoe_crc_eof_old);

	/*
	 * copy fc crc and eof to the skb buff
	 * Use utility buffer in the fc_frame part of the sk_buff for the
	 * trailer.
	 * We don't do a get_page for this frag, since that page may not be
	 * managed that way.  So that skb_free() doesn't do that either, we
	 * setup the destructor to remove this frag.
	 */
	if (skb_is_nonlinear(skb)) {
		struct page *p;
		cp = (struct fcoe_crc_eof *)fp->fr_util;
		p = virt_to_page((char *)cp);
		/*
		 * set the destructor and inc the cleanup_needed
		 * this is done because before unloading the module we should
		 * wait for all the calls to fcoe_skb_destroy()
		 */
		skb->destructor = fcoe_skb_destroy;
		atomic_inc(&fci->cleanup_needed);
		flg = 1;
		skb_fill_page_desc(skb, skb_shinfo(skb)->nr_frags,
				   p, offset_in_page((char *)cp), tlen);
		skb->len += tlen;
		skb->data_len += tlen;
		skb->truesize += tlen;
	} else {
		cp = (struct fcoe_crc_eof *)skb_put(skb, tlen);
	}

	/*
	 * This true for Target Driver only
	 * Initiator does not use frame destructor
	 */
	if (unlikely(fp->fr_destructor)) {
		skb->destructor = fcoe_skb_destroy;
		if (flg == 0)
			atomic_inc(&fci->cleanup_needed);
	}

	net8_put(&cp->fcoe_eof, eof);
	cp->fcoe_crc32 = ~crc;
	if (tlen == sizeof(*cp))
		memset(cp->fcoe_resvd, 0, sizeof(cp->fcoe_resvd));
	wlen = (skb->len - tlen + sizeof(crc)) / FCOE_WORD_TO_BYTE;

	/*
	 *      Fill in the control structures
	 */
	skb->ip_summed = CHECKSUM_NONE;
	eh = (struct ethhdr *)skb_push(skb, hlen + sizeof(struct ethhdr));
	if (fc->address_mode == FCOE_FCOUI_ADDR_MODE)
		fc_fcoe_set_mac(eh->h_dest, (net24_t *) &fh->fh_d_id);
	else
		/* insert GW address */
		memcpy(eh->h_dest, fc->dest_addr, ETH_ALEN);

	if (unlikely(flogi_in_progress))
		memcpy(eh->h_source, fc->ctl_src_addr, ETH_ALEN);
	else
		memcpy(eh->h_source, fc->data_src_addr, ETH_ALEN);

	eh->h_proto = htons(ETH_P_FCOE);
	skb->protocol = htons(ETH_P_802_3);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,21)
	skb_reset_mac_header(skb);
	skb_reset_network_header(skb);
#else
	skb->mac.raw = skb->data;
	skb->nh.raw = skb->data;
#endif
	if (hlen == sizeof(struct fcoe_hdr)) {
		struct fcoe_hdr *hp;

		hp = (struct fcoe_hdr *)(eh + 1);
		memset(hp, 0, sizeof(*hp));
		if (FC_FCOE_VER)
			FC_FCOE_ENCAPS_VER(hp, FC_FCOE_VER);
		hp->fcoe_sof = sof;
	} else {
		struct fcoe_hdr_old *ohp;

		ASSERT(hlen == sizeof(*ohp));
		ohp = (struct fcoe_hdr_old *)(eh + 1);
		net16_put(&ohp->fcoe_plen, FC_FCOE_ENCAPS_LEN_SOF(wlen, sof));
	}

	stats = fc_dev->dev_stats[smp_processor_id()];
	stats->TxFrames++;
	stats->TxWords += wlen;
	skb->dev = fc->real_dev;

	fp->fr_dev = (void *)fc_dev;
	if (fc->fcoe_pending_queue.qlen)
		rc = fcoe_check_wait_queue(fc_dev);

	if (rc == 0)
		rc = fcoe_start_io(skb);

	if (rc) {
		fcoe_insert_wait_queue(fc_dev, skb);
		if (fc->fcoe_pending_queue.qlen > FCOE_MAX_QUEUE_DEPTH)
			openfc_pause(fc_dev);
	}

	return 0;
}

int fcoe_percpu_receive_thread(void *arg)
{
	struct fcoe_percpu_s *p = arg;
	u_int32_t fr_len;
	u_int hlen;
	u_int tlen;
	struct fcdev *fc_dev;
	struct fcoe_rcv_info *fr;
	struct fcoe_dev_stats *stats;
	struct fc_frame_header *fh;
	struct sk_buff *skb;
	struct fcoe_crc_eof *cp;
	enum fc_sof sof;
	struct fc_frame *fp;
	u64 mac = 0;
	struct fcoe_softc *fc;

	set_user_nice(current, 19);

	while (!kthread_should_stop()) {

		spin_lock_bh(&p->fcoe_rx_list.lock);
		while ((skb = __skb_dequeue(&p->fcoe_rx_list)) == NULL) {
			set_current_state(TASK_INTERRUPTIBLE);
			spin_unlock_bh(&p->fcoe_rx_list.lock);
			schedule();
			set_current_state(TASK_RUNNING);
			if (kthread_should_stop())
				return 0;
			spin_lock_bh(&p->fcoe_rx_list.lock);
		}
		spin_unlock_bh(&p->fcoe_rx_list.lock);
		fr = fcoe_dev_from_skb(skb);
		fc_dev = fr->fd;
		if (unlikely(fc_dev == NULL)) {
			OFC_DBG("invalid HBA Structure");
			kfree_skb(skb);
			continue;
		}

		stats = fc_dev->dev_stats[smp_processor_id()];

		if (unlikely(debug_fcoe)) {
			OFC_DBG("skb_info: len:%d data_len:%d head:%p data:%p "
			       "tail:%p end:%p sum:%d dev:%s",
				skb->len, skb->data_len,
				skb->head, skb->data, skb_tail_pointer(skb),
				skb_end_pointer(skb), skb->csum,
				skb->dev ? skb->dev->name : "<NULL>");
		}

		/*
		 * Save source MAC address before discarding header.
		 */
 		fc = fc_dev->drv_priv;
		if (unlikely(fc->flogi_progress))
			mac = net48_get((net48_t *) eth_hdr(skb)->h_source);

		/*
		 * Check the header and pull it off.
		 */
		hlen = ((struct fcoe_softc *)fc_dev->drv_priv)->fcoe_hlen;
		if (hlen == sizeof(struct fcoe_hdr)) {
			struct fcoe_hdr *hp;

			hp = (struct fcoe_hdr *)skb->data;
			if (unlikely(FC_FCOE_DECAPS_VER(hp) != FC_FCOE_VER)) {
				if (stats->ErrorFrames < 5)
					OFC_DBG("unknown FCoE version %x",
					       FC_FCOE_DECAPS_VER(hp));
				stats->ErrorFrames++;
				kfree_skb(skb);
				continue;
			}
			sof = hp->fcoe_sof;
			skb_pull(skb, sizeof(*hp));
			fr_len = skb->len - sizeof(struct fcoe_crc_eof);
			tlen = sizeof(struct fcoe_crc_eof);
		} else {
			struct fcoe_hdr_old *fchp;
			u_int len;

			ASSERT(hlen == sizeof(*fchp));
			fchp = (struct fcoe_hdr_old *)skb->data;
			len = net16_get(&fchp->fcoe_plen);
			skb_pull(skb, sizeof(*fchp));
			fr_len = FC_FCOE_DECAPS_LEN(len);
			fr_len = fr_len * FCOE_WORD_TO_BYTE;
			if (unlikely(fr_len < sizeof(struct fc_frame_header) +
			    sizeof(cp->fcoe_crc32))) {
				if (stats->ErrorFrames < 5)
					OFC_DBG("length error: len_sof %x", len);
				stats->ErrorFrames++;
				kfree_skb(skb);
				continue;
			}
			sof = FC_FCOE_DECAPS_SOF(len);
			fr_len -= sizeof(cp->fcoe_crc32);
			tlen = sizeof(struct fcoe_crc_eof_old);
		}

		if (skb_is_nonlinear(skb))
			skb_linearize(skb);	/* not ideal */
		if (unlikely(fr_len + tlen > skb->len)) {
			if (stats->ErrorFrames < 5)
				OFC_DBG("length error fr_len 0x%x skb->len 0x%x",
				       fr_len + tlen, skb->len);
			stats->ErrorFrames++;
			kfree_skb(skb);
			continue;
		}
		stats->RxFrames++;
		stats->RxWords += fr_len / FCOE_WORD_TO_BYTE;

		fp = &p->frame;
		fc_frame_init_static(fp);
		fp->fr_hdr = (struct fc_frame_header *)skb->data;
		fp->fr_len = fr_len;
		cp = (struct fcoe_crc_eof *)(skb->data + fr_len);
		fp->fr_eof = cp->fcoe_eof;
		fp->fr_sof = sof;

		/*
		 * Check the CRC here, unless it's solicited data for SCSI.
		 * In that case, the SCSI layer can check it during the copy,
		 * and it'll be more cache-efficient.
		 */
		fh = fc_frame_header_get(fp);
		ASSERT(fh);
		if (fh->fh_r_ctl == FC_RCTL_DD_SOL_DATA &&
		    fh->fh_type == FC_TYPE_FCP) {
			fp->fr_flags |= FCPHF_CRC_UNCHECKED;
			openfc_rcv(fc_dev, fp);
		} else if (cp->fcoe_crc32 ==
			   ~crc32_sb8_64_bit(~0, skb->data, fr_len)) {
			if (unlikely(fc->flogi_progress))
				fcoe_recv_flogi(fc, fp, mac);
			openfc_rcv(fc_dev, fp);
		} else {
			if (debug_fcoe || stats->InvalidCRCCount < 5) {
				printk(KERN_WARNING \
					"fcoe: dropping frame with CRC error");
				fc_print_frame_hdr("frame with CRC error", fp);
			}
			stats->InvalidCRCCount++;
			stats->ErrorFrames++;
			fc_frame_free(fp);
		}
		ASSERT(fc_frame_freed_static(fp));
		kfree_skb(skb);
	}
	return 0;
}
			
/*
 * Snoop potential response to FLOGI or even incoming FLOGI.
 */
static void fcoe_recv_flogi(struct fcoe_softc *fc, struct fc_frame *fp, u64 sa)
{
	struct fc_frame_header *fh;
	u8 op;

	fh = fc_frame_header_get(fp);
	if (fh->fh_type != FC_TYPE_ELS)
		return;
	op = fc_frame_payload_op(fp);
	if (op == ELS_LS_ACC && fh->fh_r_ctl == FC_RCTL_ELS_REP &&
	    fc->flogi_oxid == net16_get(&fh->fh_ox_id)) {
		/*
		 * FLOGI accepted.
		 * If the src mac addr is FC_OUI-based, then we mark the
		 * address_mode flag to use FC_OUI-based Ethernet DA.
		 * Otherwise we use the FCoE gateway addr
		 */
		if (sa == FC_FCOE_FLOGI_MAC) {
			fc->address_mode = FCOE_FCOUI_ADDR_MODE;
		} else {
			net48_put((net48_t *)&fc->dest_addr, sa);
			fc->address_mode = FCOE_GW_ADDR_MODE;
		}
#ifdef HAVE_SET_RX_MODE
		/*
		 * Remove any previously-set unicast MAC filter.
		 * Add secondary FCoE MAC address filter for our OUI.
		 */
		rtnl_lock();
		if (net48_get((net48_t *)&fc->data_src_addr))
			dev_unicast_delete(fc->real_dev, fc->data_src_addr,
					   ETH_ALEN);
		fc_fcoe_set_mac(fc->data_src_addr, (net24_t *) &fh->fh_d_id);
		dev_unicast_add(fc->real_dev, fc->data_src_addr, ETH_ALEN);
		rtnl_unlock();
#else
		fc_fcoe_set_mac(fc->data_src_addr, (net24_t *) &fh->fh_d_id);
#endif /* HAVE_SET_RX_MODE */
		fc->flogi_progress = 0;
	} else if (op == ELS_FLOGI && fh->fh_r_ctl == FC_RCTL_ELS_REQ && sa) {
		/*
		 * Save source MAC for point-to-point responses.
		 */
		net48_put((net48_t *)&fc->dest_addr, sa);
		fc->address_mode = FCOE_GW_ADDR_MODE;
	}
}

void fcoe_watchdog(ulong vp)
{
	struct fcdev *fd;
	struct fcoe_softc *fc;
	struct fcoe_info *fci = &fcoei;
	int paused = 0;

	read_lock(&fci->fcoe_hostlist_lock);
	list_for_each_entry(fc, &fci->fcoe_hostlist, list) {
		fd = fc->fd;
		if (fd) {
			if (fc->fcoe_pending_queue.qlen > FCOE_MAX_QUEUE_DEPTH)
				paused = 1;
			if (fcoe_check_wait_queue(fd) <  FCOE_MAX_QUEUE_DEPTH) {
				if (paused)
					openfc_unpause(fd);
			}
		}
	}
	read_unlock(&fci->fcoe_hostlist_lock);

	fci->timer.expires = jiffies + (1 * HZ);
	add_timer(&fci->timer);
}

/*
 * the wait_queue is used when the skb transmit fails. skb will go
 * in the wait_queue which will be emptied by the time function OR
 * by the next skb transmit.
 *
 */

/*
 * Function name : fcoe_check_wait_queue()
 *
 * Return Values : 0 or error
 *
 * Description	 : empties the wait_queue
 *		   dequeue the head of the wait_queue queue and
 *		   calls fcoe_start_io() for each packet
 *		   if all skb have been transmitted, return 0
 *		   if a error occurs, then restore wait_queue and try again
 *		   later
 *
 */

static int fcoe_check_wait_queue(struct fcdev *fd)
{
	int rc, unpause = 0;
	int paused = 0;
	struct sk_buff *skb;
	struct fcoe_softc *fc;

	fc = (struct fcoe_softc *)fd->drv_priv;
	spin_lock_bh(&fc->fcoe_pending_queue.lock);

	/*
 	 * is this interface paused?
 	 */
	if (fc->fcoe_pending_queue.qlen > FCOE_MAX_QUEUE_DEPTH)
		paused = 1;	
	if (fc->fcoe_pending_queue.qlen) {
		while ((skb = __skb_dequeue(&fc->fcoe_pending_queue)) != NULL) {
			spin_unlock_bh(&fc->fcoe_pending_queue.lock);
			rc = fcoe_start_io(skb);
			if (rc) {
				fcoe_insert_wait_queue_head(fd, skb);
				return rc;
			}
			spin_lock_bh(&fc->fcoe_pending_queue.lock);
		}
		if (fc->fcoe_pending_queue.qlen < FCOE_MAX_QUEUE_DEPTH)
			unpause = 1;
	}
	spin_unlock_bh(&fc->fcoe_pending_queue.lock);
	if ((unpause) && (paused))
		openfc_unpause(fd);
	return fc->fcoe_pending_queue.qlen;
}

static void fcoe_insert_wait_queue_head(struct fcdev *fd, struct sk_buff *skb)
{
	struct fcoe_softc *fc;

	fc = (struct fcoe_softc *)fd->drv_priv;
	spin_lock_bh(&fc->fcoe_pending_queue.lock);
	__skb_queue_head(&fc->fcoe_pending_queue, skb);
	spin_unlock_bh(&fc->fcoe_pending_queue.lock);
}

static void fcoe_insert_wait_queue(struct fcdev *fd, struct sk_buff *skb)
{
	struct fcoe_softc *fc;

	fc = (struct fcoe_softc *)fd->drv_priv;
	spin_lock_bh(&fc->fcoe_pending_queue.lock);
	__skb_queue_tail(&fc->fcoe_pending_queue, skb);
	spin_unlock_bh(&fc->fcoe_pending_queue.lock);
}
