/*
 * Copyright(c) 2007 Intel Corporation. All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Maintained at www.Open-FCoE.org
 */

/*
 * FCOE protocol file
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <scsi/scsi_tcq.h>
#include <scsi/scsicam.h>
#include <scsi/scsi_transport.h>
#include <scsi/scsi_transport_fc.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/if_ether.h>
#include <linux/if_vlan.h>
#include <net/rtnetlink.h>

/*
 * Non Linux header definitions
 */
#include "sa_assert.h"
#include "fc_types.h"
#include "fc_frame.h"
#include "fc_print.h"
#include "fc_encaps.h"
#define FCOE_T11_AUG07		/* use new FCoE version */
#include "fc_fcoe.h"
#include "fc_fcoe_old.h"
#include "fc_fs.h"
#include "fc_els.h"
#include "fcdev.h"
#include "fcoeioctl.h"
#include "openfc_ioctl.h"
#include "ofc_dbg.h"
#include "fcoe_def.h"

#ifndef BUILD_VERSION
#define BUILD_VERSION ""
#endif /* BUILD_VERSION */

/*
 * transport functions
 */
static struct openfc_port_operations fcoe_port_ops = {
	.owner = THIS_MODULE,
	.send = fcoe_xmit,
	.frame_alloc = fcoe_frame_alloc,
};

/*
 * debug = 1 --> enable debug
 * debug = 0 --> disable debug
 */
int debug_fcoe;
extern struct fcoe_percpu_s *fcoe_percpu[];
/*
 * This is called from driver exit routine or deleting an interface
 * this finction cleans up all registrations and
 * frees up all alocated memory
 * 1) disables receiving packets from the networks
 * 2) remove "fc_dev" from the device list so that receive processing thread
 *    will drop all packet that has been queued
 * 3) unregister from openfc
 * 4) free per cpu stats structure
 * 5) call openfc to free the fc_dev structure
 * 6) Can not be called with lock held
 */
int fcoe_destroy_interface(struct fcdev *fd)
{
	int cpu, idx;
	struct fcoe_dev_stats *p;
	struct fcoe_percpu_s *pp;
	struct fcoe_softc *fc;
	struct fcoe_rcv_info *fr;
	struct fcoe_info *fci = &fcoei;
	struct sk_buff_head *list;
	struct sk_buff *skb, *next;
	struct sk_buff *head;
#ifdef HAVE_SET_RX_MODE
	net8_t flogi_maddr[ETH_ALEN];
#endif

	ASSERT(fd != NULL);

	openfc_unregister(fd);

	fc = (struct fcoe_softc *)fd->drv_priv;

	/*
	 * cleanup the openfc hba i/f
	 */
	dev_remove_pack(&fc->fcoe_packet_type);

	rtnl_lock();
#ifdef HAVE_SET_RX_MODE
	/*
	 * delete secondary MAC addresses
	 */
	net48_put((net48_t *) flogi_maddr, FC_FCOE_FLOGI_MAC);
	dev_unicast_delete(fc->real_dev, flogi_maddr, ETH_ALEN);
	if (net48_get((net48_t *) fc->data_src_addr))
		dev_unicast_delete(fc->real_dev, fc->data_src_addr, ETH_ALEN);
#else
	/* turn off promscuity */
	dev_set_promiscuity(fc->real_dev, -1);
#endif
	rtnl_unlock();

	for (idx = 0; idx < NR_CPUS; idx++) {
		if (fci->fcoe_percpu[idx]) {
			pp = fci->fcoe_percpu[idx];
			spin_lock_bh(&pp->fcoe_rx_list.lock);
			list = &pp->fcoe_rx_list;
			head = list->next;
			for (skb = head; skb != (struct sk_buff *)list;
			     skb = next) {
				next = skb->next;
				fr = fcoe_dev_from_skb(skb);
				if (fr->fd == fc->fd) {
					__skb_unlink(skb, list);
					kfree_skb(skb);
				}
			}
			spin_unlock_bh(&pp->fcoe_rx_list.lock);
		}
	}

	fcoe_clean_pending_queue(fd);
	for_each_online_cpu(cpu) {
		p = fd->dev_stats[cpu];
		if (p) {
			fd->dev_stats[cpu] = NULL;
			kfree(p);
		}
	}

	fcoe_put_dev(fd);
	return 0;
}

/*
 * Return non-zero if link is OK for use by FCoE.
 * Any permanently-disqualifying conditions have been previously checked.
 * This checks pause settings, which can change with link.
 * This also updates the speed setting, which may change with link for 100/1000.
 */
int fcoe_link_ok(struct fcdev *fdev)
{
	struct fcoe_softc *fc = (struct fcoe_softc *)fdev->drv_priv;
	struct net_device *dev = fc->real_dev;
	struct ethtool_pauseparam pause = { ETHTOOL_GPAUSEPARAM };
	struct ethtool_cmd ecmd = { ETHTOOL_GSET };
	int rc = 0;

	if ((dev->flags & IFF_UP) && netif_carrier_ok(dev)) {
		dev = fc->phys_dev;
		dev->ethtool_ops->get_pauseparam(dev, &pause);
		if (dev->ethtool_ops->get_settings) {
			dev->ethtool_ops->get_settings(dev, &ecmd);
			fdev->fd_speed_support &=
			    ~(OFC_SPEED_1GBIT | OFC_SPEED_10GBIT);
			if (ecmd.supported & (SUPPORTED_1000baseT_Half |
					      SUPPORTED_1000baseT_Full))
				fdev->fd_speed_support |= OFC_SPEED_1GBIT;
			if (ecmd.supported & SUPPORTED_10000baseT_Full)
				fdev->fd_speed_support |= OFC_SPEED_10GBIT;
			if (ecmd.speed == SPEED_1000)
				fdev->fd_speed = OFC_SPEED_1GBIT;
			if (ecmd.speed == SPEED_10000)
				fdev->fd_speed = OFC_SPEED_10GBIT;

			/*
			 * for 10 G (and faster), ignore autoneg requirement.
			 */
			if (ecmd.speed >= SPEED_10000)
				pause.autoneg = 1;
		}
		if (pause.autoneg && pause.tx_pause && pause.rx_pause)
			rc = 1;
	}
	return rc;
}

/*
 * This function creates the fcoe interface
 * create struct fcdev which is a shared structure between opefc
 * and transport level protocol.
 */

int fcoe_create_interface(struct fcoe_info *fci, void *ptr)
{
	struct fcdev *fdev;
	struct fcoe_cfg *cfg = ptr;
	struct fcoe_softc *fc;
	struct fcoe_dev_stats *p;
	struct net_device *net_dev;
	int rc = 0;
	int i;
#ifdef HAVE_SET_RX_MODE
	net8_t flogi_maddr[ETH_ALEN];
#endif

	ASSERT(fci != NULL);
	fdev = openfc_alloc_dev(&fcoe_port_ops, sizeof(struct fcoe_softc));
	if (unlikely(!fdev)) {
		OFC_DBG("couldn't allocate space for hba struct");
		rc = -ENOMEM;
		goto out;
	}

	fc = (struct fcoe_softc *)fdev->drv_priv;
	fc->fcoe_hlen = (cfg->fcoe_type & FCOE_CFG_TYPE_T11) ?
	    sizeof(struct fcoe_hdr) : sizeof(struct fcoe_hdr_old);

	/*
	 * get the pointer the real eth device
	 */
	fc->real_dev = (void *)dev_get_by_name(&init_net, cfg->local_ifname);
	if (fc->real_dev == NULL) {
		OFC_DBG("could not get network device for %s",
		       cfg->local_ifname);
		rc = -ENODEV;
		goto out;
	}

	if (fcoe_find_fcdev(fc->real_dev->name) != NULL) {
		rc = -EEXIST;
		goto out;
	}

	/*
	 * Do not support for bonding device
	 */
	if ((fc->real_dev->priv_flags & IFF_MASTER_ALB) ||
	    (fc->real_dev->priv_flags & IFF_SLAVE_INACTIVE) ||
	    (fc->real_dev->priv_flags & IFF_MASTER_8023AD)) {
		rc = -EOPNOTSUPP;
		goto out;
	}

	/*
	 * Require support for get_pauseparam ethtool op.
	 */
	net_dev = fc->real_dev;
	if (!net_dev->ethtool_ops && (net_dev->priv_flags & IFF_802_1Q_VLAN))
		net_dev = VLAN_DEV_INFO(net_dev)->real_dev;
	if (!net_dev->ethtool_ops || !net_dev->ethtool_ops->get_pauseparam) {
		rc = -EOPNOTSUPP;
		goto out;
	}
	fc->phys_dev = net_dev;
	fdev->fd_link_status = TRANS_LINK_DOWN;
	if (fcoe_link_ok(fdev))
		fdev->fd_link_status = TRANS_LINK_UP;

	if (fc->real_dev->features & NETIF_F_SG)
		fdev->capabilities = TRANS_C_SG;

	skb_queue_head_init(&fc->fcoe_pending_queue);

	/*
	 * copy the ifname
	 */
	memcpy(fdev->ifname, fc->real_dev->name, IFNAMSIZ);

	write_lock_bh(&fci->fcoe_hostlist_lock);
	list_add_tail(&fc->list, &fci->fcoe_hostlist);
	write_unlock_bh(&fci->fcoe_hostlist_lock);
	/*
	 * setup Source Mac Address
	 */
	memcpy(fc->ctl_src_addr, fc->real_dev->dev_addr,
	       fc->real_dev->addr_len);
	fdev->fd_wwnn = cfg->wwnn;
	fdev->fd_wwpn = cfg->wwpn;
	fdev->framesize = fc->real_dev->mtu - (sizeof(struct fcoe_hdr) +
					       sizeof(struct fcoe_crc_eof));
	if (cfg->fcoe_mtu >= FC_MIN_MAX_FRAME) {
		cfg->fcoe_mtu += sizeof(struct fc_frame_header);
		if (cfg->fcoe_mtu < fdev->framesize)
			fdev->framesize = cfg->fcoe_mtu;
	}

	/*
	 * FCoE protocol receives frames on multiple unicast MAC Addresses.
	 */
	rtnl_lock();
#ifdef HAVE_SET_RX_MODE
	/*
	 * Add FLOGI MAC address
	 */
	net48_put((net48_t *) flogi_maddr, FC_FCOE_FLOGI_MAC);
	dev_unicast_add(fc->real_dev, flogi_maddr, ETH_ALEN);
#else
	/*
	 * Additional unicast MAC address is not supported in Linux
	 * older than Linux 2.6.23, therefore  turn on promscuity
	 */
	dev_set_promiscuity(fc->real_dev, 1);
#endif
	rtnl_unlock();

	snprintf(fdev->drv_info.model, 64, FCOE_DRIVER_NAME);
	snprintf(fdev->drv_info.vendor, 64, FCOE_DRIVER_VENDOR);
	snprintf(fdev->drv_info.model_desc, 64, FCOE_DRIVER_NAME);
	snprintf(fdev->drv_info.drv_version, 64, __stringify(BUILD_VERSION));
	snprintf(fdev->drv_info.drv_name, 64, fdev->ifname);

	/*
	 * allocate per cpu stats block
	 */
	for_each_online_cpu(i) {
		p = kzalloc(sizeof(struct fcoe_dev_stats), GFP_KERNEL);
		if (p)
			fdev->dev_stats[i] = p;
	}

	/*
	 * setup
	 * receive function from ether net driver
	 * ether type for the given device
	 */
	fc->fcoe_packet_type.func = fcoe_rcv;
	fc->fcoe_packet_type.type = __constant_htons(ETH_P_FCOE);
	fc->fcoe_packet_type.dev = fc->real_dev;
	dev_add_pack(&fc->fcoe_packet_type);
	fc->fd = fdev;

	/*
	 * call to openfc for device registration
	 */
	rc = openfc_register(fdev);
	return rc;
out:
	fcoe_put_dev(fdev);
	return rc;
}

void fcoe_put_dev(struct fcdev *dev)
{
	struct net_device *real_dev;
	struct fcoe_softc *fc;

	if (dev) {
		fc = (struct fcoe_softc *)dev->drv_priv;
		real_dev = fc->real_dev;

		openfc_put_dev(dev);

		if (real_dev)
			dev_put(real_dev);
	}

	return;
}

struct fcoe_softc *fcoe_find_fcdev(char *name)
{
	struct fcoe_softc *fc;
	struct fcdev *fd;
	struct fcoe_info *fci = &fcoei;

	read_lock(&fci->fcoe_hostlist_lock);
	list_for_each_entry(fc, &fci->fcoe_hostlist, list) {
		fd = fc->fd;
		if (!strncmp(name, fd->ifname, IFNAMSIZ)) {
			read_unlock(&fci->fcoe_hostlist_lock);
			return (fc);
		}
	}
	read_unlock(&fci->fcoe_hostlist_lock);
	return NULL;
}

void fcoe_clean_pending_queue(struct fcdev *fd)
{
	struct fcoe_softc  *fc = fd->drv_priv;
	struct sk_buff *skb;

	spin_lock_bh(&fc->fcoe_pending_queue.lock);
	while ((skb = __skb_dequeue(&fc->fcoe_pending_queue)) != NULL) {
		spin_unlock_bh(&fc->fcoe_pending_queue.lock);
		kfree_skb(skb);
		spin_lock_bh(&fc->fcoe_pending_queue.lock);
	}
	spin_unlock_bh(&fc->fcoe_pending_queue.lock);
}
