/*
 * Copyright(c) 2007 Intel Corporation. All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Maintained at www.Open-FCoE.org
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/spinlock.h>
#include <linux/cpu.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/if_ether.h>
#include <linux/fs.h>

#include "sa_assert.h"
#include "fc_types.h"
#include "fc_frame.h"
#include "fcdev.h"
#include "fcoeioctl.h"
#include "ofc_dbg.h"

#include "fcoe_def.h"

MODULE_AUTHOR("Open-FC.org");
MODULE_DESCRIPTION("FCOE Protocol driver");
MODULE_LICENSE("GPL");

/*
 * Static functions and variables definations
 */
#ifdef CONFIG_HOTPLUG_CPU
static int fcoe_cpu_callback(struct notifier_block *, ulong, void *);
#endif /* CONFIG_HOTPLUG_CPU */
static int fcoe_device_notification(struct notifier_block *, ulong, void *);
extern int fcoe_percpu_receive_thread(void *arg);
extern void fcoe_unreg_char_dev(void);
extern int fcoe_reg_char_dev(void);
extern void msleep(unsigned int);

static void fcoe_dev_setup(void);
static void fcoe_dev_cleanup(void);

struct fcoe_info fcoei = {
	.fcoe_hostlist = LIST_HEAD_INIT(fcoei.fcoe_hostlist),
};

//struct fcoe_info *fci = &fcoei;
struct fcoe_percpu_s *fcoe_percpu[NR_CPUS];

#ifdef CONFIG_HOTPLUG_CPU
static struct notifier_block fcoe_cpu_notifier = {
	.notifier_call = fcoe_cpu_callback,
};
#endif /* CONFIG_HOTPLUG_CPU */

/*
 * notification function from net device
 */
static struct notifier_block fcoe_notifier = {
	.notifier_call = fcoe_device_notification,
};

#ifdef CONFIG_HOTPLUG_CPU
/*
 * create percpu stats block
 * called by cpu add/remove notifier
 */
static void fcoe_create_percpu_data(int cpu)
{
	struct fcdev *fc_dev;
	struct fcoe_softc *fc;
	struct fcoe_dev_stats *p;
	struct fcoe_info *fci = &fcoei;

	write_lock_bh(&fci->fcoe_hostlist_lock);
	list_for_each_entry(fc, &fci->fcoe_hostlist, list) {
		fc_dev = fc->fd;
		if (fc_dev->dev_stats[cpu] == NULL) {
			p = kzalloc(sizeof(struct fcoe_dev_stats), GFP_KERNEL);
			if (p)
				fc_dev->dev_stats[cpu] = p;
		}
	}
	write_unlock_bh(&fci->fcoe_hostlist_lock);
}

/*
 * destroy percpu stats block
 * called by cpu add/remove notifier
 */
static void fcoe_destroy_percpu_data(int cpu)
{
	struct fcoe_dev_stats *p;
	struct fcdev *fc_dev;
	struct fcoe_softc *fc;
	struct fcoe_info *fci = &fcoei;

	write_lock_bh(&fci->fcoe_hostlist_lock);
	list_for_each_entry(fc, &fci->fcoe_hostlist, list) {
		fc_dev = fc->fd;
		p = fc_dev->dev_stats[cpu];
		if (p != NULL) {
			fc_dev->dev_stats[cpu] = NULL;
			kfree(p);
		}
	}
	write_unlock_bh(&fci->fcoe_hostlist_lock);
}

/*
 * Get notified when a cpu comes on/off. Be hotplug friendly.
 */
static int fcoe_cpu_callback(struct notifier_block *nfb, unsigned long action,
			     void *hcpu)
{
	unsigned int cpu = (unsigned long)hcpu;

	switch (action) {
	case CPU_ONLINE:
		fcoe_create_percpu_data(cpu);
		break;
	case CPU_DEAD:
		fcoe_destroy_percpu_data(cpu);
		break;
	default:
		break;
	}
	return NOTIFY_OK;
}
#endif /* CONFIG_HOTPLUG_CPU */

/*
 * function to setup link change notification interface
 */
static void fcoe_dev_setup(void)
{
	/*
	 * here setup a interface specific wd time to
	 * monitor the link state
	 */
	register_netdevice_notifier(&fcoe_notifier);
}

/*
 * function to cleanup link change notification interface
 */
static void fcoe_dev_cleanup(void)
{
	unregister_netdevice_notifier(&fcoe_notifier);
}

/*
 * This function is called by the ethernet driver
 * this is called in case of link change event
 */
static int fcoe_device_notification(struct notifier_block *notifier,
				    ulong event, void *ptr)
{
	struct net_device *real_dev = ptr;
	struct fcoe_softc *fc;
	struct fcdev *fc_dev = NULL;
	struct fcoe_dev_stats *stats;
	struct fcoe_info *fci = &fcoei;
	int new_status;
	int rc;

	read_lock(&fci->fcoe_hostlist_lock);
	list_for_each_entry(fc, &fci->fcoe_hostlist, list) {
		if (fc->real_dev == real_dev) {
			fc_dev = fc->fd;
			break;
		}
	}
	read_unlock(&fci->fcoe_hostlist_lock);
	if (fc_dev == NULL) {
		rc = NOTIFY_DONE;
		goto out;
	}
	new_status = fc_dev->fd_link_status;
	switch (event) {
	case NETDEV_DOWN:
	case NETDEV_GOING_DOWN:
		new_status = TRANS_LINK_DOWN;
		break;
	case NETDEV_UP:
	case NETDEV_CHANGE:
		new_status = TRANS_LINK_DOWN;
		if (fcoe_link_ok(fc_dev))
			new_status = TRANS_LINK_UP;
		break;
	case NETDEV_REGISTER:
		rc = NOTIFY_OK;
		break;
	default:
		OFC_DBG("unknown event %ld call", event);
		rc = NOTIFY_OK;
	}
	if (fc_dev->fd_link_status != new_status) {
		fc_dev->fd_link_status = new_status;
		if (new_status == TRANS_LINK_UP) {
			openfc_linkup(fc_dev);
		} else {
			stats = fc_dev->dev_stats[smp_processor_id()];
			stats->LinkFailureCount++;
			openfc_linkdown(fc_dev);
			fcoe_clean_pending_queue(fc_dev);
		}
	}
out:
	return rc;
}

/*
 * Initialization routine
 * 1. Will create fc transport software structure
 * 2. initialize the link list of port information structure
 */
static int __init fcoeinit(void)
{
	int rc = 0;
	int cpu;
	struct fcoe_percpu_s *p;
	struct fcoe_info *fci = &fcoei;

	rwlock_init(&fci->fcoe_hostlist_lock);
	atomic_set(&fci->cleanup_needed, 0);
	/*
	 * This routine will create the char dev entry and a major number
	 * returns 0 if major number for char dev is greater then -1
	 */
	fci->fcoe_major = fcoe_reg_char_dev();
	if (fci->fcoe_major < 0) {
		OFC_DBG("failed to register the control device\n");
		rc = -ENODEV;
		goto out;
	}
#ifdef CONFIG_HOTPLUG_CPU
	register_cpu_notifier(&fcoe_cpu_notifier);
#endif /* CONFIG_HOTPLUG_CPU */
	fci->fcoe_class = class_create(THIS_MODULE, FCOE_CLASS_NAME);
	if (IS_ERR(fci->fcoe_class)) {
		rc = -ENODEV;
		goto out_chrdev;
	}
	class_device_create(fci->fcoe_class, NULL,
			    MKDEV(fci->fcoe_major, 0), NULL, FCOE_CTL_DEV_NAME);

	/*
	 * initialize per CPU interrupt thread
	 */
	for_each_online_cpu(cpu) {
		if (cpu > openfc_cpu_mask)
			break;

		p = kzalloc(sizeof(struct fcoe_percpu_s), GFP_KERNEL);
		if (p) {
			p->thread = kthread_create(fcoe_percpu_receive_thread,
						   (void *)p,
						   "fcoethread/%d", cpu);

			/*
			 * if there is no error then bind the thread to the cpu
			 * initialize the semaphore and skb queue head
			 */
			if (likely(!IS_ERR(p->thread))) {
				p->cpu = cpu;
				fci->fcoe_percpu[cpu] = p;
				skb_queue_head_init(&p->fcoe_rx_list);
				kthread_bind(p->thread, cpu);
				wake_up_process(p->thread);
			} else {
				fci->fcoe_percpu[cpu] = NULL;
				kfree(p);

			}
		}
	}
	if (rc < 0) {
		OFC_DBG("failed to initialize proc intrerface\n");
		rc = -ENODEV;
		goto out_chrdev;
	}

	/*
	 * setup link change notification
	 */
	fcoe_dev_setup();

	init_timer(&fci->timer);
	fci->timer.data = (ulong) fci;
	fci->timer.function = fcoe_watchdog;
	fci->timer.expires = (jiffies + (10 * HZ));
	add_timer(&fci->timer);

	return 0;

out_chrdev:
#ifdef CONFIG_HOTPLUG_CPU
	unregister_cpu_notifier(&fcoe_cpu_notifier);
#endif /* CONFIG_HOTPLUG_CPU */
	fcoe_unreg_char_dev();
out:
	return rc;
}

static void __exit fcoe_exit(void)
{
	uint32_t idx;
	struct fcoe_softc *fc, *tmp;
	struct fcoe_info *fci = &fcoei;
	struct fcoe_percpu_s *p;
	struct sk_buff *skb;

	/*
	 * Stop all call back interfaces
	 */
#ifdef CONFIG_HOTPLUG_CPU
	unregister_cpu_notifier(&fcoe_cpu_notifier);
#endif /* CONFIG_HOTPLUG_CPU */
	class_device_destroy(fci->fcoe_class, MKDEV(fci->fcoe_major, 0));
	class_destroy(fci->fcoe_class);
	fcoe_unreg_char_dev();
	fcoe_dev_cleanup();

	/*
	 * stop timer
	 */
	del_timer_sync(&fci->timer);

	/*
	 * assuming that at this time there will be no
	 * ioctl in prograss, therefore we do not need to lock the
	 * list.
	 */
	list_for_each_entry_safe(fc, tmp, &fci->fcoe_hostlist, list) {
		list_del(&fc->list);
		fcoe_destroy_interface(fc->fd);
	}

	for (idx = 0; idx < NR_CPUS; idx++) {
		if (fci->fcoe_percpu[idx]) {
			kthread_stop(fci->fcoe_percpu[idx]->thread);
			p = fci->fcoe_percpu[idx];
			spin_lock_bh(&p->fcoe_rx_list.lock);
			while ((skb = __skb_dequeue(&p->fcoe_rx_list)) != NULL)
				kfree_skb(skb);
			spin_unlock_bh(&p->fcoe_rx_list.lock);
			kfree(fci->fcoe_percpu[idx]);
		}
	}

	/*
	 * wait until all the skbs are free
	 */
	while (atomic_read(&fci->cleanup_needed) != 0) {
		msleep(100);
		printk(" Waiting for %d skbs free \n",
		       atomic_read(&fci->cleanup_needed));
	}

}

module_init(fcoeinit);
module_exit(fcoe_exit);
