/*
 * Copyright(c) 2007 Intel Corporation. All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Maintained at www.Open-FCoE.org
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/fs.h>

#include "sa_assert.h"
#include "fc_types.h"
#include "fc_frame.h"
#include "fcdev.h"
#include "fcoeioctl.h"
#include "ofc_dbg.h"

#include "fcoe_def.h"

#define FCOE_CTL_DEV_NAME  "fcoe"
/*
 * using this ioctl interface user can create multiple
 * instance of fcoe hba (dcehba). each hba structure will
 * be mapped to a ethernet interface.  persistent binding
 * table information and fc configuration information
 * will be passed to opfc driver.
 */
int fcoe_reg_char_dev(void);
void fcoe_unreg_char_dev(void);
int fcoe_set_config(void __user *);
int fcoe_delete_config(void __user *);
/*
 * Static functions and variables definations
 */
static int fcoe_ioctl(struct inode *, struct file *, u_int32_t, ulong);
static int fcoe_open(struct inode *, struct file *);
static int fcoe_close(struct inode *, struct file *);

#ifdef CONFIG_COMPAT
static long fcoe_compat_ioctl(struct file *, u_int32_t, ulong);
static int do_ioctl(struct file *, unsigned, unsigned long);
#endif

static struct file_operations fcoe_file_ops = {
	.owner = THIS_MODULE,
	.open = fcoe_open,
	.release = fcoe_close,
	.ioctl = fcoe_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = fcoe_compat_ioctl,
#endif

};

/*
 *  fcoe driver's open routine
 */
static int fcoe_open(struct inode *inode, struct file *file)
{
	return 0;
}

/*
 *  fcoe driver's close routine
 */
static int fcoe_close(struct inode *inode, struct file *file)
{
	return 0;
}

/*
 * IOCTL interface of the lower level driver
 * THis is only needed by fcip and FCoE transport
 */
static int fcoe_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		      unsigned long arg)
{
	int rc = 0;
	void __user *_arg = (void __user *)arg;

	if (!capable(CAP_SYS_ADMIN))
		return -EACCES;

	if (_IOC_TYPE(cmd) != FCOE_IOCTL_MAGIC)
		return -ENOTTY;

	if (cmd == FCOE_CREATE_INTERFACE)
		rc = fcoe_set_config(_arg);
	else if (cmd == FCOE_DESTROY_INTERFACE)
		rc = fcoe_delete_config(_arg);
	else
		rc = -ENOIOCTLCMD;

	return rc;
}

#ifdef CONFIG_COMPAT

static int do_ioctl(struct file *f, unsigned cmd, unsigned long arg)
{
	int rc;
	lock_kernel();
	rc = fcoe_ioctl(f->f_dentry->d_inode, f, cmd, arg);
	unlock_kernel();
	return rc;
}

static long fcoe_compat_ioctl(struct file *file, unsigned int cmd,
			      unsigned long arg)
{
	long rc = 0;

	switch (cmd) {
	case FCOE_CREATE_INTERFACE:
	case FCOE_DESTROY_INTERFACE:
		rc = do_ioctl(file, cmd, arg);
		break;
	default:
		rc = -ENOIOCTLCMD;
	}

	return rc;
}
#endif

int fcoe_reg_char_dev(void)
{
	/*
	 * This routine will create the char dev entry and a major number
	 * returns 0 if major number for char dev is greater then -1
	 */
	return register_chrdev(0, FCOE_CTL_DEV_NAME, &fcoe_file_ops);
}

void fcoe_unreg_char_dev(void)
{
	struct fcoe_info *fci = &fcoei;

	unregister_chrdev(fci->fcoe_major, FCOE_CTL_DEV_NAME);
}

/*
 * This routine is for configuration of fcoe/openfc instance
 */
int fcoe_set_config(void __user *arg)
{
	int rc;
	struct fcoe_cfg cfg;
	struct fcoe_info *fci = &fcoei;

	if (copy_from_user(&cfg, arg, sizeof(struct fcoe_cfg))) {
		return -EFAULT;
	}
	rc = fcoe_create_interface(fci, &cfg);
	return rc;
}

int fcoe_delete_config(void __user *arg)
{
	int rc = -ENXIO;
	struct fcoe_cfg cfg;
	struct fcoe_info *fci = &fcoei;
	struct fcoe_softc *fc;

	if (copy_from_user(&cfg, arg, sizeof(struct fcoe_cfg)))
		return -EFAULT;

	fc = fcoe_find_fcdev(cfg.local_ifname);
	if (fc && fc->fd) {
		write_lock_bh(&fci->fcoe_hostlist_lock);
		list_del(&fc->list);
		write_unlock_bh(&fci->fcoe_hostlist_lock);
		rc = fcoe_destroy_interface(fc->fd);
	}
	return rc;
}
