/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/log2.h>
#include <scsi/scsi.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport.h>
#include <scsi/scsi_transport_fc.h>
#include <scsi/scsi_tcq.h>
#include <linux/mlx4/driver.h>
#include <linux/mlx4/cmd.h>
#include <linux/mlx4/cq.h>

#include "mlx4_fc.h"
#include "mlx4_ofc.h"
#include "bitmap.h"
#include "fc_frame.h"
#include "fc_els.h"
#include "fc_remote_port.h"
#include "../openfc/openfc_scsi_pkt.h"
#include "../openfc/fc_sess_impl.h"

static enum mlx4_fcmd_io_dir scsi_dir_translate(enum dma_data_direction dmadir)
{
	switch (dmadir) {
	case DMA_BIDIRECTIONAL:
		return FCMD_IO_DIR_BIDI;
	case DMA_FROM_DEVICE:
		return FCMD_IO_DIR_READ;
	case DMA_TO_DEVICE:
		return FCMD_IO_DIR_WRITE;
	case DMA_NONE:
		return FCMD_IO_DIR_TARGET; /* also used for initiator when no RDMA expected */
	default:
		FCPRINT("unknown DMA dir: %d\n", dmadir);
	}
	return -1;
}

static void mlx4_fc_fcmd_tx_comp(struct mlx4_fc_priv *priv, struct mlx4_cqe *cqe)
{
	struct mlx4_fc_dev 	*mfcdev = priv->mfcdev;
	struct mlx4_fc_q 	*sq = &priv->fcmd.fc_qp.sq;
	struct mlx4_fc_data_seg *data;
	struct mlx4_fc_fcmd_tx_desc 	*tx_desc;
	u64 dma;
	u32 count;
	int wqe_idx = be16_to_cpu(cqe->wqe_index) & sq->size_mask;

	tx_desc = sq->buf + wqe_idx * FCMD_SQ_BB_SIZE;
	
	data = &tx_desc->data;
	dma = be64_to_cpu(data->addr);
	count = be32_to_cpu(data->count);

	pci_unmap_single(mfcdev->dev->pdev, dma, count, PCI_DMA_TODEVICE);
	sq->cons++;
}

static int mlx4_fc_map_sg_to_fmr(struct mlx4_fc_dev *mfcdev, struct scatterlist *sglist, int nents, 
				 struct mlx4_fc_fexch *fexch, enum dma_data_direction dir)
{
	struct mlx4_fmr *fmr = &fexch->fmr;
	struct scatterlist *sg;
	int page_cnt, sg_cnt;
	unsigned int total_len;
	int i;
	u64 fmr_page_mask = SHIFT_TO_MASK(fmr->page_shift); /* masks off low bits */
	u64 dma;
	u64 page_list[MLX4_FC_MAX_FMR_PAGES];
	unsigned int fmr_page_size = SHIFT_TO_SIZE(fmr->page_shift);
	u32 rkey, lkey;
	int rc = 0;
	int err;

	#if DEBUG_SG
	char buf[1024];
	char *p = buf;
	#endif
	
	sg_cnt = pci_map_sg(mfcdev->dev->pdev, sglist, nents, dir);
	ASSERT(sg_cnt > 0);

#if (!USE_FBO)
	/* if we can't use FBO, we exit here if first dma is not alligned */
	if (sg_dma_address(sglist) & ~fmr_page_mask) {
		rc = -EAGAIN;
		goto out_unmap;
	}
#endif

	#if DEBUG_SG
	p += sprintf(p, "FMR_MAP: fmr_pg_sz: 0x%x sg_cnt=%d fbo=0x%llx ", 
		     SHIFT_TO_SIZE(fmr->page_shift), sg_cnt, sglist & ~fmr_page_mask);
	#endif

	page_cnt = 0;
	total_len = 0;
	for_each_sg(sglist, sg, sg_cnt, i) {
		total_len += sg_dma_len(sg);

		#if DEBUG_SG
		p += sprintf(p, "sg[%d]={0x%llx, 0x%x} ", i, sg_dma_address(sg), dma_len);
		#endif

		if (sg_dma_address(sg) & ~fmr_page_mask) {
			if (i > 0) {
				FCPRINT("Non-first page is partial, will retry with smaller granularity FMR\n");
				rc = -EINVAL;
				goto out_unmap;
			}
		}
		if ((sg_dma_address(sg) + sg_dma_len(sg)) & ~fmr_page_mask) {
			if (i < sg_cnt - 1) {
				FCPRINT("Non-last page is partial, will retry with smaller granularity FMR\n");
				rc = -EINVAL;
				goto out_unmap;
			}
		}

		for (dma = (sg_dma_address(sg) & fmr_page_mask);
		     dma < sg_dma_address(sg) + sg_dma_len(sg);
		     dma += fmr_page_size) {
			ASSERT(page_cnt < fmr->max_pages);
			#if DEBUG_SG
			p += sprintf(p, "pg[%d]=0x%llx ", page_cnt, dma);
			#endif

			page_list[page_cnt] = dma;
			++page_cnt;
		}
	}

	#if DEBUG_SG
	p += sprintf(p, "| len=0x%x", total_len);
	ASSERT(p < buf + sizeof(buf));
	FCPRINT("%s\n", buf);
	#endif

	ASSERT(page_cnt);
	err = mlx4_map_phys_fmr_fbo(mfcdev->dev, fmr, page_list,
				    page_cnt,	/* num_pages */
				    0,		/* iova */
				    sg_dma_address(sglist) & ~fmr_page_mask,
				    total_len,
				    &lkey, &rkey, /* fake, don't need them */
				    1		/* use same key */
				    );

	if (err) {
		mlx4_err(mfcdev->dev, "Could not map FMR to given SG\n");
		rc = err;
		goto out_unmap;
	}
	
	return 0;
	
out_unmap:
	pci_unmap_sg(mfcdev->dev->pdev, sglist, nents, dir);
	return rc;
}

static int mlx4_fc_create_bounce(struct mlx4_fc_dev *mfcdev, struct scsi_cmnd *cmd, 
				 struct mlx4_fc_fexch *fexch)
{
	void *bouncebuff;
	int bounceoff = fexch->bounce_off;
	unsigned int total_len;
	struct scatterlist *sg;
	void *page_addr;
	
	int i;

	bouncebuff = kmalloc(BOUNCESIZE, GFP_ATOMIC);
	if (!bouncebuff)
		return -ENOMEM;
#if DEBUG_BOUNCE
	memset(bouncebuff, MLX_MAGIC, BOUNCESIZE);
#endif
	ASSERT(cmd->request_bufflen < BOUNCESIZE - bounceoff);

	total_len = 0;
	scsi_for_each_sg(cmd, sg, scsi_sg_count(cmd), i) {
		// if write, copy data from orig sg to bounce buff
		if (cmd->sc_data_direction == DMA_TO_DEVICE) {
			page_addr = kmap_atomic(sg_page(sg), KM_SOFTIRQ0);
			memcpy(bouncebuff + bounceoff + total_len,
			       page_addr + (sg->offset & ~PAGE_MASK),
			       sg->length);
			kunmap_atomic(page_addr, KM_SOFTIRQ0);
		}
		total_len += sg->length;
	}
	ASSERT_PR(cmd->request_bufflen == total_len,
		  "request_bufflen=%d, total len in sg=%d\n",
		  cmd->request_bufflen, total_len);

	sg_init_one(&fexch->bounce_sg[0], bouncebuff + bounceoff, cmd->request_bufflen);
	fexch->bouncebuff = bouncebuff;

	return 0;
}

static int mlx4_fc_map_fmr(struct mlx4_fc_dev *mfcdev, struct scsi_cmnd *cmd,
			   struct mlx4_fc_fexch *fexch)
{
	int err;


	if (cmd->sc_data_direction == DMA_NONE)
		return 0;

	ASSERT(cmd->sc_data_direction != DMA_BIDIRECTIONAL);
	if (cmd->sc_data_direction == DMA_BIDIRECTIONAL) {
		mlx4_warn(mfcdev->dev, "BIDI SCSI commands not supported\n");
		return -EINVAL;
	}

	ASSERT(scsi_sg_count(cmd));

	fexch->bouncebuff = NULL;
	fexch->bounce_off = 0;
	err = mlx4_fc_map_sg_to_fmr(mfcdev, 
				    scsi_sglist(cmd), scsi_sg_count(cmd), 
				    fexch, cmd->sc_data_direction);
	if (err == -EAGAIN) {
		err = mlx4_fc_create_bounce(mfcdev, cmd, fexch);
		if (err)
			return err;
		err = mlx4_fc_map_sg_to_fmr(mfcdev, &fexch->bounce_sg[0], 1, 
					    fexch, cmd->sc_data_direction);
	}
	
	return err;
}

static void mlx4_fc_unmap_fmr_sg(struct mlx4_fc_dev *mfcdev, struct scatterlist *sglist, int nents, 
				 struct mlx4_fmr *fmr, enum dma_data_direction dir)
{
	u32 dummy_lkey, dummy_rkey;

	pci_unmap_sg(mfcdev->dev->pdev, sglist, nents, dir);
	mlx4_fmr_unmap(mfcdev->dev, fmr, &dummy_lkey, &dummy_rkey);
}

static void mlx4_fc_destroy_bounce(struct mlx4_fc_dev *mfcdev, struct scsi_cmnd *cmd, struct mlx4_fc_fexch *fexch)
{
        struct scatterlist *sg;
	int i;
        unsigned long total_len;
        char *page_addr;

	if (cmd->sc_data_direction == DMA_FROM_DEVICE) {
#if DEBUG_BOUNCE
		// check write before start of buffer (as given to fmr)
		for (p = fexch->bouncebuff; p < fexch->bouncebuff + fexch->bounce_off; ++p)
			ASSERT_PR(*p == MLX_MAGIC,
			      "WRITE BEFORE START OF BUFFER\n"
			      "  buff=%p (dma=0x%llx), bounceoff=0x%x\n"
			      "  at p=%p magic removed, instead got:\n"
			      "  [p-2]=0x%02x [p-1]=0x%02x [p]=0x%02x [p+1]=0x%02x [p+2]=0x%02x [p+3]=0x%02x\n",
			      fexch->bouncebuff, fexch->bounce_dma, fexch->bounce_off,
			      p, *(p-2), *(p-1), *p, *(p+1), *(p+2), *(p+3) );
		
		// check write beyond buffer len (as given to fmr)
		for (p = fexch->bouncebuff + fexch->bounce_off + cmd->request_bufflen; 
		      p < fexch->bouncebuff + fexch->bounce_off; 
		      ++p)
			ASSERT_PR(*p == MLX_MAGIC,
			      "WRITE BEYOND END OF BUFFER\n"
			      "  buff=%p (dma=0x%llx), bounceoff=0x%x, request_buflen=0x%x\n"
			      "  at p=%p magic removed, instead got:\n"
			      "  [p-2]=0x%02x [p-1]=0x%02x [p]=0x%02x [p+1]=0x%02x [p+2]=0x%02x [p+3]=0x%02x\n",
			      fexch->bouncebuff, fexch->bounce_dma, fexch->bounce_off, cmd->request_bufflen,
			      p, *(p-2), *(p-1), *p, *(p+1), *(p+2), *(p+3) );
#endif

		total_len=0;
		scsi_for_each_sg(cmd, sg, scsi_sg_count(cmd), i) {
			page_addr = kmap_atomic(sg_page(sg), KM_SOFTIRQ0);
			memcpy(page_addr + (sg->offset & ~PAGE_MASK),
			       fexch->bouncebuff + fexch->bounce_off + total_len,
                               sg->length);
			kunmap_atomic(page_addr, KM_SOFTIRQ0);
			total_len += sg->length;
		}
		ASSERT(cmd->request_bufflen == total_len);
	}

        kfree(fexch->bouncebuff);
        fexch->bouncebuff = NULL;
}

static void mlx4_fc_unmap_fmr(struct mlx4_fc_dev *mfcdev, struct scsi_cmnd *cmd, struct mlx4_fc_fexch *fexch)
{
	if (cmd->sc_data_direction == DMA_NONE)
		return;

	if (cmd->sc_data_direction == DMA_BIDIRECTIONAL) {
		mlx4_warn(mfcdev->dev, "BIDI SCSI commands not supported\n");
		return;
	}

	if (fexch->bouncebuff) {
		mlx4_fc_unmap_fmr_sg(mfcdev, &fexch->bounce_sg[0], 1, 
				     &fexch->fmr, cmd->sc_data_direction);
		mlx4_fc_destroy_bounce(mfcdev, cmd, fexch);
	} else
		mlx4_fc_unmap_fmr_sg(mfcdev, scsi_sglist(cmd), scsi_sg_count(cmd),
				     &fexch->fmr, cmd->sc_data_direction);
}

void mlx4_fc_clean_exch(void *drv_priv, struct scsi_cmnd *cmd, int xno)
{
	struct mlx4_fc_priv *priv = drv_priv;
	struct mlx4_fc_fexch *fexch = &priv->fexch[xno];

	mlx4_fc_unmap_fmr(priv->mfcdev, cmd, fexch);
	mlx4_fc_bitmap_slot_free(priv->fexch_bm, xno);
}

/*
 * FEXCH completion - pay attention: ethernet header is stripped.
 */
static void mlx4_fc_fexch_rx_comp(struct mlx4_fc_priv *priv, struct mlx4_cqe *generic_cqe)
{
	struct mlx4_fc_fexch_cqe *cqe = (struct mlx4_fc_fexch_cqe *)generic_cqe;
	struct mlx4_fc_fexch *fexch;
	struct mlx4_fc_q *rq;
	struct sk_buff *skb;
	int wqe_idx;
	struct mlx4_fc_rcv_info *fr;
	struct mlx4_fc_rx_desc *rx_desc;
	int xno;
	u32 qpn;

	qpn = be32_to_cpu(cqe->my_qpn) & ((1 << 24) - 1);
	xno = qpn - priv->base_fexch_qpn;
	fexch = &priv->fexch[xno];
	rq = &fexch->fc_qp.rq;
	wqe_idx = be16_to_cpu(cqe->wqe_index) & rq->size_mask;
	rx_desc = rq->buf + (wqe_idx * rq->stride);

	pci_unmap_single(priv->mfcdev->dev->pdev,
			 be64_to_cpu(rx_desc->data[0].addr),
			 be32_to_cpu(rx_desc->data[0].count),
			 PCI_DMA_FROMDEVICE);

	#ifdef DEBUG_CPU_BINDING
	if (xno % num_online_cpus() != smp_processor_id())
		FCPRINT("SMP: got completion for xno %d on cpu %d instead of %d\n", 
			xno, smp_processor_id(), xno % num_online_cpus());
	#endif
	
	ASSERT(rq->info);
	
	skb = (struct sk_buff *)rq->info[wqe_idx].buf;
	fr = mlx4_fc_rcv_info_from_skb(skb);

	fr->txcnt = be32_to_cpu(cqe->xmit_byte_count);
	fr->rxcnt = be32_to_cpu(cqe->rcv_byte_count);
	fr->bcnt = be32_to_cpu(cqe->byte_cnt);
	fr->xno = xno;
	fr->fmr = &fexch->fmr;

	skb_put(skb, fr->bcnt);
	skb_pull(skb, 2); /* skip FCOE ethertype */
	
	mlx4_fc_rcv(priv->mfcdev->fdev[priv->port], skb);
	
	fexch->fc_qp.rq.cons++;
}

int mlx4_fc_create_fexch(struct mlx4_fc_priv *priv, int xno)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_fexch *fexch = &priv->fexch[xno];
	struct mlx4_fc_qp *qp = &fexch->fc_qp;
	struct mlx4_fc_q *rq = &qp->rq;
	int err=0;

	rq->prod   = 0;
	rq->cons   = 0xffffffff;
	rq->stride = FEXCH_RQ_WQE_SIZE;
	rq->size = FEXCH_RQ_NUM_WQES;
	rq->size_mask = rq->size - 1;
	rq->info = vmalloc(rq->size * sizeof(struct mlx4_fc_info));
	if (!rq->info) {
		mlx4_err(mfcdev->dev, "Could not alloc info for fexch %d rq\n", xno);
		err = -ENOMEM;
		goto err_out;
	}
	memset(rq->info, 0, rq->size * sizeof(struct mlx4_fc_info));

	qp->buf_size = (FEXCH_SQ_BB_SIZE * FEXCH_SQ_NUM_BBS) + (rq->size * rq->stride);

        err = mlx4_alloc_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size, qp->buf_size);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate man for FEXCH %d\n", xno);
		goto err_free_rxinfo;
	}

	rq->buf = qp->wqres.buf.u.direct.buf + (FEXCH_SQ_BB_SIZE * FEXCH_SQ_NUM_BBS);

	*qp->wqres.db.db = 0;

	err = mlx4_qp_alloc(mfcdev->dev, priv->base_fexch_qpn + xno, &qp->mqp);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate QP number 0x%x\n", qp->mqp.qpn);
		goto err_free_man;
	}
	
	qp->doorbell_qpn = swab32(qp->mqp.qpn << 8);

	ASSERT(MLX4_FC_FMR_PAGE_SHIFT >= (ffs(mfcdev->dev->caps.page_size_cap) - 1));
	err = mlx4_fmr_alloc_reserved(mfcdev->dev, priv->base_fexch_mpt + xno, mfcdev->priv_pdn | MLX4_MPT_ENABLE_INVALIDATE,
				      MLX4_PERM_REMOTE_WRITE | MLX4_PERM_REMOTE_READ,
				      MLX4_FC_MAX_FMR_PAGES, /* max_pages */
				      1, /* max_maps */
				      MLX4_FC_FMR_PAGE_SHIFT, &fexch->fmr);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate FMR for fexch %d, err=%d\n", xno, err);
		goto err_free_qp;
	}

	err = mlx4_fmr_enable(mfcdev->dev, &fexch->fmr);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not enable FMR for fexch %d, err=%d\n", xno, err);
		goto err_free_fmr;
	}
	
	return 0;

err_free_fmr:
	mlx4_fmr_free_reserved(mfcdev->dev, &fexch->fmr);
err_free_qp:
	mlx4_qp_remove(mfcdev->dev, &fexch->fc_qp.mqp);
	mlx4_qp_free(mfcdev->dev, &fexch->fc_qp.mqp);
err_free_man:
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
err_free_rxinfo:
	vfree(rq->info);
	rq->info = NULL;
err_out:
	return err;
}

void mlx4_fc_destroy_fexch(struct mlx4_fc_priv *priv, int xno)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_fexch *fexch = &priv->fexch[xno];
	struct mlx4_fc_qp *qp = &fexch->fc_qp;
	struct mlx4_fc_q *rq = &qp->rq;

	mlx4_fmr_free_reserved(mfcdev->dev, &fexch->fmr);
	mlx4_qp_to_reset(mfcdev->dev, &qp->mqp);
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
	vfree(rq->info);
}

int mlx4_fc_init_fexch(struct mlx4_fc_priv *priv, int xno)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_fexch *fexch = &priv->fexch[xno];
	struct mlx4_fc_qp *qp = &fexch->fc_qp;
	enum mlx4_qp_state qp_state;
	int err = 0;
	
	struct mlx4_qp_context context = {
		.flags 		= cpu_to_be32(QPC_SERVICE_TYPE_FEXCH << 16),
		.pd 		= cpu_to_be32(mfcdev->priv_pdn),
		.mtu_msgmax	= 0x77,	/* Raw-ETH requirement */
		.sq_size_stride = 0, 	/* this means SQ_NUM_BBS=1, and SQ_BB_SIZE=1 */
		.rq_size_stride	= ilog2(FEXCH_RQ_NUM_WQES) << 3 | ilog2(FEXCH_RQ_WQE_SIZE >> 4),
		.usr_page 	= cpu_to_be32(mfcdev->priv_uar.index),
		.local_qpn 	= cpu_to_be32(qp->mqp.qpn),
		.pri_path.sched_queue = 0x83 | (priv->port - 1) << 6,
		.pri_path.counter_index = 0xff,
		.pri_path.ackto = MLX4_LINK_TYPE_ETH,
		.pri_path.grh_mylmc = priv->rfci[RFCI_DATA].fc_mac_idx, /* Source MAC index */
		.params2	= cpu_to_be32((qp->wqres.buf.u.direct.map & (PAGE_SIZE-1)) & 0xfc0), /* page_offset */
		.cqn_send	= cpu_to_be32(priv->fexch_cq[xno % num_online_cpus()].mcq.cqn),
		.cqn_recv	= cpu_to_be32(priv->fexch_cq[xno % num_online_cpus()].mcq.cqn),
		.db_rec_addr 	= cpu_to_be64(qp->wqres.db.dma),
		.srqn		= 0,
		.my_fc_id_idx	= 0,
	};

	if (priv->fc_vlan_idx != -1) {
		context.pri_path.fl 	= 0x40;
		context.pri_path.snooper_flags = priv->fc_vlan_idx;
	}

	err = mlx4_qp_to_ready(mfcdev->dev, &qp->wqres.mtt, &context, &qp->mqp, &qp_state);
	if (qp_state != MLX4_QP_STATE_RTS) {
		mlx4_err(mfcdev->dev, "Error bringing FEXCH %d QP to RTS state\n", xno);
		return err;
	}
	return 0;
}

/*
 * Attention: This function could be called from interrupt context
 */
int mlx4_fc_create_fcmd(struct mlx4_fc_priv *priv)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_qp *qp = &priv->fcmd.fc_qp;
        struct mlx4_fc_q *sq = &qp->sq;
	int err=0;
	int i, eqidx, cpu;

	spin_lock_init(&priv->fcmd.lock);
	
	err = mlx4_fc_q_init(sq, FCMD_SQ_BB_SIZE, FCMD_SQ_NUM_BBS);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not alloc info for fcmd sq\n");
		goto err_out;
	}
	
	priv->fcmd.fc_qp.buf_size = (sq->size *  sq->stride) + (FCMD_RQ_NUM_WQES * FCMD_RQ_WQE_SIZE);
	
	err = mlx4_alloc_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size, qp->buf_size);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate man for fcmd of port %d, err=%d\n", priv->port, err);
		goto err_free_txinfo;
	}

	sq->buf = qp->wqres.buf.u.direct.buf;
	*qp->wqres.db.db = 0;

	mlx4_fc_stamp_sq(sq);
	
	err = mlx4_qp_reserve_range(mfcdev->dev, 1, 1, &qp->mqp.qpn);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not reserve QPN for fcmd of port %d, err=%d\n", priv->port, err);
		goto err_free_man;
	}
	FCPRINT("Creating FCMD for port %d using QPN=0x%x\n", priv->port, qp->mqp.qpn);

	err = mlx4_qp_alloc(mfcdev->dev, qp->mqp.qpn, &qp->mqp);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate QP number 0x%x\n", qp->mqp.qpn);
		goto err_release_qp;
	}

	qp->doorbell_qpn = swab32(qp->mqp.qpn << 8);

	err = mlx4_fc_create_cq(priv, &priv->fcmd.fc_cq, MLX4_FC_FCMD_CQ_ENTRIES,
				MLX4_LEAST_ATTACHED_VECTOR, /* don't mind which cpu */
				0, /* don't arm the cq */
				NULL, mlx4_fc_fcmd_tx_comp, "FCMD");
	if (err) {
		mlx4_err(mfcdev->dev, "Failed creating FCMD CQ for port %d, err=%d\n", priv->port, err);
		goto err_free_qp;
	}

	FCPRINT("Created FCMD CQ successfull for port %d, CQN=%d\n", priv->port, priv->fcmd.fc_cq.mcq.cqn);

	/* Create FEXCHs for this FCMD */
	priv->num_fexch = MIN(MLX4_FC_MAX_FEXCH_PER_PORT, mfcdev->num_fexch_qps_perport);
	priv->fexch_bm = mlx4_fc_bitmap_alloc(priv->num_fexch);
	if (!priv->fexch_bm) {
		mlx4_err(mfcdev->dev, "Error allocating bitmap for fexch on port %d\n", priv->port);
		goto err_free_cq;
	}

	eqidx = 0;
	for_each_online_cpu(cpu) {
		err = mlx4_fc_create_cq(priv, &priv->fexch_cq[eqidx], 
					priv->num_fexch / num_online_cpus(),
					(eqidx % num_online_cpus()) % mfcdev->dev->caps.num_comp_vectors,
					1, /* arm cq */
					mlx4_fc_fexch_rx_comp, NULL,
					"FEXCH");
		if (err) {
			mlx4_err(mfcdev->dev, "Failed creating FEXCH CQ %d for port %d, err=%d\n", eqidx, priv->port, err);
			goto err_destroy_fexch_cq;
		}
		FCPRINT("Created FEXCH CQ %d successfull for port %d, CQN=%d\n", eqidx, priv->port, priv->fexch_cq[eqidx].mcq.cqn);

		++eqidx;
	}

	for (i = 0; i < priv->num_fexch; i++) {
		err = mlx4_fc_create_fexch(priv, i);
		if (err) {
			mlx4_err(mfcdev->dev, "Failed creating FEXCH %d for port %d, err=%d\n", i, priv->port, err);
			goto err_destroy_fexch;
		}
	}
	FCPRINT("Created %d FEXCHs for port %d\n", i, priv->port);

	return 0;


err_destroy_fexch:
	while (--i >= 0)
		mlx4_fc_destroy_fexch(priv, i);
err_destroy_fexch_cq:
	while (--eqidx >= 0)
		mlx4_fc_destroy_cq(&priv->fexch_cq[eqidx]);
	mlx4_fc_bitmap_free(priv->fexch_bm);
err_free_cq:
	mlx4_fc_destroy_cq(&priv->fcmd.fc_cq);
err_free_qp:
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
err_release_qp:
	mlx4_qp_release_range(mfcdev->dev, qp->mqp.qpn, 1);
err_free_man:
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
err_free_txinfo:
	vfree(sq->info);
	sq->info = NULL;
err_out:
	return err;
}

void mlx4_fc_destroy_fcmd(struct mlx4_fc_priv *priv)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_qp *qp = &priv->fcmd.fc_qp;
	struct mlx4_fc_q *sq = &qp->sq;
	int i;

	for (i = 0; i < priv->num_fexch; ++i) {
		mlx4_fc_destroy_fexch(priv, i);
	}
	for (i = 0; i < num_online_cpus(); ++i) {
		mlx4_fc_destroy_cq(&priv->fexch_cq[i]);
	}
	mlx4_fc_bitmap_free(priv->fexch_bm);
	
	mlx4_qp_release_range(mfcdev->dev, priv->fcmd.fc_cq.mcq.cqn, 1);
	mlx4_fc_destroy_cq(&priv->fcmd.fc_cq);
	mlx4_qp_to_reset(mfcdev->dev, &qp->mqp);
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
	vfree(sq->info);
}

int mlx4_fc_init_fcmd(struct mlx4_fc_priv *priv)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_qp *qp = &priv->fcmd.fc_qp;
	enum mlx4_qp_state qp_state;
	int err = 0;
	int i;
	
	struct mlx4_qp_context context = {
		.flags 		= cpu_to_be32(QPC_SERVICE_TYPE_FCMD << 16),
		.pd 		= cpu_to_be32(mfcdev->priv_pdn),
		.mtu_msgmax	= 0x77,	/* Raw-ETH requirement */
		.sq_size_stride = ilog2(FCMD_SQ_NUM_BBS) << 3 | ilog2(FCMD_SQ_BB_SIZE >> 4) | SQ_NO_PREFETCH,
		.rq_size_stride	= 0,	/* this means RQ_NUM_WQES=1, and RQ_WQE_SIZE=1 */
		.usr_page 	= cpu_to_be32(mfcdev->priv_uar.index),
		.local_qpn 	= cpu_to_be32(qp->mqp.qpn),
		.pri_path.sched_queue = 0x83 | (priv->port - 1) << 6,
		.pri_path.counter_index = 0xff,
		.pri_path.ackto = MLX4_LINK_TYPE_ETH,
		.pri_path.grh_mylmc = priv->rfci[RFCI_DATA].fc_mac_idx, /* Source MAC index */
		.params2	= cpu_to_be32((qp->wqres.buf.u.direct.map & (PAGE_SIZE-1)) & 0xfc0),
		.cqn_send	= cpu_to_be32(priv->fcmd.fc_cq.mcq.cqn),
		.cqn_recv	= cpu_to_be32(priv->fcmd.fc_cq.mcq.cqn),
		.db_rec_addr 	= cpu_to_be64(qp->wqres.db.dma),
		.srqn		= 0,
		.VE		= 0,
		.exch_base	= cpu_to_be16(priv->base_fexch_qpn),
		.exch_size	= ilog2(mfcdev->num_fexch_qps_perport)-1,
		.my_fc_id_idx	= 0,
	};
	if (priv->fc_vlan_idx != -1) {
		context.pri_path.fl 	= 0x40;
		context.pri_path.snooper_flags = priv->fc_vlan_idx;
	}

	FCPRINT("Setting FCMD to use 0x%x exchs (log2=0x%x), starting at qpn 0x%x\n",
		mfcdev->num_fexch_qps_perport, ilog2(mfcdev->num_fexch_qps_perport), priv->base_fexch_qpn);
	
	err = mlx4_qp_to_ready(mfcdev->dev, &qp->wqres.mtt, &context, &qp->mqp, &qp_state);
	if (qp_state != MLX4_QP_STATE_RTS) {
		mlx4_err(mfcdev->dev, "Error bringing FCMD QP to RTS state\n");
		return err;
	}

	/* bring FEXCHs to ready state */
	for (i = 0; i < priv->num_fexch; i++) {
		err = mlx4_fc_init_fexch(priv, i);
		if (err) {
			mlx4_err(mfcdev->dev, "Failed init of FEXCH %d for port %d, err=%d\n", i, priv->port, err);
			return err;
		}
	}
	FCPRINT("Initialized %d FEXCHs for port %d\n", i, priv->port);

	return 0;
}

static inline void set_ctrl_seq(struct mlx4_fc_ctrl_seg *ctrl, u8 seqid, u8 info, u8 ls, u32 task_retry_id)
{
	ctrl->size = cpu_to_be16( ((FCMD_SQ_BB_SIZE / 16) & 0x3f) | (1 << 7));
	ctrl->flags = cpu_to_be32(
		MLX4_FC_BIT_TX_COMP | 	/* request completion */
		(seqid << 24) |
		(info << 20) |
		(ls << 16) );
	ctrl->parameter = cpu_to_be32(task_retry_id);
}

static inline int prepare_fexch(struct mlx4_fc_priv *priv, struct fc_scsi_pkt *fsp)
{
	struct mlx4_fc_fexch 	*fexch;
	struct mlx4_fc_dev 	*mfcdev = priv->mfcdev;
	int fexch_idx;
	int err;

	fexch_idx = mlx4_fc_bitmap_slot_alloc(priv->fexch_bm);
	if (fexch_idx == -1) {
		mlx4_err(mfcdev->dev, "No free FEXCH\n");
		return -1;
	}
	fexch = &priv->fexch[fexch_idx];

	err = mlx4_fc_map_fmr(mfcdev, fsp->cmd, fexch);
	ASSERT_PR(err == 0, "xno=%d\n", fexch_idx);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not map SCSI sg to FMR\n");
		return -1;
	}

	err = mlx4_fc_post_rx_buf(mfcdev, &fexch->fc_qp, MLX4_FC_FCMD_RX_SKB_BUFSIZE, fsp);
	ASSERT_PR(!err, "could not post a single recv buff on fexch %d\n", fexch_idx);

	mlx4_fc_ring_db_rx(&fexch->fc_qp);

	return fexch_idx;
}

static inline void set_init_seg(struct mlx4_fc_init_seg *init, int frame_size, fc_fid_t remote_fid, 
				enum dma_data_direction data_dir, int fexch_idx)
{
	enum mlx4_fcmd_io_dir mlx_dir = scsi_dir_translate(data_dir);

	ASSERT(mlx_dir >= 0);
	ASSERT(mlx_dir != FCMD_IO_DIR_BIDI);
	
	init->pe = 0;		/* priority enable, goes to F_CTL[17] */
	init->cs_ctl = 0;	/* CS_CTL/Priority field */
	init->seq_id_tx = 0;	/* seq. id to be used in FCP_DATA frames */
	init->mtu = cpu_to_be16(frame_size / 4);
	init->remote_fid[2] = (remote_fid) & 0xff;
	init->remote_fid[1] = (remote_fid >> 8) & 0xff;
	init->remote_fid[0] = (remote_fid >> 16) & 0xff;

	init->flags = 
		(1 << 1) | 			/* originator, 1=initiator */
		(scsi_dir_translate(data_dir) << 3) | /* scsi cmd direction */
		(0x0 << 6);			/* abort code, goes to F_CTL[5:4] */
	init->remote_exch = cpu_to_be16(0xffff); /* initiators never know remote exch no. at beginning of exch */
	init->local_exch_idx = cpu_to_be16(fexch_idx);	/* alloc free exchange, put index here */
}


DEFINE_SPINLOCK(cq_poll);

int mlx4_do_fcmd_send(void *drv_priv, struct fc_scsi_pkt *fsp)
{
	struct mlx4_fc_priv 	*priv = drv_priv;
	struct mlx4_fc_dev 	*mfcdev = priv->mfcdev;
	struct mlx4_dev 	*mdev = mfcdev->dev;
	struct mlx4_fc_q 	*sq = &priv->fcmd.fc_qp.sq;
	int			fexch_idx;

	struct mlx4_fc_fcmd_tx_desc 	*tx_desc;
	struct mlx4_fc_data_seg *data;

	dma_addr_t dma;
	void *cmd_data;
	int cmd_len;
	u32 index, prod;
	__be32 op_own;
	unsigned long flags;

	if (spin_trylock_irqsave(&cq_poll, flags)) {
		mlx4_fc_cq_clean(&priv->fcmd.fc_cq);
		spin_unlock_irqrestore(&cq_poll, flags);
	}
	
	/* Check available SQ BBs + 1 spare SQ BB for owenership */
	spin_lock_irqsave(&priv->fcmd.lock, flags);
	if (unlikely((u32) (sq->prod - sq->cons - 1) > sq->size - 2)) {
		mlx4_info(mdev, "fcmd_send: Out of send queue BBs\n");
		spin_unlock_irqrestore(&priv->fcmd.lock, flags);
		return -1;
	}
	prod = sq->prod;
	++sq->prod;
	spin_unlock_irqrestore(&priv->fcmd.lock, flags);

	index = prod & sq->size_mask;
	sq->info[index].buf = NULL;
	tx_desc = sq->buf + index * FCMD_SQ_BB_SIZE;

	/* set ctrl segment: for initiator we must have info=6
	   and ls=0. seqid and task_retry_id set to 0 */
	set_ctrl_seq(&tx_desc->ctrl, 0, 6, 0, 0);

	/* allocate and prepare FEXCH for command */
	fexch_idx = prepare_fexch(priv, fsp);
	if (fexch_idx < 0)
		return -1;

	/* set init segment */
	set_init_seg(&tx_desc->init, fsp->rp->rp_sess->fs_max_payload, fsp->rp->rp_fid, 
		     fsp->cmd->sc_data_direction, fexch_idx);

	/* set addr segment */
	tx_desc->addr.static_rate = 0;
	if (priv->address_mode == FCOE_FCOUI_ADDR_MODE)
		fc_set_mac(tx_desc->addr.dmac, (net24_t *)tx_desc->init.remote_fid);
	else
		memcpy(tx_desc->addr.dmac, priv->dest_addr, ETH_ALEN);
	
	/* prepare cdb command in buffer */
	/* TODO: copy this directly to WQE using immediate data segment */
	net32_put(&fsp->cdb_cmd.fc_dl, fsp->data_len);
	fsp->cdb_cmd.fc_flags = fsp->req_flags & ~FCP_CFL_LEN_MASK;
	net64_put((net64_t *) fsp->cdb_cmd.fc_lun, (fsp->lun << 48));
	memcpy(fsp->cdb_cmd.fc_cdb, fsp->cmd->cmnd, fsp->cmd->cmd_len);

	/* set data segment */
	cmd_data = &fsp->cdb_cmd;
	cmd_len = sizeof(fsp->cdb_cmd);
	dma = pci_map_single(mfcdev->dev->pdev, cmd_data, cmd_len, PCI_DMA_TODEVICE);
	data = &tx_desc->data;
	data->addr = cpu_to_be64(dma);
	data->count = cpu_to_be32(cmd_len);
	data->mem_type = cpu_to_be32(mfcdev->mr.key); /* always snoop */

	op_own = cpu_to_be32(MLX4_FC_FCMD_OP_SEND) | 
		((prod & sq->size) ? cpu_to_be32(MLX4_FC_BIT_DESC_OWN) : 0);
	
	/* Ensure new descirptor (and ownership of next descirptor) hits memory
	 * before setting ownership of this descriptor to HW */
	wmb();
	tx_desc->ctrl.op_own = op_own;

	//FCPRINT("Posting command, FEXCH=%d, WQ_idx=%d, scsi reqlen=%d, scsi_dir=%d\n", 
	//	fexch_idx, index, fsp->cmd->request_bufflen, fsp->cmd->sc_data_direction);

	/* Ring doorbell! */
	wmb();
	writel(priv->fcmd.fc_qp.doorbell_qpn, mfcdev->uar_map + MLX4_SEND_DOORBELL);

	return 0;
}

