/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>

#include <linux/mlx4/driver.h>
#include <linux/mlx4/cmd.h>
#include <linux/mlx4/qp.h>
#include <linux/mlx4/device.h>

#include "mlx4_fc.h"

MODULE_AUTHOR(MLX4FC_DRIVER_AUTHOR);
MODULE_DESCRIPTION(MLX4FC_DRIVER_DESC);
MODULE_LICENSE(MLX4FC_DRIVER_LICENSE);
MODULE_VERSION(MLX4FC_DRIVER_VERSION);

int mlx4_fc_vlan_id = MLX4_DEFAULT_VLAN_ID;
module_param_named(vlan_id, mlx4_fc_vlan_id, int, 0644);
MODULE_PARM_DESC(vlan_id, "VLAN ID. default is " __stringify(MLX4_DEFAULT_VLAN_ID));

static const char mlx4_fc_version[] __devinitdata =
	MLX4FC_DRIVER_NAME ": " MLX4FC_DRIVER_DESC " v"
	MLX4FC_DRIVER_VERSION " (" MLX4FC_DRIVER_RELDATE ")\n";

int mlx4_fc_q_init(struct mlx4_fc_q *q, u16 stride, u32 size)
{
	q->prod   = 0;
	q->cons   = 0xffffffff;
	q->stride = stride;
	q->size = size;
	q->size_mask = q->size - 1;
	q->info = vmalloc(q->size * sizeof(struct mlx4_fc_info));
	if (!q->info) {
		return -ENOMEM;
	}

	memset(q->info, 0, q->size * sizeof(struct mlx4_fc_info));

	return 0;
}

void mlx4_fc_stamp_sq(struct mlx4_fc_q *sq)
{
	__be32 *p;
	int i;
	
	/* stamp first dword of every 64 byte */
	for (i = 0; i < sq->size; ++i) {
		p = sq->buf + i * sq->stride;
		*p = cpu_to_be32(1 << 31); // that's the op_own field of the WQE
	}

}

static void *mlx4_fc_add(struct mlx4_dev *mdev)
{
	struct mlx4_fc_dev *mfcdev;
	int port;
	int err;

	FCPRINT("mlx4 device added: %.*s\n", MLX4_BOARD_ID_LEN, mdev->board_id);

	mfcdev = (struct mlx4_fc_dev *)kzalloc(sizeof(struct mlx4_fc_dev), GFP_KERNEL);
	if (!mfcdev) {
		dev_err(&mdev->pdev->dev, "Device struct alloc failed, "
			"aborting.\n");
		err = -ENOMEM;
		goto err;
	}

	if (mlx4_pd_alloc(mdev, &mfcdev->priv_pdn))
		goto err_free_dev;

	FCPRINT("PDN: 0x%x\n", mfcdev->priv_pdn);

	err = mlx4_mr_alloc(mdev, mfcdev->priv_pdn, 0, ~0ull,
				MLX4_PERM_LOCAL_WRITE | MLX4_PERM_LOCAL_READ, 0, 0, &mfcdev->mr);

	if (err) {
		dev_err(&mdev->pdev->dev, "mr alloc failed");
		err = -ENOMEM;
		goto err_free_pd;
	}
	
	err = mlx4_mr_enable(mdev, &mfcdev->mr);
	if (err) {
		dev_err(&mdev->pdev->dev, "mr enable failed");
		err = -ENOMEM;
		goto err_free_mr;
	}

	FCPRINT("Memory key is 0x%x\n", mfcdev->mr.key);

	if (mlx4_uar_alloc(mdev, &mfcdev->priv_uar))
		goto err_free_mr;

	mfcdev->uar_map = ioremap(mfcdev->priv_uar.pfn << PAGE_SHIFT, PAGE_SIZE);
	if (!mfcdev->uar_map)
		goto err_free_uar;
	MLX4_INIT_DOORBELL_LOCK(&mfcdev->uar_lock);

	INIT_LIST_HEAD(&mfcdev->pgdir_list);
	mutex_init(&mfcdev->pgdir_mutex);
	
	mfcdev->dev = mdev;
	mfcdev->dma_device = &mdev->pdev->dev;	
	mfcdev->fc_ports = mlx4_get_ports_of_type(mdev, MLX4_PORT_TYPE_ETH);
	mfcdev->num_fc_ports = 0;
	mlx4_foreach_port(port, mfcdev->fc_ports)
		mfcdev->num_fc_ports++;
	FCPRINT("FC ports bitmap: 0x%x, %d ports\n", mfcdev->fc_ports, mfcdev->num_fc_ports);

	mlx4_query_steer_cap(mdev, &mfcdev->log_num_mac, &mfcdev->log_num_vlan, &mfcdev->log_num_prio);
	mlx4_qp_get_region(mdev, MLX4_QP_REGION_FC_ADDR, &mfcdev->base_rfci_qpn, &mfcdev->num_rfci_qps);
	mlx4_qp_get_region(mdev, MLX4_QP_REGION_FC_EXCH, &mfcdev->base_fexch_qpn, &mfcdev->num_fexch_qps);
	mfcdev->base_fexch_mpt = mlx4_get_fexch_mpts_base(mfcdev->dev);
	mfcdev->num_fexch_qps_perport = rounddown_pow_of_two(mfcdev->num_fexch_qps / mfcdev->num_fc_ports);
	
	FCPRINT("macs=%d, vlans=%d, prios=%d, rfci_qpn=0x%x, rfci_cnt=0x%x, fexch_qpn=0x%x, fexch_cnt=0x%x, fexch_basempt=0x%x\n",
		mfcdev->log_num_mac, mfcdev->log_num_vlan, mfcdev->log_num_prio,
		mfcdev->base_rfci_qpn, mfcdev->num_rfci_qps,
		mfcdev->base_fexch_qpn, mfcdev->num_fexch_qps,
		mfcdev->base_fexch_mpt);

	mlx4_foreach_port(port, mfcdev->fc_ports)
		mlx4_fc_init_port(mfcdev, port);

	return mfcdev;
				
err_free_uar:
	mlx4_uar_free(mdev, &mfcdev->priv_uar);
err_free_mr:
	mlx4_mr_free(mfcdev->dev, &mfcdev->mr);
err_free_pd:
	mlx4_pd_free(mdev, mfcdev->priv_pdn);
err_free_dev:
	kfree(mfcdev);
err:
	return NULL;
}

static void mlx4_fc_remove(struct mlx4_dev *dev, void *fcdev_ptr)
{
	struct mlx4_fc_dev *mfcdev = fcdev_ptr;
	int port;

	mlx4_foreach_port(port, mfcdev->fc_ports) {
		mlx4_fc_free_port(mfcdev, port);
	}

	mlx4_uar_free(dev, &mfcdev->priv_uar);
	mlx4_mr_free(dev, &mfcdev->mr);
	mlx4_pd_free(dev, mfcdev->priv_pdn);

	kfree(mfcdev);
	
	FCPRINT("mlx4 device removed: %.*s\n", MLX4_BOARD_ID_LEN, dev->board_id);
}

static void mlx4_fc_event(struct mlx4_dev *dev, void *mfcdev_ptr,
			  enum mlx4_dev_event event, int subtype,
			  int port)
{
	struct mlx4_fc_dev *mfcdev = (struct mlx4_fc_dev *)mfcdev_ptr;
	struct fcdev *fdev;
	uint8_t new_status;

	switch (event) {
	case MLX4_EVENT_TYPE_PORT_CHANGE:
		switch (subtype) {
		case MLX4_PORT_CHANGE_SUBTYPE_ACTIVE:
			FCPRINT("mlx4_port_change, active\n");
			new_status = TRANS_LINK_UP;
			break;
		case MLX4_PORT_CHANGE_SUBTYPE_DOWN:
			FCPRINT("mlx4_port_change, deactive\n");
			new_status = TRANS_LINK_DOWN;
			break;
		default:
			FCPRINT("mlx4_port_change, error\n");
			return;
		}
		break;
		
	case MLX4_EVENT_TYPE_LOCAL_CATAS_ERROR:
		FCPRINT("fatal error\n");
		return;

	default:
		FCPRINT("unknown event: %d\n",event);
		return;
	}


	fdev = mfcdev->fdev[port];
	if (!fdev) return;

	if (fdev->fd_link_status != new_status) {
		fdev->fd_link_status = new_status;
		if (new_status == TRANS_LINK_UP) {
			FCPRINT("Link up\n");
			openfc_linkup(fdev);
		} else {
			FCPRINT("Link down\n");
			openfc_linkdown(fdev);
		}
	}
}

static struct mlx4_interface mlx4_fc_interface = {
	.add	= mlx4_fc_add,
	.remove	= mlx4_fc_remove,
	.event	= mlx4_fc_event
};

static int __init mlx4_fc_init(void)
{
	int err;

	FCPRINT("module loaded\n");

	/* mlx4_enable_wc(); */
	init_nport_id_table();
	err = mlx4_register_interface(&mlx4_fc_interface);
	if (err) 
		return err;
	
	err = mlx4_fc_init_rx();
	if (err)
		goto err_unreg; 

	return 0;

err_unreg:
	mlx4_unregister_interface(&mlx4_fc_interface);

	return err;
}

static void __exit mlx4_fc_cleanup(void)
{
	mlx4_fc_kill_rx();
	mlx4_unregister_interface(&mlx4_fc_interface);
	/* mlx4_disable_wc(); */
}

module_init(mlx4_fc_init);
module_exit(mlx4_fc_cleanup);
