/*
 * Copyright(c) 2007 Intel Corporation. All rights reserved.
 * Copyright(c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under the terms and conditions of the GNU
 * General Public License (GPL) Version 2, as published by the Free Software 
 * Foundation.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/kthread.h>

#include <linux/mlx4/driver.h>
#include <linux/mlx4/cmd.h>
#include <linux/mlx4/qp.h>
#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_transport_fc.h>

#include "mlx4_fc.h"
#include "fc_frame.h"
#include "fc_print.h"
#include "fc_els.h"
#include "fc_remote_port.h"
#include "fc_fs.h"
#include "openfc.h"
#include "../openfc/openfc_scsi_pkt.h"

#include "mlx4_ofc.h"

#define CMD_SP(Cmnd)		((Cmnd)->SCp.ptr)

struct mlx4_fc_rx_thread *rx_thread_percpu[MLX4_NR_CPUS];

MODULE_AUTHOR("Oren Duer <oren@mellanox.co.il>");
MODULE_DESCRIPTION("Mellanox-OpenFCoE connector");
MODULE_LICENSE("GPL");

int mlx4_fc_auto_start = 1;
module_param_named(auto_start, mlx4_fc_auto_start, int, 0644);
MODULE_PARM_DESC(auto_start, "Automatic FC login. Default = 1 (yes)");

int mlx4_fc_send(struct fcdev *fc_dev, struct fc_frame *fp)
{
	struct fc_accel_dev 	*fcadev = fc_dev->drv_priv;
	struct mlx4_fc_priv	*priv = fcadev->drv_priv;
	struct fc_frame_header 	*fh;
	struct sk_buff 		*skb;
	uint8_t 		sof, eof;
	u_int 			hlen;		/* header length implies the version */
	u_int 			tlen;		/* trailer length */
	int 			wlen;
	struct ethhdr 		*eh;
	struct fcoe_crc_eof 	*cp;
	int 			flogi_in_progress = 0;
	struct fcoe_hdr_old 	*ohp;

	ASSERT((fp->fr_len % sizeof(uint32_t)) == 0);

	/*
	 * if it is a flogi then we need to learn gw-addr
	 * and my own fcid
	 */
	fh = fc_frame_header_get(fp);

	if (unlikely(fh->fh_r_ctl == FC_RCTL_ELS_REQ)) {
		if (fc_frame_payload_op(fp) == ELS_FLOGI) {
			priv->flogi_oxid = net16_get(&fh->fh_ox_id);
			priv->address_mode = FCOE_FCOUI_ADDR_MODE;
			priv->flogi_progress = 1;
			flogi_in_progress = 1;
			priv->rfci_rx_enabled = 1;
		} else if (priv->flogi_progress && net24_get(&fh->fh_s_id) != 0) {
			fc_set_mac(priv->rfci[RFCI_DATA].mac, &fh->fh_s_id);
			priv->my_fid = fh->fh_s_id;
			priv->flogi_progress = 0;
			priv->need_update_data_mac = 1; /* Don'y know why - but this context has in_atomic()=1, so we can't add RFCI here
							 * Will do it later in fcoe_percpu_receive_thread() */
		}
	}

	skb = (struct sk_buff *)fp->fr_free_priv;
	ASSERT(skb->data == (unsigned char *)fp->fr_hdr);

	sof = fp->fr_sof;
	eof = fp->fr_eof;
	hlen = sizeof(struct fcoe_hdr_old);
	tlen = sizeof(struct fcoe_crc_eof_old);

	ASSERT(!skb_is_nonlinear(skb));
	cp = (struct fcoe_crc_eof *)skb_put(skb, tlen);

	wlen = (skb->len - tlen + sizeof(u32)) / FCOE_WORD_TO_BYTE; /* u32 is CRC */

	eh = (struct ethhdr *)skb_push(skb, hlen + sizeof(struct ethhdr));
	if (priv->address_mode == FCOE_FCOUI_ADDR_MODE)
		fc_set_mac(eh->h_dest, (net24_t *) &fh->fh_d_id);
	else
		/* insert GW address */
		memcpy(eh->h_dest, priv->dest_addr, ETH_ALEN);

	if (unlikely(flogi_in_progress))
		memcpy(eh->h_source, priv->rfci[RFCI_CTRL].mac, ETH_ALEN);
	else
		memcpy(eh->h_source, priv->rfci[RFCI_DATA].mac, ETH_ALEN);
	
	eh->h_proto = htons(ETHERTYPE_FCOE);

	ASSERT(hlen == sizeof(*ohp));
	ohp = (struct fcoe_hdr_old *)(eh + 1);
	net16_put(&ohp->fcoe_plen, FCOE_ENCAPS_LEN_SOF(wlen, sof));

	fp->fr_dev = (void *)fc_dev;

	if (priv->need_update_data_mac) {
		priv->deffered_plogi_skb = skb;
		priv->deffered_plogi_eof = eof;
		FCPRINT("Deferring this RFCI send\n");
		return 0;
	}

	return fcadev->accel_ops.do_send(priv, RFCI_FLOGI, skb, eof);
}

int mlx4_fc_send_scsi(struct fcdev *fdev, struct fc_scsi_pkt *fsp)
{
	struct fc_accel_dev *fcadev = fdev->drv_priv;
	int err;

	err = fcadev->accel_ops.do_scsi_send(fcadev->drv_priv, fsp);
	if (err) {
		printk("Error do_send, err=%d\n", err);
		goto err_out;
	}

	return 0;

err_out:
	return err;
}

static int mlx4_fc_timeout(struct fcdev *dev, struct fc_scsi_pkt *fsp)
{
	struct openfc_softc *openfcp = fsp->openfcp;

	FCPRINT("Timeout called\n");

	if (!spin_trylock(&openfcp->outstandingcmd[fsp->idx].scsi_pkt_lock)) {
		int idx;
		struct os_lun *disk;

		idx = fsp->idx;
		fsp->state = OPENFC_SRB_ABORT_PENDING;
		if (fsp->seq_ptr) {
			FCPRINT("Abort exchange...\n");
		}
		fsp->seq_ptr = NULL;
		fsp->status_code = OPENFC_CMD_TIME_OUT;
		fsp->cdb_status = 0;
		fsp->io_status = 0;
		disk = fsp->disk;
		disk->error_cnt++;
		/*
		 * if we get 5 consecutive errors
		 * for the same command then offline the device.
		 */
		if (disk->error_cnt >= FC_MAX_ERROR_CNT) {
			disk->state = OPENFC_LUN_ERR;
			fsp->status_code = OPENFC_HRD_ERROR;
		} else {
			fsp->io_status = (SUGGEST_RETRY << 24);
		}
		if (fsp->done)
			(*fsp->done) (fsp);
	}

	spin_unlock(&openfcp->outstandingcmd[fsp->idx].scsi_pkt_lock);

	return OPENFC_TIMEOUT_SUCESS;
}

static void mlx4_frame_free(struct fc_frame *fp)
{
	kfree_skb((struct sk_buff *)fp->fr_free_priv);
}

static struct fc_frame *mlx4_frame_alloc(size_t len)
{
	struct fc_frame *fp;
	struct sk_buff *skb;

	ASSERT((len % sizeof(uint32_t)) == 0);
	len += sizeof(struct fc_frame_header);
	skb = dev_alloc_skb(len + sizeof(*fp) +
			    FC_FRAME_HEADROOM + FC_FRAME_TAILROOM);
	if (!skb)
		return NULL;
	fp = (struct fc_frame *)skb->head;
	fc_frame_init(fp);
	fp->fr_free = mlx4_frame_free;
	fp->fr_free_priv = skb;
	skb_reserve(skb, sizeof(*fp) + FC_FRAME_HEADROOM);
	skb_put(skb, len);
	fp->fr_hdr = (struct fc_frame_header *)skb->data;
	fp->fr_len = len;
	fp->fr_destructor = NULL;
	return fp;
}

struct openfc_port_operations mlx4_openfc_ops = {
	.owner = THIS_MODULE,
	.send = mlx4_fc_send,
	.frame_alloc = mlx4_frame_alloc,
	.send_scsi = mlx4_fc_send_scsi,
	.timeout_hdlr = mlx4_fc_timeout,
};

struct fc_accel_dev *mlx4_alloc_openfc(struct fc_accel_dev_ops fops, int sizeof_priv)
{
	struct fcdev *fdev;
	struct fc_accel_dev *accel_dev;
	
	fdev = openfc_alloc_dev(&mlx4_openfc_ops, sizeof(struct fc_accel_dev) + sizeof_priv);
	if (!fdev) {
		printk("could not allocate openfc dev, required priv size = %lu\n", sizeof(struct fc_accel_dev) + sizeof_priv);
		return NULL;
	}

	accel_dev = (struct fc_accel_dev *)fdev->drv_priv;
	accel_dev->fdev = fdev;
	accel_dev->accel_ops = fops;
	accel_dev->drv_priv = (void *)(accel_dev + 1);
	
	return accel_dev;
}
EXPORT_SYMBOL(mlx4_alloc_openfc);

int mlx4_init_openfc(struct fcdev *fdev, int auto_start)
{
	struct fc_accel_dev *fcadev = fdev->drv_priv;
	struct mlx4_fc_priv *priv = fcadev->drv_priv;
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	int i;
	int err;
	
	fdev->fd_link_status = TRANS_LINK_DOWN;
	fdev->capabilities = TRANS_C_SG;
	if (auto_start)
		fdev->options = TRANS_O_FCS_AUTO; /* automatically start state machine after registration */
	snprintf(fdev->ifname, IFNAMSIZ, "mlxfc%d", priv->port); /* validate less than IFNAMESIZ */
	FCPRINT("setting fdev ifname to %s\n", fdev->ifname);

	fdev->fd_wwnn = mfcdev->dev->caps.def_mac[priv->port] | ((u64)0x1000 << 48);
	fdev->fd_wwpn = mfcdev->dev->caps.def_mac[priv->port] | ((u64)0x2000 << 48);
	fdev->framesize = priv->fc_frame_size + 24 /*fc*/;
	snprintf(fdev->drv_info.model, 64, MLX4FC_DRIVER_NAME);
	snprintf(fdev->drv_info.vendor, 64, MLX4FC_DRIVER_VENDOR);
	snprintf(fdev->drv_info.model_desc, 64, MLX4FC_DRIVER_NAME);
	snprintf(fdev->drv_info.drv_version, 64, MLX4FC_DRIVER_VERSION);
	snprintf(fdev->drv_info.drv_name, 128, "%s", fdev->ifname);

	/* per cpu stats */
	for_each_online_cpu(i) {
		void *p = kzalloc(sizeof(struct fcoe_dev_stats), GFP_KERNEL);
		if (p)
			fdev->dev_stats[i] = p;
	}

	err = openfc_register(fdev);
	if (err < 0) {
		mlx4_err(mfcdev->dev, "Error registering with openfc, err=%d\n", err);
		return err;
	}
	FCPRINT("Registered with openfs, scsi host No. is %d\n", err);

	return 0;
}
EXPORT_SYMBOL(mlx4_init_openfc);

void mlx4_destroy_openfc(struct fcdev *fdev)
{
	int cpu;

	openfc_unregister(fdev);

	for_each_online_cpu(cpu) {
		void *p = fdev->dev_stats[cpu];
		if (p) {
			fdev->dev_stats[cpu] = NULL;
			kfree(p);
		}
	}

	FCPRINT("unregistered from openfc\n");
}
EXPORT_SYMBOL(mlx4_destroy_openfc);

/* This function is called from thread context - not the ISR */
int mlx4_fc_recv_scsi(struct fcdev *fc_dev, struct sk_buff *skb)
{
	struct fc_accel_dev *fca_dev = fc_dev->drv_priv;
	struct mlx4_fc_priv *priv = fca_dev->drv_priv;
	struct mlx4_fc_rcv_info *fr;
	struct fc_scsi_pkt *fsp = NULL;
	/* from openfc_scsi_rcv() */
	enum dma_data_direction data_dir;
	u_int8_t cdb_op;
	struct fcp_resp_ext *rp_ex;
	struct fcp_resp_rsp_info *fc_rp_info;
	u_int32_t respl = 0;
	u_int32_t snsl = 0;
	struct fcp_resp *fc_rp;

	fr = mlx4_fc_rcv_info_from_skb(skb);
	fsp = fr->fsp;

	/* things used to happen during states transient of sequence */
	CMD_SP(fsp->cmd) = MLX4_POISON;
	fsp->xfer_len = fr->rxcnt; // ... or txcnt, depends on dir

	/* things from openfc_scsi_rcv(), RESPONSE branch */
	fc_rp = (struct fcp_resp *)(skb->data + 26); // FCOE(2) + FC(24)
	fsp->seq_ptr = NULL;	/* released below with complete */
	fsp->cdb_status = fc_rp->fr_status;
	fsp->scsi_comp_flags = fc_rp->fr_flags;
	fsp->status_code = OPENFC_COMPLETE;
	if ((fc_rp->fr_flags == 0) && (fc_rp->fr_status == 0)) {
		data_dir = fsp->cmd->sc_data_direction;
		cdb_op = fsp->cmd->cmnd[0];
		if (data_dir == DMA_FROM_DEVICE) {
			if (fsp->data_len > fsp->xfer_len) {
				if ((cdb_op != READ_10 ||
				     cdb_op != READ_6 ||
				     cdb_op != WRITE_10 ||
				     cdb_op != WRITE_6)) {
					fsp->status_code = OPENFC_ERROR;
					fsp->io_status =
					    (SUGGEST_RETRY << 24);
					FCPRINT("fsp->data_len (%x) > fsp->xfer_len (%zx), and not READ/WRITE-10/6: SUGGEST_RETRY\n",
						fsp->data_len, fsp->xfer_len);
				} else {
					FCPRINT("xfer len less then "
						"data_len: xfer_len %zx"
						" data_len %x, OPENFC_DATA_UNDRUN \n",
						fsp->xfer_len,
						fsp->data_len);
					fsp->status_code =
					    OPENFC_DATA_UNDRUN;
				}
			} else if (fsp->data_len < fsp->xfer_len) {
				if ((cdb_op != READ_10 ||
				     cdb_op != READ_6 ||
				     cdb_op != WRITE_10 ||
				     cdb_op != WRITE_6)) {
					fsp->status_code = OPENFC_ERROR;
					fsp->io_status =
					    (SUGGEST_RETRY << 24);
					FCPRINT("fsp->data_len (%x) < fsp->xfer_len (%zx), and not READ/WRITE-10/6: SUGGEST_RETRY\n",
						fsp->data_len, fsp->xfer_len);
				} else {
					FCPRINT("xfer len greater than "
						"data_len xfer_len %zx"
						" data_len %x, OPENFC_DATA_OVRRUN \n",
						fsp->xfer_len,
						fsp->data_len);
					fsp->status_code =
					    OPENFC_DATA_OVRRUN;
				}
			}
		}
	} else {
		rp_ex = (void *)(fc_rp + 1);
		fc_rp_info = (struct fcp_resp_rsp_info *)(rp_ex + 1);
		if (fc_rp->fr_flags & FCP_RSP_LEN_VAL) {
			respl = net32_get(&rp_ex->fr_rsp_len);
			if ((respl != 0 && respl != 4 && respl != 8) ||
			    (fc_rp_info->rsp_code != FCP_TMF_CMPL)) {
				fsp->status_code = OPENFC_ERROR;
				FCPRINT("something about bad respl: %d, and rsp_code isn't FCP_TMF_CMPL... OPENFC_ERROR\n", respl);
			}
		}
		if (fc_rp->fr_flags & FCP_SNS_LEN_VAL) {
			snsl = net32_get(&rp_ex->fr_sns_len);
			if (snsl > SCSI_SENSE_BUFFERSIZE)
				snsl = SCSI_SENSE_BUFFERSIZE;
			memcpy(fsp->cmd->sense_buffer,
			       &fc_rp_info->_fr_resvd[0] + respl, snsl);
			//FCPRINT("Copied SENSE data\n");
		}
		if (fc_rp->fr_flags & FCP_RESID_UNDER) {
			fsp->scsi_resid = net32_get(&rp_ex->fr_resid);
			/*
			 * The cmnd->underflow is the minimum number of
			 * bytes that must be transfered for this
			 * command.  Provided a sense condition is not
			 * present, make sure the actual amount
			 * transferred is at least the underflow value
			 * or fail.
			 */
			if (!(fc_rp->fr_flags & FCP_SNS_LEN_VAL) &&
			    (fc_rp->fr_status == 0) &&
			    (fsp->cmd->request_bufflen -
			     fsp->scsi_resid) < fsp->cmd->underflow) {
				fsp->status_code = OPENFC_ERROR;
				FCPRINT("OPENFC_ERROR - FCP_RESID_UNDER...\n"); 
			}
		} else if (fc_rp->fr_flags & FCP_RESID_OVER) {
			fsp->status_code = OPENFC_ERROR;
			FCPRINT("OPENFC_ERROR - FCP_RESID_OVER\n");
		}
	}
	/* ... done */

	/* Clean our stuff */
	ASSERT_PR(fr->fmr, "fr: xno=0x%x, rxcnt=0x%x, txcnt=0x%x, bcnt=0x%x\n", 
		  fr->xno, fr->rxcnt, fr->txcnt, fr->bcnt);

	
	fca_dev->accel_ops.clean_exch(priv, fsp->cmd, fr->xno);
	kfree_skb(skb);

	fsp->state = OPENFC_SRB_RCV_STATUS;
	(*fsp->done) (fsp);

	return 0;
}

static void fcoe_recv_flogi(struct fcdev *fc_dev, struct fc_frame *fp, u64 sa)
{
	struct fc_accel_dev *fcadev = fc_dev->drv_priv;
	struct mlx4_fc_priv *priv = fcadev->drv_priv;
	struct fc_frame_header *fh;
	u8 op;

	fh = fc_frame_header_get(fp);
	if (fh->fh_type != FC_TYPE_ELS)
		return;
	op = fc_frame_payload_op(fp);
	if (op == ELS_LS_ACC && fh->fh_r_ctl == FC_RCTL_ELS_REP &&
	    priv->flogi_oxid == net16_get(&fh->fh_ox_id)) {
		if (sa == FCOE_FLOGI_MAC) {
			priv->address_mode = FCOE_FCOUI_ADDR_MODE;
		} else {
			net48_put((net48_t *)&priv->dest_addr, sa);
			priv->address_mode = FCOE_GW_ADDR_MODE;
		}

		priv->my_fid = fh->fh_d_id;
		fc_set_mac(priv->rfci[RFCI_DATA].mac, &fh->fh_d_id);
		fcadev->accel_ops.flogi_finished(fc_dev);
		priv->flogi_progress = 0;
	} else if (op == ELS_FLOGI && fh->fh_r_ctl == FC_RCTL_ELS_REQ && sa) {
		net48_put((net48_t *)&priv->dest_addr, sa);
		priv->address_mode = FCOE_GW_ADDR_MODE;
	}
}

int mlx4_fc_rx_thread_func(void *arg)
{
	struct mlx4_fc_rx_thread *p = arg;
	u_int32_t fr_len;
	u_int hlen;
	u_int tlen;
	struct fcdev *fc_dev;
	struct mlx4_fc_rcv_info *fr;
	struct fcoe_dev_stats *stats;
	struct fc_frame_header *fh;
	struct sk_buff *skb;
	struct fcoe_crc_eof_old *cp;
	enum fc_sof sof;
	struct fc_frame *fp;
	u64 mac = 0;
	struct fc_accel_dev *fcadev;
	struct mlx4_fc_priv *priv;
	struct fcoe_hdr_old *fchp;
	u_int len;

	set_user_nice(current, 19);

	while (!kthread_should_stop()) {
		long flags = 0;

		spin_lock_irqsave(&p->rx_list.lock, flags);
		while ((skb = __skb_dequeue(&p->rx_list)) == NULL) {
			set_current_state(TASK_INTERRUPTIBLE);
			spin_unlock_irqrestore(&p->rx_list.lock, flags);
			schedule();
			set_current_state(TASK_RUNNING);
			if (kthread_should_stop())
				return 0;
			spin_lock_irqsave(&p->rx_list.lock, flags);
		}
		spin_unlock_irqrestore(&p->rx_list.lock, flags);
		fr = mlx4_fc_rcv_info_from_skb(skb);
		fc_dev = fr->fd;
		if (unlikely(fc_dev == NULL)) {
			FCPRINT("invalid HBA Structure");
			kfree_skb(skb);
			continue;
		}

		stats = fc_dev->dev_stats[smp_processor_id()];

		#ifdef DEBUG_CPU_BINDING
		if (fr->xno % num_online_cpus() != smp_processor_id())
			FCPRINT("SMP: processing xno %d on thread running on cpu %d instead of %d\n", 
				fr->xno, smp_processor_id(), fr->xno % num_online_cpus());
		#endif

		fcadev = (struct fc_accel_dev *)fc_dev->drv_priv;
 		priv = fcadev->drv_priv;

		if (fr->fsp) {
			// This is a fexch completion
			mlx4_fc_recv_scsi(fc_dev, skb);
			continue;
		}

		/*
		 * Save source MAC address before discarding header.
		 */
		if (unlikely(priv->flogi_progress))
			mac = net48_get((net48_t *) eth_hdr(skb)->h_source);

		/*
		 * Check the header and pull it off.
		 */
		hlen = priv->fcoe_hlen;
		ASSERT(hlen == sizeof(*fchp));
		fchp = (struct fcoe_hdr_old *)skb->data;
		len = net16_get(&fchp->fcoe_plen);
		skb_pull(skb, sizeof(*fchp));
		fr_len = FCOE_DECAPS_LEN(len);
		fr_len = fr_len * FCOE_WORD_TO_BYTE;
		if (unlikely(fr_len < sizeof(struct fc_frame_header) +
			     sizeof(cp->fcoe_crc32))) {
			if (stats->ErrorFrames < 5)
				FCPRINT("length error: len_sof %x", len);
			stats->ErrorFrames++;
			kfree_skb(skb);
			continue;
		}
		sof = FCOE_DECAPS_SOF(len);
		fr_len -= sizeof(cp->fcoe_crc32);
		tlen = sizeof(struct fcoe_crc_eof_old);

		if (skb_is_nonlinear(skb))
			skb_linearize(skb);	/* not ideal */
		if (unlikely(fr_len + tlen > skb->len)) {
			if (stats->ErrorFrames < 5)
				FCPRINT("length error fr_len 0x%x skb->len 0x%x",
					fr_len + tlen, skb->len);
			stats->ErrorFrames++;
			kfree_skb(skb);
			continue;
		}
		stats->RxFrames++;
		stats->RxWords += fr_len / FCOE_WORD_TO_BYTE;

		fp = &p->frame;
		fc_frame_init_static(fp);
		fp->fr_hdr = (struct fc_frame_header *)skb->data;
		fp->fr_len = fr_len;
		cp = (struct fcoe_crc_eof_old *)(skb->data + fr_len);
		fp->fr_eof = cp->fcoe_eof;
		fp->fr_sof = sof;

		fh = fc_frame_header_get(fp);
		ASSERT(fh);
		if (unlikely(priv->flogi_progress))
			fcoe_recv_flogi(fc_dev, fp, mac);

		openfc_rcv(fc_dev, fp);

		ASSERT(fc_frame_freed_static(fp));
		kfree_skb(skb);
		if (priv->need_update_data_mac) {
			priv->need_update_data_mac = 0;

			fcadev->accel_ops.flogi_finished(fc_dev);
		}
	}

	return 0;
}

int mlx4_fc_rcv(struct fcdev *fdev, struct sk_buff *skb)
{
	struct mlx4_fc_rcv_info *fr;
	struct mlx4_fc_priv *priv = (struct mlx4_fc_priv *)fdev->drv_priv;
	
	u_int8_t *data;
	struct fc_frame_header *fh;
	ushort oxid;
	int cpu_idx;
	struct mlx4_fc_rx_thread *p;
	long flags = 0;
	
	data = skb->data;
	data += priv->fcoe_hlen;
	fh = (struct fc_frame_header *)data;
	oxid = net16_get(&fh->fh_ox_id);

	fr = mlx4_fc_rcv_info_from_skb(skb);
	fr->fd = fdev;

	cpu_idx = oxid % num_online_cpus();
	p = rx_thread_percpu[cpu_idx];
	ASSERT(p);

	spin_lock_irqsave(&p->rx_list.lock, flags);
	__skb_queue_tail(&p->rx_list, skb);
	if (p->rx_list.qlen == 1)
		wake_up_process(p->thread);

	spin_unlock_irqrestore(&p->rx_list.lock, flags);

	return 0;
}
EXPORT_SYMBOL(mlx4_fc_rcv);

int mlx4_fc_init_rx(void)
{
	struct mlx4_fc_rx_thread *p;
	int cpu;
	
	for_each_online_cpu(cpu) {
		p = kzalloc(sizeof(struct mlx4_fc_rx_thread), GFP_KERNEL);
		if (!p) {
			printk("Could not allocate rx_thread\n");
			return -ENOMEM;
		}
		p->thread = kthread_create(mlx4_fc_rx_thread_func,
					   (void *)p,
					   "mlx4fc_rx/%d", cpu);

		if (!IS_ERR(p->thread)) {
			p->cpu = cpu;
			rx_thread_percpu[cpu] = p;
			skb_queue_head_init(&p->rx_list);
			kthread_bind(p->thread, cpu);
			wake_up_process(p->thread);
		} else {
			rx_thread_percpu[cpu] = NULL;
			kfree(p);
			return -ENOMEM;
		}
	}

	return 0;
}
EXPORT_SYMBOL(mlx4_fc_init_rx);

void mlx4_fc_kill_rx(void)
{
	struct mlx4_fc_rx_thread *p;
	struct sk_buff *skb;
	int idx;
	
	for (idx = 0; idx < MLX4_NR_CPUS; idx++) {
		if (rx_thread_percpu[idx]) {
			long flags;
			kthread_stop(rx_thread_percpu[idx]->thread);
			p = rx_thread_percpu[idx];
			spin_lock_irqsave(&p->rx_list.lock, flags);
			while ((skb = __skb_dequeue(&p->rx_list)) != NULL)
				kfree_skb(skb);
			spin_unlock_irqrestore(&p->rx_list.lock, flags);
			kfree(rx_thread_percpu[idx]);
		}
	}
}
EXPORT_SYMBOL(mlx4_fc_kill_rx);

int __init mlx4_ofc_init(void)
{
	return 0;
}

void __exit mlx4_ofc_cleanup(void)
{
}

module_init(mlx4_ofc_init);
module_exit(mlx4_ofc_cleanup);

