/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/delay.h>
#include <linux/if_vlan.h>

#include <linux/mlx4/device.h>
#include <linux/mlx4/cmd.h>

#include "mlx4_fc.h"
#include "mlx4_ofc.h"

int mlx4_fc_frame_size = MLX4_DEFAULT_FC_MTU;
module_param_named(frame_size, mlx4_fc_frame_size, int, 0644);
MODULE_PARM_DESC(frame_size, "Frame size default is " __stringify(MLX4_DEFAULT_FC_MTU));

int mlx4_fc_auto_start = 1;
module_param_named(auto_start, mlx4_fc_auto_start, int, 0644);
MODULE_PARM_DESC(auto_start, "Automatic FC login. Default = 1 (yes)");

static struct nport_id_entry fc_nport_id_table[MLX4_MAX_PORTS][MLX4_FC_NUM_NPORT_IDS];

static int mlx4_fc_SET_VLAN_FLTR(struct mlx4_dev *dev, u8 port, int vid)
{
        struct mlx4_cmd_mailbox *mailbox;
        struct mlx4_set_vlan_fltr_mbox *filter;
        int i, j, index = 0;
        u32 entry;
        int err = 0;

        mailbox = mlx4_alloc_cmd_mailbox(dev);
        if (IS_ERR(mailbox))
                return PTR_ERR(mailbox);

        filter = mailbox->buf;
        if (vid != -1) {
                memset(filter, 0, sizeof *filter);
                for (i = VLAN_FLTR_SIZE - 1; i >= 0; i--) {
                        entry = 0;
                        for (j = 0; j < 32; j++)
                                if (vid == index++)
                                        entry |= 1 << j;
                        filter->entry[i] = cpu_to_be32(entry);
                }
        } else {
                /* When no vlans are configured we disable the filter
                * (i.e., pass all vlans) because we ignore them anyhow */
                memset(filter, 0xff, sizeof(*filter));
        }
        err = mlx4_cmd(dev, mailbox->dma, port, 0, MLX4_CMD_SET_VLAN_FLTR,
                       MLX4_CMD_TIME_CLASS_B);
        mlx4_free_cmd_mailbox(dev, mailbox);
        return err;
}

void init_nport_id_table(void)
{
	memset(&fc_nport_id_table, 0, sizeof(fc_nport_id_table));
}

static int mlx4_CONFIG_FC_BASIC(struct mlx4_dev *dev, u8 port, struct mlx4_fc_basic_config_params *params)
{
	struct mlx4_cmd_mailbox *mailbox;
	int err;

#define CONFIG_FC_FEXCH_BASE_OFFSET 	0x0
#define CONFIG_FC_NM_OFFSET		0x5
#define CONFIG_FC_NV_OFFSET		0x6
#define CONFIG_FC_NP_OOFSET		0x7
#define CONFIG_FC_BASEMPT_OFFSET	0x8
#define CONFIG_FC_NUM_RFCI_OFFSET	0xc
#define CONFIG_FC_RFCI_BASE_OFFSET	0xd
#define CONFIG_FC_PROMISC_QPN_OFFSET	0x14
#define CONFIG_FC_MCAST_QPN_OFFSET	0x18

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		return PTR_ERR(mailbox);

	memset(mailbox->buf, 0, 256);

	MLX4_PUT(mailbox->buf, params->fexch_base, 	CONFIG_FC_FEXCH_BASE_OFFSET);
	MLX4_PUT(mailbox->buf, params->nm, 		CONFIG_FC_NM_OFFSET);
	MLX4_PUT(mailbox->buf, params->nv,		CONFIG_FC_NV_OFFSET);
	MLX4_PUT(mailbox->buf, params->np,		CONFIG_FC_NP_OOFSET);
	MLX4_PUT(mailbox->buf, (params->fexch_base_mpt << 8), CONFIG_FC_BASEMPT_OFFSET);
	MLX4_PUT(mailbox->buf, params->rfci_base | (((u32)params->log_num_rfci) << 24),
							CONFIG_FC_NUM_RFCI_OFFSET);
	MLX4_PUT(mailbox->buf, params->def_fcoe_promisc_qpn, CONFIG_FC_PROMISC_QPN_OFFSET);
	MLX4_PUT(mailbox->buf, params->def_fcoe_mcast_qpn, CONFIG_FC_MCAST_QPN_OFFSET);

	FCPRINT("mlx4_cmd, InMod=0x%x, OpMod=0x%x, Op=0x%x, mailbox:\n",
		MLX4_CMD_INMOD_BASIC_CONF | port, 
		MLX4_CMD_MOD_FC_ENABLE, 
		MLX4_CMD_CONFIG_FC);
	err = 0;
	err = mlx4_cmd(dev, mailbox->dma, 
		       MLX4_CMD_INMOD_BASIC_CONF | port, 
		       MLX4_CMD_MOD_FC_ENABLE, 
		       MLX4_CMD_CONFIG_FC,
		       MLX4_CMD_TIME_CLASS_B);
	
	mlx4_free_cmd_mailbox(dev, mailbox);
	return err;

}

static int mlx4_CONFIG_FC_NPORT_ID(struct mlx4_dev *dev, u8 port, struct nport_id_entry *table)
{
	struct mlx4_cmd_mailbox *mailbox;
	int err;

	mailbox = mlx4_alloc_cmd_mailbox(dev);
	if (IS_ERR(mailbox))
		return PTR_ERR(mailbox);

	memcpy(mailbox->buf, table, MLX4_FC_NUM_NPORT_IDS * sizeof(u32));

	err = 0;
	
	err = mlx4_cmd(dev, mailbox->dma, 
		       MLX4_CMD_INMOD_NPORT_TAB | port, 
		       MLX4_CMD_MOD_FC_ENABLE, 
		       MLX4_CMD_CONFIG_FC,
		       MLX4_CMD_TIME_CLASS_B);
	
	mlx4_free_cmd_mailbox(dev, mailbox);
	return err;

}

int mlx4_fc_config_nport_id(struct mlx4_fc_priv *priv, int idx, net24_t nport_id)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	int port = priv->port;
	
	if ( (idx < 0) || (idx >= MLX4_FC_NUM_NPORT_IDS) )
		return -1;

	fc_nport_id_table[port - 1][idx].fid = nport_id;
	return mlx4_CONFIG_FC_NPORT_ID(mfcdev->dev, port, fc_nport_id_table[port - 1]);
}

int mlx4_QUERY_PORT(struct mlx4_dev *dev, u8 port, int *link_up)
{
        struct mlx4_cmd_mailbox *mailbox;
        struct mlx4_query_port_context *context;
        int err = 0;

        mailbox = mlx4_alloc_cmd_mailbox(dev);
        if (IS_ERR(mailbox))
                return PTR_ERR(mailbox);

        err = mlx4_cmd_box(dev, 0, mailbox->dma, port, 0, MLX4_CMD_QUERY_PORT,
                           MLX4_CMD_TIME_CLASS_B);
        if (err) {
                mlx4_err(dev, "Query port returned error: %d\n", err);
                goto query_err;
        }

        context = (struct mlx4_query_port_context *) mailbox->buf;

        /* this is a double-checking... mlx4_core has done it for us */
        if (!(context->supported_port_type & MLX4_PORT_TYPE_ETH)) {
                mlx4_err(dev, "Port does not support ETH\n");
                goto query_err;
        }

        *link_up = context->supported_port_type & QUERY_PORT_LINK_MASK;
        FCPRINT("Port %d link is %s\n", port, (*link_up ? "UP" : "DOWN"));

query_err:
        mlx4_free_cmd_mailbox(dev, mailbox);
        return err;
}

int mlx4_fc_flogi_finished(struct fcdev *fc_dev)
{
	struct fc_accel_dev *fcadev = fc_dev->drv_priv;
	struct mlx4_fc_priv *priv = fcadev->drv_priv;
	int err = 0;
	net24_t fid = priv->my_fid;

	FCPRINT_MAC("Finishing flogi process - starting RFCI for mac:", priv->rfci[RFCI_DATA].mac);
	err = mlx4_start_rfci_data(priv, mac_to_u64(priv->rfci[RFCI_DATA].mac));
	if (err) {
		mlx4_err(priv->mfcdev->dev, "Could not init raw data packets channel, err=%d\n", err);
		goto err;
	}

	err = mlx4_fc_init_fcmd(priv);
	if (err) {
		mlx4_err(priv->mfcdev->dev, "Could not init FCMD , err=%d\n", err);
		goto err;
	}
	FCPRINT("FCMD in state RTS\n");

	err = mlx4_fc_config_nport_id(priv, 0, fid);
	if (err) {
		mlx4_err(priv->mfcdev->dev, "Could not config NPort ID to port %d\n", priv->port);
		goto err;
	}
	FCPRINT("CONFIG_FC NPort ID %x:%x:%x successfull for port %d\n", fid.net_data[0], fid.net_data[1], fid.net_data[2], priv->port);

	if (priv->deffered_plogi_skb) {
		FCPRINT("Sending a deferred RFCI packet\n");
		mlx4_do_rfci_xmit(priv, RFCI_DATA, priv->deffered_plogi_skb, priv->deffered_plogi_eof);
		priv->deffered_plogi_skb = NULL;
	}

err:
	return err;
}

struct fc_accel_dev_ops fcaops = {
	.clean_exch = mlx4_fc_clean_exch,
	.do_scsi_send = mlx4_do_fcmd_send,
	.do_send = mlx4_do_rfci_xmit,
	.flogi_finished = mlx4_fc_flogi_finished,
};

int mlx4_fc_init_port(struct mlx4_fc_dev *mfcdev, int port)
{
	struct fc_accel_dev *afdev;
	struct fcdev *fdev;
	struct mlx4_fc_priv *priv;
	int err = 0;
	int mvp = (1 << mfcdev->log_num_mac) * (1 << mfcdev->log_num_vlan) * (1 << mfcdev->log_num_prio);
	struct mlx4_caps *caps = &mfcdev->dev->caps;
	int link_up = -1;
	
	FCPRINT("port %d is ETH, binding.\n", port);
	
	afdev = mlx4_alloc_openfc(fcaops, sizeof(struct mlx4_fc_priv));
	if (!afdev) {
		err = -ENOMEM;
		goto err_out;
	}

	fdev = afdev->fdev;
	mfcdev->fdev[port] = fdev;
	
	priv = (struct mlx4_fc_priv *)afdev->drv_priv;

	priv->is_initialized = 0;

	priv->port = port;
	priv->mfcdev = mfcdev;
	priv->base_rfci_qpn = mfcdev->base_rfci_qpn + (port - 1) * mvp;
	priv->base_fexch_qpn = mfcdev->base_fexch_qpn + (port - 1) * mfcdev->num_fexch_qps_perport;
	priv->base_fexch_mpt = mfcdev->base_fexch_mpt + (port - 1) * mfcdev->num_fexch_qps_perport;
	priv->rfci[RFCI_DATA].fc_mac_idx = -1;
	priv->need_update_data_mac = 0;
	priv->deffered_plogi_skb = NULL;
	priv->n_m = 2;
	priv->n_v = mfcdev->log_num_vlan;
	priv->rfci_rx_enabled = 0;

	priv->fcoe_hlen = sizeof(struct fcoe_hdr_old); /*(cfg->fcoe_type & FCOE_CFG_TYPE_T11) ?
	    sizeof(struct fcoe_hdr) : sizeof(struct fcoe_hdr_old); */

	FCPRINT("port %d default mac=0x%llx, max_mtu=%d\n",
		port, caps->def_mac[priv->port], caps->eth_mtu_cap[priv->port]);

	/* Configure port */
	priv->fc_frame_size = MIN(mlx4_fc_frame_size, (caps->eth_mtu_cap[priv->port] 
				  - 14/*mac*/ - 2/*fcoe*/ - 24/*fc*/ - 5/*crceof*/) & ~3);
	FCPRINT("Setting port FC frame size to %d\n", priv->fc_frame_size);
	
	{
		struct mlx4_fc_basic_config_params params = {
			.fexch_base = priv->base_fexch_qpn,
			.nm = priv->n_m,
			.nv = priv->n_v,
			.np = 0,
			.fexch_base_mpt = priv->base_fexch_mpt,
			.log_num_rfci = ilog2(mvp),
			.rfci_base = priv->base_rfci_qpn,
			.def_fcoe_promisc_qpn = 0x34,
			.def_fcoe_mcast_qpn = (2 << 30) | 0x333,
		};
		FCPRINT("base_fexch=0x%x, base_mpt=0x%x, base_rfci=0x%x, num_rfci=0x%x\n", 
			params.fexch_base, params.fexch_base_mpt, params.rfci_base, (1 << params.log_num_rfci));

		err = mlx4_CONFIG_FC_BASIC(mfcdev->dev, port, &params);
		if (err) {
			mlx4_err(mfcdev->dev, "Could not issue CONFIG_FC Basic command on port %d\n", port);
			goto err_free_fdev;
		}
		FCPRINT("CONFIG_FC Basic successfull for port %d\n", port);
	}

	/* Init port */
	FCPRINT("Initializing port\n");
	err = mlx4_INIT_PORT(mfcdev->dev, priv->port);
	if (err) {
		mlx4_err(mfcdev->dev, "Failed Initializing port\n");
		goto err_free_fdev;
	}
	FCPRINT("port %d initialized\n", port);

	mdelay(10); /* Let link state stabilize if cable was connected */

	mlx4_fc_SET_VLAN_FLTR(mfcdev->dev, port, mlx4_fc_vlan_id);

	priv->fc_vlan_idx = -1;
	if (mlx4_fc_vlan_id != -1) {
		err = mlx4_register_vlan(mfcdev->dev, port, mlx4_fc_vlan_id, &priv->fc_vlan_idx); 
		if (err) {
			mlx4_err(mfcdev->dev, "Could not register VLAN %d. err = %x\n", mlx4_fc_vlan_id, err);
			goto err_free_fdev;
		}
		FCPRINT("Registered vlan %d to index %d\n", mlx4_fc_vlan_id, priv->fc_vlan_idx);
	}

	u64_to_mac(priv->rfci[RFCI_CTRL].mac, caps->def_mac[priv->port]);

	err = mlx4_fc_create_rfci(priv, &priv->rfci[RFCI_CTRL], caps->def_mac[priv->port]);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not create default MAC RFCI QP for port %d, err=%d\n", port, err);
		goto err_unreg_vlan;
	}
	FCPRINT("Created RFCI for real dev MAC on port %d\n", port);

	err = mlx4_fc_create_rfci(priv, &priv->rfci[RFCI_FLOGI], FCOE_FLOGI_MAC);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not create FLOGI RFCI QP for port %d, err=%d\n", port, err);
		goto err_destroy_rfci_ctrl;
	}
	FCPRINT("Created RFCI for FLOGI MAC on port %d\n", port);
	
	err = mlx4_fc_create_fcmd(priv);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not create FCMD QP for port %d, err=%d\n", port, err);
		goto err_destroy_rfci_flogi;
	}
	FCPRINT("FCMD created for port %d\n", port);

	err = mlx4_init_openfc(fdev, mlx4_fc_auto_start);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not init openfc device\n");
		goto err_destroy_fcmd;
	}

	err = mlx4_QUERY_PORT(mfcdev->dev, priv->port, &link_up);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not get port state\n");
		goto err_destroy_fcmd;
	}
	
	if (link_up) {
		fdev->fd_link_status = TRANS_LINK_UP;
		openfc_linkup(fdev);
	}

	err = mlx4_fc_init_rfci(priv, &priv->rfci[RFCI_CTRL]);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not init RFCI for port %d, err=%d\n", port, err);
		goto err_destroy_fcmd;
	}
	FCPRINT("RFCI in state RTS for port %d\n", port);

	err = mlx4_fc_init_rfci(priv, &priv->rfci[RFCI_FLOGI]);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not init FLOGI RFCI for port %d, err=%d\n", port, err);
		goto err_destroy_fcmd;
	}
	FCPRINT("RFCI in state RTS for port %d\n", port);

	priv->is_initialized = 1;
	
	return 0;

err_destroy_fcmd:
	mlx4_fc_destroy_fcmd(priv);

err_destroy_rfci_flogi:
        mlx4_fc_destroy_rfci(priv, &priv->rfci[RFCI_FLOGI]);
err_destroy_rfci_ctrl: 
	mlx4_fc_destroy_rfci(priv, &priv->rfci[RFCI_CTRL]);
err_unreg_vlan:
	if (priv->fc_vlan_idx != -1)
		mlx4_unregister_vlan(mfcdev->dev, port, priv->fc_vlan_idx); 
err_free_fdev:
	openfc_put_dev(fdev);
err_out:
	return err;
}

void mlx4_fc_free_port(struct mlx4_fc_dev *mfcdev, int port)
{
	struct fcdev *fdev = mfcdev->fdev[port];
	struct mlx4_fc_priv *priv = ((struct fc_accel_dev *)fdev->drv_priv)->drv_priv;

	if (!priv->is_initialized)
		return;

	openfc_linkdown(fdev);

	mlx4_destroy_openfc(fdev);
	mlx4_fc_destroy_fcmd(priv);
	mlx4_fc_destroy_rfci(priv, &priv->rfci[RFCI_FLOGI]);
	mlx4_fc_destroy_rfci(priv, &priv->rfci[RFCI_CTRL]);

	if (priv->rfci[RFCI_DATA].fc_mac_idx != -1)
		mlx4_fc_destroy_rfci(priv, &priv->rfci[RFCI_DATA]);

	if (priv->fc_vlan_idx != -1)
		mlx4_unregister_vlan(mfcdev->dev, port, priv->fc_vlan_idx); 

	openfc_put_dev(fdev);
	mfcdev->fdev[port] = NULL;
	FCPRINT("port %d free\n", port);
}


