/*
 * Copyright (c) 2008 Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>

#include <linux/mlx4/driver.h>
#include <linux/mlx4/cmd.h>
#include <linux/mlx4/qp.h>
#include <linux/mlx4/cq.h>

#include "mlx4_fc.h"
#include "mlx4_ofc.h"

#define MLX4_CQE_QPN_MASK 0x00ffffff

static struct mlx4_fc_rfci *rfci_by_qpn(struct mlx4_fc_priv *priv, int qpn)
{
	int i;

	for (i = 0; (i < RFCI_NUM) && (priv->rfci[i].fc_qp.mqp.qpn != qpn); i++);
	
	ASSERT_PR(i < RFCI_NUM, "Couldn't find matching rfci for qpn %d\n", qpn);

	return &priv->rfci[i];
}

static void mlx4_fc_rfci_tx_comp(struct mlx4_fc_priv *priv, struct mlx4_cqe *cqe) 
{
	struct mlx4_fc_rfci *rfci;
	struct mlx4_fc_q *sq;
	struct sk_buff *skb;
	struct mlx4_fc_rfci_tx_desc *tx_desc;
	u32 index;

	rfci = rfci_by_qpn(priv, be32_to_cpu(cqe->my_qpn) & MLX4_CQE_QPN_MASK);

	sq = &rfci->fc_qp.sq;
	index = be16_to_cpu(cqe->wqe_index) & sq->size_mask;
	tx_desc = sq->buf + index * RFCI_SQ_BB_SIZE;
	
	pci_unmap_single(priv->mfcdev->dev->pdev,
			 be64_to_cpu(tx_desc->data.addr),
			 be32_to_cpu(tx_desc->data.count),
			 PCI_DMA_TODEVICE);

	ASSERT(sq->info);
	ASSERT(sq->info[index].buf);
	skb = (struct sk_buff *)sq->info[index].buf;
	kfree_skb(skb);
	sq->info[index].buf = NULL;

	++sq->cons;
}

static void mlx4_fc_rfci_rx_comp(struct mlx4_fc_priv *priv, struct mlx4_cqe *cqe) 
{
	struct mlx4_fc_rfci *rfci;
	struct mlx4_fc_q *rq;
	struct mlx4_fc_rx_desc *rx_desc;
	u32 index;
	int len;
	struct sk_buff *skb;
	int ret;

	rfci = rfci_by_qpn(priv, be32_to_cpu(cqe->my_qpn) & MLX4_CQE_QPN_MASK);
	rq = &rfci->fc_qp.rq;
	index = be16_to_cpu(cqe->wqe_index) & rq->size_mask;
	rx_desc = rq->buf + (index * rq->stride);

	pci_unmap_single(priv->mfcdev->dev->pdev,
			 be64_to_cpu(rx_desc->data[0].addr),
			 be32_to_cpu(rx_desc->data[0].count),
			 PCI_DMA_FROMDEVICE);

	ASSERT(rq->info);
	skb = rq->info[index].buf;
	ASSERT(skb);

	len = be32_to_cpu(cqe->byte_cnt);

	skb_put(skb, len);
	skb_set_mac_header(skb, 0);
	skb_pull(skb, ETH_HLEN);

	if (priv->rfci_rx_enabled) {
		mlx4_fc_rcv(priv->mfcdev->fdev[priv->port], skb);
	} else {
		FCPRINT("RFCI RX ignored till host started\n");
		kfree_skb(skb);
	}

	rfci->fc_qp.rq.cons++;

	ret = mlx4_fc_post_rx_buf(priv->mfcdev, &rfci->fc_qp, MLX4_FC_RFCI_RX_SKB_BUFSIZE, NULL);
	ASSERT_PR(!ret, "Could not post a single recv buff on rfci\n");
}

int mlx4_fc_create_rfci(struct mlx4_fc_priv *priv, struct mlx4_fc_rfci *rfci, u64 mac)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_qp *qp = &rfci->fc_qp;
	struct mlx4_fc_q *sq = &qp->sq;
	struct mlx4_fc_q *rq = &qp->rq;
	u32 qpn;
	int err=0;
	int i;

	FCPRINT("creating RFCI for mac 0x%llx\n", mac);

	spin_lock_init(&rfci->lock);
	
	err = mlx4_register_mac(mfcdev->dev, priv->port, mac, &rfci->fc_mac_idx); 
	if (err) {
		mlx4_err(mfcdev->dev, "Could not register MAC\n");
		goto err_out;
	}
	
	err = mlx4_fc_q_init(sq, RFCI_SQ_BB_SIZE, RFCI_SQ_NUM_BBS);
	if (err) {
		mlx4_err(mfcdev->dev, "Error initializing rfci sq\n");
		goto err_unreg_mac;
	}

	err = mlx4_fc_q_init(rq, RFCI_RQ_WQE_SIZE, RFCI_RQ_NUM_WQES);
	if (err) {
		mlx4_err(mfcdev->dev, "Error initializing rfci rq\n");
		err = -ENOMEM;
		goto err_free_txinfo;
	}
	
	qp->buf_size = (sq->size *  sq->stride) + (rq->size * rq->stride);

	err = mlx4_alloc_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size, qp->buf_size);
	if (err)
		goto err_free_rxinfo;

	if (RFCI_SQ_BB_SIZE >= RFCI_RQ_WQE_SIZE) {
		sq->buf = qp->wqres.buf.u.direct.buf;
		rq->buf = sq->buf + (sq->size * sq->stride);
	} else {
		rq->buf = qp->wqres.buf.u.direct.buf;
		sq->buf = rq->buf + (rq->size * rq->stride);
	}
	
	*qp->wqres.db.db = 0;

	mlx4_fc_stamp_sq(sq);

	qpn = priv->base_rfci_qpn | (rfci->fc_mac_idx << priv->n_v);
	if (priv->fc_vlan_idx != -1)
		qpn |= priv->fc_vlan_idx;

	FCPRINT("using QPN=0x%x (mfcdev base is 0x%x, priv base is 0x%x)\n", qpn, 
		mfcdev->base_rfci_qpn, priv->base_rfci_qpn);

	err = mlx4_qp_alloc(mfcdev->dev, qpn, &rfci->fc_qp.mqp);
	if (err) {
		mlx4_err(mfcdev->dev, "Could not allocate QP number 0x%x\n", qpn);
		goto err_free_man;
	}

	qp->doorbell_qpn = swab32(qp->mqp.qpn << 8);

	err = mlx4_fc_create_cq(priv, &rfci->fc_cq, MLX4_FC_RFCI_CQ_ENTRIES, 
				0, /* don't care which cpu */
				1, /* arm cq */
				mlx4_fc_rfci_rx_comp, mlx4_fc_rfci_tx_comp, "RFCI");
	if (err) {
		mlx4_err(mfcdev->dev, "Failed creating RFCI CQ for port %d, err=%d\n", priv->port, err);
		goto err_free_qp;
	}

	FCPRINT("Created RFCI CQ successfull for port %d, CQN=0x%x\n", priv->port, rfci->fc_cq.mcq.cqn);

	for (i = 0; i < rq->size - 1; i++) {
		err =  mlx4_fc_post_rx_buf(mfcdev, &rfci->fc_qp, MLX4_FC_RFCI_RX_SKB_BUFSIZE, NULL);
		if (err) {
			FCPRINT("Failed preparing RX descriptor[%d] for RFCI\n", i);
			goto err_free_qp;
		}
	}
	mlx4_fc_ring_db_rx(&rfci->fc_qp);
	
	return 0;

err_free_qp:
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
err_free_man:
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
err_free_rxinfo:
	vfree(rq->info);
	rq->info = NULL;
err_free_txinfo:
	vfree(sq->info);
	sq->info = NULL;
err_unreg_mac:
	mlx4_unregister_mac(mfcdev->dev, priv->port, rfci->fc_mac_idx);

err_out:
	return err;
}

void mlx4_fc_destroy_rfci(struct mlx4_fc_priv *priv, struct mlx4_fc_rfci *rfci)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_qp *qp = &rfci->fc_qp;
	struct mlx4_fc_q *sq = &qp->sq;
	struct mlx4_fc_q *rq = &qp->rq;

	mlx4_fc_destroy_cq(&rfci->fc_cq);
	mlx4_qp_to_reset(mfcdev->dev, &qp->mqp);
	mlx4_qp_remove(mfcdev->dev, &qp->mqp);
	mlx4_qp_free(mfcdev->dev, &qp->mqp);
	mlx4_free_hwq_res(mfcdev->dev, &qp->wqres, qp->buf_size);
	vfree(rq->info);
	rq->info = NULL;
	vfree(sq->info);
	sq->info = NULL;
	mlx4_unregister_mac(mfcdev->dev, priv->port, rfci->fc_mac_idx);
	rfci->fc_mac_idx = -1;
	
	FCPRINT("RFCI of port %d destroyed\n", priv->port);
}

int mlx4_fc_init_rfci(struct mlx4_fc_priv *priv, struct mlx4_fc_rfci *rfci)
{
	struct mlx4_fc_dev *mfcdev = priv->mfcdev;
	struct mlx4_fc_qp *qp = &rfci->fc_qp;
	enum mlx4_qp_state qp_state;
	int err = 0;
	
	struct mlx4_qp_context context = {
		.flags 		= cpu_to_be32(QPC_SERVICE_TYPE_RFCI << 16),
		.pd 		= cpu_to_be32(mfcdev->priv_pdn),
		.mtu_msgmax	= 0x77, /* Raw-ETH requirement */
		.sq_size_stride = ilog2(RFCI_SQ_NUM_BBS) << 3 | ilog2(RFCI_SQ_BB_SIZE >> 4),
		.rq_size_stride	= ilog2(RFCI_RQ_NUM_WQES) << 3 | ilog2(RFCI_RQ_WQE_SIZE >>4),
		.usr_page 	= cpu_to_be32(mfcdev->priv_uar.index),
		.local_qpn 	= cpu_to_be32(qp->mqp.qpn),
		.pri_path.sched_queue = 0x83 | (priv->port - 1) << 6,
		.pri_path.counter_index = 0xff,
		.pri_path.ackto = MLX4_LINK_TYPE_ETH,
		.params2	= cpu_to_be32((qp->wqres.buf.u.direct.map & (PAGE_SIZE-1)) & 0xfc0),
		.cqn_send	= cpu_to_be32(rfci->fc_cq.mcq.cqn),
		.cqn_recv	= cpu_to_be32(rfci->fc_cq.mcq.cqn),
		.db_rec_addr 	= cpu_to_be64(qp->wqres.db.dma), /* we can assume that db.dma is aligned */
		.srqn		= 0,
	};

	err = mlx4_qp_to_ready(mfcdev->dev, &qp->wqres.mtt, &context, &qp->mqp, &qp_state);
	if (qp_state != MLX4_QP_STATE_RTS) {
		mlx4_err(mfcdev->dev, "Error bringing RFCI QP to RTS state\n");
		return err;
	}
	return 0;
}

int mlx4_do_rfci_xmit(void *drv_priv, int channel, struct sk_buff *skb, u8 fceof)
{
	struct mlx4_fc_priv 	*priv = drv_priv;
	struct mlx4_fc_rfci 	*rfci = &priv->rfci[channel];
	struct mlx4_fc_dev 	*mfcdev = priv->mfcdev;
	struct mlx4_dev 	*mdev = mfcdev->dev;
	struct mlx4_fc_q 	*sq = &rfci->fc_qp.sq;

	struct mlx4_fc_rfci_tx_desc 	*tx_desc;
	struct mlx4_fc_data_seg *data;
	
	unsigned int desc_size;
	unsigned int pad_size;
	dma_addr_t dma;
	u32 index, prod;
	__be32 op_own;
	u32 flags;

	desc_size = CTRL_SIZE + DS_SIZE;
	pad_size = RFCI_SQ_BB_SIZE - desc_size;


	spin_lock_irqsave(&rfci->lock, flags);
	if (unlikely((u32) (sq->prod - sq->cons - 1) > sq->size - 2)) {
		mlx4_info(mdev, "rfci_xmit: Out of send queue BBs\n");
		spin_unlock_irqrestore(&rfci->lock, flags);
		return 0;
	}
	prod = sq->prod;
	++sq->prod;
	spin_unlock_irqrestore(&rfci->lock, flags);

	index = prod & sq->size_mask;
	sq->info[index].buf = skb;

	tx_desc = sq->buf + index * RFCI_SQ_BB_SIZE;

	tx_desc->ctrl.size = cpu_to_be16((desc_size / 16) & 0x3f);
	if (mlx4_fc_vlan_id != -1) {
		tx_desc->ctrl.size |= cpu_to_be16(MLX4_FC_BIT_INS_VLAN);
		tx_desc->ctrl.vlan = cpu_to_be16(mlx4_fc_vlan_id);
	}
	
	tx_desc->ctrl.flags = cpu_to_be32(MLX4_FC_BIT_TX_COMP | 
					  MLX4_FC_BIT_NO_ICRC | 
					  MLX4_FC_BIT_TX_FCRC_CS);

	op_own = cpu_to_be32(MLX4_FC_RFCI_OP_SEND) | cpu_to_be32((u32)fceof << 16) | 
		((prod & sq->size) ? cpu_to_be32(MLX4_FC_BIT_DESC_OWN) : 0);
	data = &tx_desc->data;

	dma = pci_map_single(mfcdev->dev->pdev, skb->data, skb->len - 5, PCI_DMA_TODEVICE);
	data->addr = cpu_to_be64(dma);
	data->count = cpu_to_be32(skb->len - 5);
	data->mem_type = cpu_to_be32(mfcdev->mr.key); /* always snoop */

	/* Ensure new descirptor (and ownership of next descirptor) hits memory
	 * before setting ownership of this descriptor to HW */
	wmb();
	tx_desc->ctrl.op_own = op_own;

	/* Ring doorbell! */
	wmb();
	writel(rfci->fc_qp.doorbell_qpn, mfcdev->uar_map + MLX4_SEND_DOORBELL);

	return 0;
}

int mlx4_start_rfci_data(struct mlx4_fc_priv *priv, u64 mac)
{
	int err = 0;
	
	/*
	 * Remove any previously-set unicast MAC filter.
	 * Add secondary FCoE MAC address filter for our OUI.
	 */
	if (priv->rfci[RFCI_DATA].fc_mac_idx != -1) {
		FCPRINT("Destroying old data RFCI\n");
		mlx4_fc_destroy_rfci(priv, &priv->rfci[RFCI_DATA]);
	}
	
	err = mlx4_fc_create_rfci(priv, &priv->rfci[RFCI_DATA], mac);
	if (err) {
		mlx4_err(priv->mfcdev->dev, "Could not create data RFCI QP, err=%d\n", err);
		return err;
	}
	FCPRINT("Created RFCI for data MAC\n");
	err = mlx4_fc_init_rfci(priv, &priv->rfci[RFCI_DATA]);
	if (err) {
		mlx4_err(priv->mfcdev->dev, "Could not init data RFCI QP, err=%d\n", err);
		return err;
	}
	FCPRINT("RFCI in state RTS\n");
	return err;
}


