/*
 * Copyright(c) 2007 Intel Corporation. All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 * Maintained at www.Open-FCoE.org
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/spinlock.h>
#include <linux/scatterlist.h>
#include <scsi/scsi_tcq.h>
#include <scsi/scsi.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport.h>
#include <scsi/scsi_transport_fc.h>

/*
 * non linux dot h files
 */
#include "fc_types.h"
#include "sa_assert.h"
#include "fc_event.h"
#include "fc_port.h"
#include "fc_remote_port.h"
#include "fcoeioctl.h"
#include "fcdev.h"
#include "fc_fc2.h"
#include "fc_frame.h"
#include "fc_exch.h"
#include "fc_sess.h"
#include "fc_event.h"
#include "ofc_dbg.h"
#include "openfc.h"
#include "openfc_scsi_pkt.h"
#include "crc32_le.h"

int openfc_debug;

/*
 * function prototypes
 * FC scsi I/O related functions
 */
static void openfc_scsi_recv_data(struct fc_scsi_pkt *, struct fc_frame *);
static void openfc_scsi_rcv(struct fc_seq *, struct fc_frame *, void *);
static void openfc_tm_done(struct fc_seq *, struct fc_frame *, void *);
static void openfc_scsi_error(enum fc_event, void *);
static int openfc_abort_internal(struct fcdev *, struct fc_scsi_pkt *,
				 struct fc_frame *);
void openfc_scsi_cleanup(struct fc_scsi_pkt *);
static void openfc_timeout_error(struct fc_scsi_pkt *);
void openfc_scsi_rec_rcv(struct fc_seq *, struct fc_frame *, void *);
void openfc_scsi_rec(struct fc_scsi_pkt *);

/*
 * task managment timeout value
 */
#define FCOE_TM_TIMEOUT	     10000	/* mSec */
#define OPENFC_MAX_ERROR_CNT  5

/*
 * End a request with a retry suggestion.
 */
static void openfc_scsi_retry(struct fc_scsi_pkt *fsp)
{
	struct openfc_softc *openfcp = fsp->openfcp;

	fc_seq_exch_complete(fsp->seq_ptr);
	fsp->seq_ptr = NULL;
	fsp->state = OPENFC_SRB_RCV_STATUS;
	fsp->status_code = OPENFC_ERROR;
	fsp->io_status = SUGGEST_RETRY << 24;
	openfcp->outstandingcmd[fsp->idx].ptr = NULL;
	if (fsp->done)
		(*fsp->done) (fsp);
}

/*
 * Receive SCSI data from target.
 * Called after receiving solicited data.
 */
static void openfc_scsi_recv_data(struct fc_scsi_pkt *fsp, struct fc_frame *fp)
{
	struct scsi_cmnd *sc = fsp->cmd;
	struct openfc_softc *openfcp = fsp->openfcp;
	struct fcoe_dev_stats *sp;
	struct fc_frame_header *fh;
	size_t offset;
	u32 crc;
	u32 copy_len = 0;
	size_t len;
	void *buf;

	if (!sc->request_buffer)
		return;			/* XXX possible? */
	fh = fc_frame_header_get(fp);
	offset = net32_get(&fh->fh_parm_offset);
	len = fp->fr_len - sizeof(*fh);
	buf = fc_frame_payload_get(fp, 0);

	if (offset + len > fsp->data_len) {
		/*
		 * this should never happen
		 */
		if ((fp->fr_flags & FCPHF_CRC_UNCHECKED) &&
		    fc_frame_crc_check(fp))
			goto crc_err;
		if (openfc_debug) {
			OFC_DBG("data received past end.	 "
			       "len %zx offset %zx "
			       "data_len %x\n", len, offset, fsp->data_len);
		}
		openfc_scsi_retry(fsp);
		return;
	}
	crc = 0;
	if (sc->use_sg) {
		struct scatterlist *sg;
		struct scatterlist *sg_limit;
		size_t remaining, sg_bytes;
		size_t off;
		void *page_addr;

		if (fp->fr_flags & FCPHF_CRC_UNCHECKED)
			crc = crc32_sb8_64_bit(~0, (u8 *) fh, sizeof(*fh));

		sg = (struct scatterlist *)sc->request_buffer;
		sg_limit = sg + sc->use_sg;
		remaining = len;

		while (remaining > 0 && sg < sg_limit) {
			if (offset >= sg->length) {
				offset -= sg->length;
				sg++;
				continue;
			}
			sg_bytes = min(remaining, sg->length - offset);

			/*
			 * The scatterlist item may be bigger than PAGE_SIZE,
			 * but we are limited to mapping PAGE_SIZE at a time.
			 */
			off = offset + sg->offset;
			sg_bytes = min(sg_bytes,
				       (PAGE_SIZE - (off & ~PAGE_MASK)));
			page_addr = kmap_atomic(sg_page(sg) +
						(off >> PAGE_SHIFT),
						KM_SOFTIRQ0);
			if (!page_addr)
				break;		/* XXX panic? */

			if (fsp->state != OPENFC_SRB_ABORT_PENDING) {
				if (fp->fr_flags & FCPHF_CRC_UNCHECKED) {
					crc = crc32_copy(crc,
							 (char *)page_addr +
							 (off & ~PAGE_MASK),
							 buf, sg_bytes);
				} else {
					__memcpy((char *)page_addr +
						 (off & ~PAGE_MASK),
						 buf, sg_bytes);
				}
			}
			kunmap_atomic(page_addr, KM_SOFTIRQ0);
			buf += sg_bytes;
			offset += sg_bytes;
			remaining -= sg_bytes;
			copy_len += sg_bytes;
		}
		if (fp->fr_flags & FCPHF_CRC_UNCHECKED)
			goto crc_check;
	} else if (fp->fr_flags & FCPHF_CRC_UNCHECKED) {
		crc = crc32_sb8_64_bit(~0, (u8 *)fh, sizeof(*fh));
		crc = crc32_copy(crc, (void *)sc->request_buffer + offset,
				 buf, len);
		copy_len = len;
crc_check:
		buf = fc_frame_payload_get(fp, 0);
		if (len % 4) {
			crc = crc32_sb8_64_bit(crc, buf + len, 4 - (len % 4));
			len += 4 - (len % 4);
		}
		if (~crc != le32_to_cpu(*(__le32 *)(buf + len))) {
crc_err:
			sp = openfcp->fd.dev_stats[smp_processor_id()];
			sp->ErrorFrames++;
			if (sp->InvalidCRCCount++ < 5)
				OFC_DBG("CRC error on data frame");
			/*
			 * Assume the frame is total garbage.
			 * We may have copied it over the good part
			 * of the buffer.
			 * If so, we need to retry the entire operation.
			 * Otherwise, ignore it.
			 */
			if (offset < fsp->xfer_len)
				openfc_scsi_retry(fsp);
			return;
		}
	} else {
		__memcpy((void *)sc->request_buffer + offset, buf, len);
		copy_len = len;
	}
	fsp->xfer_len += copy_len;
	if (fsp->xfer_len == fsp->data_len)
		fsp->state = OPENFC_SRB_WAIT_FOR_STATUS;
}

/*
 * Send SCSI data to target.
 * Called after receiving a Transfer Ready data descriptor.
 */
static void openfc_scsi_send_data(struct fc_scsi_pkt *fsp, struct fc_seq *sp,
				  size_t offset, size_t len,
				  struct fc_frame *oldfp, int sg_supp)
{
	struct scsi_cmnd *sc;
	struct scatterlist *fsg;
	struct scatterlist *sg;
	u_int nsg;
	struct scatterlist local_sg;
	struct fc_frame *fp = NULL;
	struct openfc_softc *openfcp = fsp->openfcp;
	size_t remaining;
	size_t mfs;
	size_t tlen;
	size_t sg_bytes;
	size_t frame_offset;
	int error;
	void *data = NULL;
	void *page_addr;
	int using_sg = sg_supp;

	if (unlikely(offset + len > fsp->data_len)) {
		/*
		 * this should not never happen
		 */
		if (openfc_debug) {
			OFC_DBG("xfer-ready past end. len %zx offset %zx\n",
			       len, offset);
		}
		openfcp->outstandingcmd[fsp->idx].ptr = NULL;
		openfc_abort_internal(openfcp->dev, fsp, oldfp);
		return;
	} else if (offset != fsp->xfer_len) {
		/*
		 *  Out of Order Data Request: We will complete the exchange
		 * and send an error notification to SCSI
		 */
		if (openfc_debug) {
			OFC_DBG("xfer-ready non-contiguous. "
			       "len %zx offset %zx\n", len, offset);
		}
	}
	mfs = fc_seq_mfs(sp);
	ASSERT(mfs >= FC_MIN_MAX_PAYLOAD);
	ASSERT(mfs <= FC_MAX_PAYLOAD);
	if (mfs > 512)
		mfs &= ~(512 - 1);	/* round down to block size */
	ASSERT(mfs >= FC_MIN_MAX_PAYLOAD);	/* won't go below 256 */
	ASSERT(len > 0);
	sc = fsp->cmd;
	if (sc->use_sg) {
		sg = (struct scatterlist *)sc->request_buffer;
		nsg = sc->use_sg;
	} else {
		/* this should not happen */
		sg_init_one(&local_sg, sc->request_buffer, sc->request_bufflen);
		sg = &local_sg;
		nsg = 1;
	}
	remaining = len;
	frame_offset = offset;
	tlen = 0;
	sp = fc_seq_start_next_fctl(sp, FC_FC_REL_OFF);
	ASSERT(sp);

	if (page_count(sg_page(sg)) == 0)
		using_sg = 0;

	while (remaining > 0 && nsg > 0) {
		if (offset >= sg->length) {
			offset -= sg->length;
			sg = sg_next(sg);
			nsg--;
			continue;
		}
		if (!fp) {
			tlen = min(mfs, remaining);
			if (using_sg) {
				fp = fc_port_frame_alloc(openfcp->fcs_port, 0);
			} else {
				fp = fc_port_frame_alloc(openfcp->fcs_port,
							 tlen);
				data = (void *)(fp->fr_hdr + 1);
			}
			ASSERT_NOTIMPL(fp != NULL);	/* XXX */
			fc_frame_setup(fp, FC_RCTL_DD_SOL_DATA, FC_TYPE_FCP);
			fc_frame_set_offset(fp, frame_offset);
		}
		sg_bytes = min(tlen, sg->length - offset);
		if (using_sg) {
			/*
			 * do not need to do get page
			 * because we can unpin the memory when the
			 * status of this I/O is received
			 */
			ASSERT(fp->fr_sg_len < FC_FRAME_SG_LEN);
			fsg = &fp->fr_sg[fp->fr_sg_len++];
			sg_set_page(fsg, sg_page(sg),
				    sg_bytes, sg->offset + offset);
		} else {
			u_int off = offset + sg->offset;

			/*
			 * The scatterlist item may be bigger than PAGE_SIZE,
			 * but we must not cross pages inside the kmap.
			 */
			sg_bytes = min(sg_bytes, (size_t) (PAGE_SIZE -
							   (off & ~PAGE_MASK)));
			page_addr = kmap_atomic(sg_page(sg) + (off >> PAGE_SHIFT),
						KM_SOFTIRQ0);
			__memcpy(data, (char *)page_addr + (off & ~PAGE_MASK),
				 sg_bytes);
			kunmap_atomic(page_addr, KM_SOFTIRQ0);
			data += sg_bytes;
		}
		fp->fr_len += sg_bytes;
		offset += sg_bytes;
		frame_offset += sg_bytes;
		tlen -= sg_bytes;
		remaining -= sg_bytes;

		if (remaining == 0) {
			error = fc_seq_send_tsi(sp, fp);
		} else if (tlen == 0) {
			error = fc_seq_send_frag(sp, fp);
		} else {
			continue;
		}
		fp = NULL;

		ASSERT(!error);
		if (error) {
			/*
			 * we need to handle this case -XXX
			 */
			openfc_scsi_retry(fsp);
			return;
		}
	}
	fsp->xfer_len += len;	/* premature count? */

	if (fsp->xfer_len == fsp->data_len)
		fsp->state = OPENFC_SRB_WAIT_FOR_STATUS;
}

/*
 * fcs exch mgr calls this routine to process scsi
 * exchanges.
 *
 * Return   : None
 * Context  : called from Soft IRQ context
 *	      can not called holding list lock
 */

static void openfc_scsi_rcv(struct fc_seq *sp, struct fc_frame *fp, void *arg)
{
	struct fc_scsi_pkt *fsp = (struct fc_scsi_pkt *)arg;
	struct openfc_cmdq *qp;
	struct openfc_softc *openfcp;
	struct fc_frame_header *fh;
	struct os_lun *disk;
	struct fc_data_desc *dd;
	struct fcp_resp *fc_rp;
	struct fcp_resp_ext *rp_ex;
	struct fcp_resp_rsp_info *fc_rp_info;
	u_int r_ctl;
	u_int32_t respl = 0;
	u_int32_t snsl = 0;
	int idx;
	u_int8_t cdb_op;
	u_int8_t sg_supp;
	enum dma_data_direction data_dir;

	fh = fc_frame_header_get(fp);
	r_ctl = fh->fh_r_ctl;
	openfcp = fsp->openfcp;

	sg_supp = (openfcp->dev->capabilities & TRANS_C_SG);
	if (!(openfcp->state & OPENFC_RUNNING))
		goto out;

	idx = fsp->idx;
	qp = &openfcp->outstandingcmd[idx];
	spin_lock(&qp->scsi_pkt_lock);

	/*
	 * make sure slot has the same srb pointer
	 * this should be always -- paranoia
	 */
	if (unlikely(qp->ptr != fsp)) {
		spin_unlock(&qp->scsi_pkt_lock);
		goto out;
	}

	/*
	 * status code could be set if we had an target reset
	 * otherwise it should be zero.
	 */
	if (unlikely(fsp->status_code)) {
		spin_unlock(&qp->scsi_pkt_lock);
		goto out;
	}
	fsp->last_pkt_time = jiffies;

	disk = fsp->disk;
	if (disk && disk->error_cnt) {
		disk->state = OPENFC_LUN_READY;
		disk->error_cnt = 0;
	}

	if (r_ctl == FC_RCTL_DD_DATA_DESC) {
		/*
		 * received XFER RDY from the target
		 * need to send data to the target
		 */
		ASSERT(!(fp->fr_flags & FCPHF_CRC_UNCHECKED));
		dd = fc_frame_payload_get(fp, sizeof(*dd));
		ASSERT_NOTIMPL(dd != NULL);
		fsp->state = OPENFC_SRB_IN_DATA_TRANS;

		/*
		 * if the I/O request does not have sg list then
		 * copy user data into packet
		 */
		if (fsp->cmd->use_sg == 0)
			sg_supp = 0;

		openfc_scsi_send_data(fsp, sp,
				      (size_t) net32_get(&dd->dd_offset),
				      (size_t) net32_get(&dd->dd_len), fp,
				      sg_supp);
		/*
		 * it here is an error on transfer then
		 * the srb will be freed , when we unlock slot
		 * could be reused by openfc_scsi_send().
		 */
		spin_unlock(&qp->scsi_pkt_lock);
	} else if (r_ctl == FC_RCTL_DD_SOL_DATA) {
		/*
		 * received a DATA frame
		 * next we will copy the data to the system buffer
		 */
		ASSERT(fp->fr_len >= sizeof(*fh));	/* len may be 0 */
		fsp->state = OPENFC_SRB_IN_DATA_TRANS;
		openfc_scsi_recv_data(fsp, fp);

		/*
		 * if there is an error on transfer then
		 * the srb will be freed when we unlock slot
		 * could be reused by openfc_scsi_send().
		 */
		spin_unlock(&qp->scsi_pkt_lock);
	} else if (r_ctl == FC_RCTL_DD_CMD_STATUS) {
		ASSERT(!(fp->fr_flags & FCPHF_CRC_UNCHECKED));

		/*
		 * NULL out the outstanding cmd pointer.
		 */
		qp->ptr = NULL;
		spin_unlock(&qp->scsi_pkt_lock);

		/*
		 * received the SCSI status
		 * if the byte 10 & 11 is zero
		 * then send the completion notification to the upper
		 * layer, otherwise if sense present then
		 * copy the sense data to the scsi sense buffer
		 */
		fc_rp = fc_frame_payload_get(fp, sizeof(*fc_rp));
		fsp->seq_ptr = NULL;	/* released below with complete */
		fsp->cdb_status = fc_rp->fr_status;
		fsp->scsi_comp_flags = fc_rp->fr_flags;
		fsp->status_code = OPENFC_COMPLETE;
		if ((fc_rp->fr_flags == 0) && (fc_rp->fr_status == 0)) {
			data_dir = fsp->cmd->sc_data_direction;
			cdb_op = fsp->cmd->cmnd[0];
			if (data_dir == DMA_FROM_DEVICE) {
				if (fsp->data_len > fsp->xfer_len) {
					if ((cdb_op != READ_10 ||
					     cdb_op != READ_6 ||
					     cdb_op != WRITE_10 ||
					     cdb_op != WRITE_6)) {
						fsp->status_code = OPENFC_ERROR;
						fsp->io_status =
						    (SUGGEST_RETRY << 24);
					} else {
						OFC_DBG("xfer len less then "
						       "data_len xfer-len %zx"
						       " len %x \n",
						       fsp->xfer_len,
						       fsp->data_len);
						fsp->status_code =
						    OPENFC_DATA_UNDRUN;
					}
				} else if (fsp->data_len < fsp->xfer_len) {
					if ((cdb_op != READ_10 ||
					     cdb_op != READ_6 ||
					     cdb_op != WRITE_10 ||
					     cdb_op != WRITE_6)) {
						fsp->status_code = OPENFC_ERROR;
						fsp->io_status =
						    (SUGGEST_RETRY << 24);
					} else {
						OFC_DBG("xfer len greater than "
						       "data_len xfer-len %zx"
						       " len %x \n",
						       fsp->xfer_len,
						       fsp->data_len);
						fsp->status_code =
						    OPENFC_DATA_OVRRUN;
					}
				}
			}
		} else {
			rp_ex = (void *)(fc_rp + 1);
			fc_rp_info = (struct fcp_resp_rsp_info *)(rp_ex + 1);
			if (fc_rp->fr_flags & FCP_RSP_LEN_VAL) {
				respl = net32_get(&rp_ex->fr_rsp_len);
				if ((respl != 0 && respl != 4 && respl != 8) ||
				    (fc_rp_info->rsp_code != FCP_TMF_CMPL)) {
					fsp->status_code = OPENFC_ERROR;
				}
			}
			if (fc_rp->fr_flags & FCP_SNS_LEN_VAL) {
				snsl = net32_get(&rp_ex->fr_sns_len);
				if (snsl > SCSI_SENSE_BUFFERSIZE)
					snsl = SCSI_SENSE_BUFFERSIZE;
				memcpy(fsp->cmd->sense_buffer,
				       &fc_rp_info->_fr_resvd[0] + respl, snsl);
			}
			if (fc_rp->fr_flags & FCP_RESID_UNDER) {
				fsp->scsi_resid = net32_get(&rp_ex->fr_resid);
				/*
				 * The cmnd->underflow is the minimum number of
				 * bytes that must be transfered for this
				 * command.  Provided a sense condition is not
				 * present, make sure the actual amount
				 * transferred is at least the underflow value
				 * or fail.
				 */
				if (!(fc_rp->fr_flags & FCP_SNS_LEN_VAL) &&
				    (fc_rp->fr_status == 0) &&
				    (fsp->cmd->request_bufflen -
				     fsp->scsi_resid) < fsp->cmd->underflow) {
					fsp->status_code = OPENFC_ERROR;
				}
			} else if (fc_rp->fr_flags & FCP_RESID_OVER) {
				fsp->status_code = OPENFC_ERROR;
			}
		}

		openfcp = fsp->openfcp;
		fc_seq_exch_complete(sp);
		fsp->state = OPENFC_SRB_RCV_STATUS;
		(*fsp->done) (fsp);
	} else {
		spin_unlock(&qp->scsi_pkt_lock);
	}
out:
	fc_frame_free(fp);
}

void openfc_scsi_abort_iocontext(struct fc_scsi_pkt *fsp)
{
	struct fc_seq *sp = fsp->seq_ptr;

	if (fsp->state != OPENFC_SRB_RCV_STATUS) {
		fsp->seq_ptr = NULL;
		if (sp)
			fc_seq_exch_complete(sp);

		fsp->state = OPENFC_SRB_RCV_STATUS;
		fsp->status_code = OPENFC_ERROR;
		fsp->io_status = (SUGGEST_RETRY << 24);
		(*fsp->done) (fsp);
	}
}

/*
 * called by upper layer protocol
 *
 * Return   : zero for success and -1 for failure
 * Context  : called from process context or timer context
 *	      can not called holding list lock
 */
int openfc_scsi_send(struct fcdev *dev, struct fc_scsi_pkt *fsp)
{
	struct openfc_softc *openfcp = openfc_get_softc(dev);
	u_int32_t idx, i;
	struct fc_frame *fp;
	struct fc_seq *sp;
	unsigned long flags;
	int rc = 0;
	void *p;

	/*
	 * Acquire the list lock
	 */
	spin_lock_irqsave(&openfcp->outstandingcmd_lock, flags);
	i = openfcp->current_cmd_indx;
	/*
	 * looking for an empty slot in outstandingcmd array
	 */
	for (idx = 1; idx < OPENFC_MAX_OUTSTANDING_COMMANDS; idx++) {
		i++;
		if (i >= OPENFC_MAX_OUTSTANDING_COMMANDS)
			i = 1;

		/*
		 * looking for an empty slot
		 * first check if the pointer is zero
		 * if it is zero then try to get the lock
		 * and if we get the lock then it is an emty slot
		 * otherwise it is in processo of getting empty
		 */
		if (openfcp->outstandingcmd[i].ptr == NULL) {
			if (spin_trylock
			    (&openfcp->outstandingcmd[i].scsi_pkt_lock))
				break;
		}
	}
	if (idx == OPENFC_MAX_OUTSTANDING_COMMANDS) {
		OFC_DBG("outstanding cmd array is full");
		spin_unlock_irqrestore(&openfcp->outstandingcmd_lock, flags);

		/*
		 * here we return to the upper layer with an error
		 * upper layer will retry the i/o later
		 */
		rc = -1;
		goto out;
	}

	/*
	 * at this time we have the slot index and the lock to
	 * the slot.
	 */
	openfcp->current_cmd_indx = i;
	openfcp->outstandingcmd[i].ptr = (void *)fsp;
	spin_unlock(&openfcp->outstandingcmd[i].scsi_pkt_lock);
	fsp->state = OPENFC_SRB_CMD_SENT;
	spin_unlock_irqrestore(&openfcp->outstandingcmd_lock, flags);
	p = xchg(&((CMD_SP(fsp->cmd))), &openfcp->outstandingcmd[i]);
	fsp->idx = i;
	net32_put(&fsp->cdb_cmd.fc_dl, fsp->data_len);
	fsp->cdb_cmd.fc_flags = fsp->req_flags & ~FCP_CFL_LEN_MASK;

	net64_put((net64_t *) fsp->cdb_cmd.fc_lun, (fsp->lun << 48));
	memcpy(fsp->cdb_cmd.fc_cdb, fsp->cmd->cmnd, fsp->cmd->cmd_len);
	sp = fc_sess_seq_alloc(fsp->rp->rp_sess,
			       openfc_scsi_rcv, openfc_scsi_error, (void *)fsp);
	if (sp == NULL) {
		rc = -1;
		goto out;
	} else {
		fp = fc_frame_alloc(openfcp->fcs_port, sizeof(fsp->cdb_cmd));
		if (fp) {
			fc_seq_hold(sp);
			fsp->seq_ptr = sp;
			memcpy(fc_frame_payload_get(fp, sizeof(fsp->cdb_cmd)),
			       &fsp->cdb_cmd, sizeof(fsp->cdb_cmd));
			fc_seq_send_req(sp, fp, FC_RCTL_DD_UNSOL_CMD,
					FC_TYPE_FCP, 0);
		} else {
			openfcp->outstandingcmd[i].ptr = (void *)NULL;
			OFC_DBG("could not allocate frame");
			fc_seq_exch_complete(sp);
			rc = -1;
			goto out;
		}
	}
out:
	return rc;
}

/*
 * transport error handler
 */
static void openfc_scsi_error(enum fc_event event, void *fsp_arg)
{
	struct fc_scsi_pkt *fsp = (struct fc_scsi_pkt *)fsp_arg;
	struct fc_scsi_pkt *ofsp;
	struct openfc_softc *openfcp = fsp->openfcp;
	struct openfc_cmdq *qp = NULL;
	struct os_lun *diskp;

	if (openfcp->state == OPENFC_GOING_DOWN)
		return;

	/*
	 * scsi pkt which are used for REC are not stored in the queue
	 */
	if (fsp->idx) {
		qp = &openfcp->outstandingcmd[fsp->idx];
		spin_lock(&qp->scsi_pkt_lock);
		if (qp->ptr == fsp)
			qp->ptr = NULL;
		spin_unlock(&qp->scsi_pkt_lock);
	}
	diskp = fsp->disk;
	ASSERT(fsp != NULL);
	switch (event) {
	case FC_EV_TIMEOUT:
		/*
		 * not sure about this case
		 * But if we get here then abort the I/O
		 */
		OFC_DBG(" exchange level timeout error");
		if (fsp->state == OPENFC_SRB_REC) {
			ofsp = fsp->old_pkt;
			diskp->tgtp->flags &= OPENFC_TGT_REC_NOT_SUPPORTED;
			openfc_scsi_pkt_release(ofsp);
			qp = &openfcp->outstandingcmd[ofsp->idx];
			spin_lock(&qp->scsi_pkt_lock);
			if (qp->ptr == ofsp)
				qp->ptr = NULL;
			spin_unlock(&qp->scsi_pkt_lock);
			openfc_timeout_error(ofsp);
			fsp->state = OPENFC_SRB_FREE;
			openfc_free_scsi_pkt(fsp);
		} else if (fsp->state == OPENFC_SRB_ABORT_PENDING) {
			fsp->state = OPENFC_SRB_ABORTED;
			if (fsp->wait_for_comp)
				complete(&fsp->tm_done);
		} else {
			/*
			 * this should not happen
			 */
			if (fsp->done) {
				fsp->status_code = OPENFC_CMD_ABORTED;
				(*fsp->done) (fsp);
			}
		}
		break;
	case FC_EV_CLOSED:
		/*
		 * if we received this we may want to
		 * cleanup the fc_transport --XXX
		 */
		OFC_DBG(" FC level close error");
		if (fsp->done) {
			fsp->status_code = OPENFC_CMD_PLOGO;
			(*fsp->done) (fsp);
		}
		break;
	case FC_EV_ACC:
		/*
		 * this is abort accepted case
		 */
		OFC_DBG(" BLS Acc event");
		if (fsp->state == OPENFC_SRB_ABORT_PENDING) {
			fsp->state = OPENFC_SRB_ABORTED;
			if (fsp->wait_for_comp)
				complete(&fsp->tm_done);
		}
		break;

	case FC_EV_RJT:
		/*
		 * this is abort rejected case
		 */
		OFC_DBG(" BLS reject event");
		if (fsp->state == OPENFC_SRB_ABORT_PENDING) {
			if (fsp->wait_for_comp)
				complete(&fsp->tm_done);
			break;
		}
		if (fsp->state == OPENFC_SRB_REC) {
			ofsp = fsp->old_pkt;
			diskp->tgtp->flags &= OPENFC_TGT_REC_NOT_SUPPORTED;
			openfc_timeout_error(ofsp);
			fsp->state = OPENFC_SRB_FREE;
			openfc_free_scsi_pkt(fsp);
			break;
		}
		break;
	default:
		OFC_DBG(" unknown event");
		if (fsp->done) {
			fsp->status_code = OPENFC_CMD_PLOGO;
			(*fsp->done) (fsp);
		}
		break;
	}
}

static int openfc_abort_internal(struct fcdev *dev, struct fc_scsi_pkt *fsp,
				 struct fc_frame *fp)
{
	int rc = -1;

	fsp->state = OPENFC_SRB_ABORT_PENDING;
	rc = openfc_abort_cmd(dev, fsp);
	if (rc == SUCCESS)
		rc = 0;
	else
		rc = -1;
	fsp->status_code = OPENFC_CMD_ABORTED;
	fsp->cdb_status = -1;
	if (fsp->done)
		(*fsp->done) (fsp);
	return rc;

}

/*
 * Scsi abort handler- calls fcs to send an abort
 * and then wait for abort completion
 */
int openfc_abort_cmd(struct fcdev *dev, struct fc_scsi_pkt *fsp)
{
	int rc = FAILED;

	if (!fsp->seq_ptr)
		return rc;
	if (fc_seq_abort_exch(fsp->seq_ptr))
		return rc;

	fsp->idx = 0;
	fsp->state = OPENFC_SRB_ABORT_PENDING;

	init_completion(&fsp->tm_done);
	fsp->wait_for_comp = 1;
	if (!wait_for_completion_timeout(&fsp->tm_done,
					 msecs_to_jiffies(FCOE_TM_TIMEOUT))) {
		OFC_DBG("target abort cmd  failed");
		rc = FAILED;
	} else if (fsp->state == OPENFC_SRB_ABORTED) {
		OFC_DBG("target abort cmd  passed");
		rc = SUCCESS;
	}
	if (fsp->seq_ptr) {
		fc_seq_exch_complete(fsp->seq_ptr);
		fsp->seq_ptr = NULL;
	}
	return rc;
}

/*
 * Scsi target reset handler- send a LUN RESET to the target
 * and wait for reset reply
 */
int openfc_target_reset(struct fcdev *dev, struct fc_scsi_pkt *fsp)
{
	struct openfc_softc *openfcp = openfc_get_softc(dev);
	unsigned long flags;
	u_int32_t idx;
	struct fc_frame *fp;
	struct fc_seq *sp;
	struct fc_scsi_pkt *ofsp;
	struct scsi_cmnd *sc_cmd;
	unsigned int pid, opid, tid, otid, lun, olun;

	fsp->idx = 0;
	net32_put(&fsp->cdb_cmd.fc_dl, fsp->data_len);
	fsp->cdb_cmd.fc_tm_flags = FCP_TMF_LUN_RESET;
	net64_put((net64_t *) fsp->cdb_cmd.fc_lun, (fsp->lun << 48));
	sp = fc_sess_seq_alloc(fsp->rp->rp_sess, openfc_tm_done,
			       openfc_scsi_error, (void *)fsp);
	if (sp) {
		fc_seq_hold(sp);
		fp = fc_frame_alloc(openfcp->fcs_port, sizeof(fsp->cdb_cmd));
		if (fp == NULL) {
			OFC_DBG("could not allocate frame");
			fc_seq_exch_complete(sp);
			return FAILED;
		}
		init_completion(&fsp->tm_done);
		memcpy(fc_frame_payload_get(fp, sizeof(fsp->cdb_cmd)),
		       &fsp->cdb_cmd, sizeof(fsp->cdb_cmd));
		fsp->state = 0xf00d;
		fc_seq_send_req(sp, fp, FC_RCTL_DD_UNSOL_CMD, FC_TYPE_FCP, 0);
		/*
		 * wait for completion of reset
		 * after that make sure all commands are terminated
		 */
		fsp->wait_for_comp = 1;
		if (!wait_for_completion_timeout(&fsp->tm_done,
						 msecs_to_jiffies
						 (FCOE_TM_TIMEOUT))) {
			fc_seq_exch_complete(sp);
			OFC_DBG("target reset failed");
			return FAILED;
		}
	}

	/*
	 * Acquire the list lock
	 */
	pid = 0;
	tid = fsp->id;
	lun = fsp->lun;
	spin_lock_irqsave(&openfcp->outstandingcmd_lock, flags);
	for (idx = 1; idx < OPENFC_MAX_OUTSTANDING_COMMANDS; idx++) {
		spin_lock(&openfcp->outstandingcmd[idx].scsi_pkt_lock);
		ofsp = openfcp->outstandingcmd[idx].ptr;
		if (ofsp) {
			/*
			 * if the cmd is internal then no scsi cmd
			 * associated with it, so ignore that scsi pkt
			 */
			if (!ofsp->cmd) {
				openfcp->outstandingcmd[idx].ptr = NULL;
				OFC_DBG("non scsi cmd in the array %d idx\n",
				       idx);
				goto out;
			}
			sc_cmd = (struct scsi_cmnd *)ofsp->cmd;
			ASSERT(sc_cmd != NULL);
			opid = pid;
			otid = sc_cmd->device->id;
			olun = sc_cmd->device->lun;

			if ((opid == pid) && (otid == tid) && (olun == lun)) {

				/*
				 * aborted cmd
				 */
				if (ofsp->status_code)
					goto out;

				ofsp->status_code = OPENFC_CMD_ABORTED;
				ofsp->cdb_status = -1;
				if (ofsp->seq_ptr)
					fc_seq_exch_complete(ofsp->seq_ptr);
				ofsp->seq_ptr = NULL;
				if (ofsp->done)
					(*ofsp->done) (ofsp);
				openfcp->outstandingcmd[idx].ptr = NULL;
			}
		}
out:
		spin_unlock(&openfcp->outstandingcmd[idx].scsi_pkt_lock);
	}

	spin_unlock_irqrestore(&openfcp->outstandingcmd_lock, flags);
	fsp->state = OPENFC_SRB_FREE;

	return SUCCESS;
}

/*
 * this routine will assume that flogo has happened
 * end we are cleaning all the commands.
 */
int openfc_inf_reset(struct fcdev *dev, struct fc_scsi_pkt *fsp)
{
	struct openfc_softc *openfcp = openfc_get_softc(dev);
	int idx;
	unsigned long flags;
	struct fc_scsi_pkt *old_fsp;

	spin_lock_irqsave(&openfcp->outstandingcmd_lock, flags);
	for (idx = 1; idx < OPENFC_MAX_OUTSTANDING_COMMANDS; idx++) {
		spin_lock(&openfcp->outstandingcmd[idx].scsi_pkt_lock);
		old_fsp = openfcp->outstandingcmd[idx].ptr;
		if (old_fsp) {
			if (timer_pending(&old_fsp->timer))
				del_timer_sync(&old_fsp->timer);
			openfcp->outstandingcmd[idx].ptr = NULL;
			old_fsp->state = OPENFC_SRB_FREE;
			openfc_free_scsi_pkt(old_fsp);
		}
		spin_unlock(&openfcp->outstandingcmd[idx].scsi_pkt_lock);
	}
	spin_unlock_irqrestore(&openfcp->outstandingcmd_lock, flags);
	return 0;
}

/*
 * Task Managment response handler
 */
static void openfc_tm_done(struct fc_seq *sp, struct fc_frame *fp, void *arg)
{
	struct fc_scsi_pkt *fsp = (struct fc_scsi_pkt *)arg;
	struct fc_frame_header *fh;
	struct fcp_resp *fc_rp;
	u_int r_ctl;

	fh = fc_frame_header_get(fp);
	r_ctl = fh->fh_r_ctl;
	if (r_ctl == FC_RCTL_DD_CMD_STATUS) {
		fc_rp = fc_frame_payload_get(fp, sizeof(*fc_rp));
		fsp->state = OPENFC_SRB_TM_DONE;
		fsp->cdb_status = fc_rp->fr_status;
		fc_seq_exch_complete(sp);
		fc_frame_free(fp);
		if (fsp->wait_for_comp)
			complete(&fsp->tm_done);
	}
}

void openfc_scsi_cleanup(struct fc_scsi_pkt *sp)
{
	struct openfc_softc *openfcp = sp->openfcp;
	int idx;
	unsigned long flags;
	struct fc_scsi_pkt *old_fsp;

	spin_lock_irqsave(&openfcp->outstandingcmd_lock, flags);
	for (idx = 1; idx < OPENFC_MAX_OUTSTANDING_COMMANDS; idx++) {
		spin_lock(&openfcp->outstandingcmd[idx].scsi_pkt_lock);
		old_fsp = openfcp->outstandingcmd[idx].ptr;
		openfcp->outstandingcmd[idx].ptr = NULL;
		spin_unlock(&openfcp->outstandingcmd[idx].scsi_pkt_lock);

		if (old_fsp) {
			old_fsp->status_code = OPENFC_HRD_ERROR;
			if (old_fsp->done) {
				(*old_fsp->done) (old_fsp);
			} else {
				old_fsp->state = OPENFC_SRB_FREE;
				openfc_free_scsi_pkt(old_fsp);
			}
		}
	}
	spin_unlock_irqrestore(&openfcp->outstandingcmd_lock, flags);
	if (sp) {
		if (sp->wait_for_comp)
			complete(&sp->tm_done);
	}
}

/*
 * openfc_scsi_timeout: called by openfc timeout handler
 * returns OPENFC_TIMEOUT_SUCCESS or OPENFC_TIMEOUT_FAIL
 *
 * Algorithm:
 * if we get the lock on fc scsi pkt then
 * if the fc scsi pkt state is
 * FCP commend sent that means we have not received any pkt form the target
 * for last 10sec. In this case we will increment abort_nedded by 1;
 * if the fc scsi pkt state is In data transfer or waiting for status then
 * we will look at the last activity time and if it is  greater then 5sec
 * then mark the IO abort needed.
 * if the abort needed is set then abort the I/o close the exchange
 * and increment the lun error count.  if the count is greater then
 * equal to MAX error then return the IO with NO connect.
 */
int openfc_scsi_timeout(struct fcdev *dev, struct fc_scsi_pkt *fsp)
{
	int send_abort = 0;
	int idx, send_rec = 0;
	int unlock = 0;
	int rc = OPENFC_TIMEOUT_FAILD;
	struct os_lun *disk;
	struct openfc_softc *openfcp = fsp->openfcp;
	u_int8_t cdb_op;

	cdb_op = fsp->cmd->cmnd[0];

	if (spin_trylock(&openfcp->outstandingcmd[fsp->idx].scsi_pkt_lock)) {
		disk = fsp->disk;
		idx = fsp->idx;
		switch (fsp->state) {
		case OPENFC_SRB_CMD_SENT:
			/*
			 * if we did not get response for the command
			 * for 10sec. we can assume that it
			 * got dropped. so abort the command
			 */
			if (!(disk->tgtp->flags & OPENFC_TGT_REC_SUPPORTED)) {
				/*
				 * if the device does not support
				 * REC ELS then abort only the read/write
				 * commands
				 */
				if ((cdb_op != READ_10) &&
				    (cdb_op != READ_6) &&
				    (cdb_op != WRITE_10) &&
				    (cdb_op != WRITE_6)) {
					unlock++;
					rc = OPENFC_TIMEOUT_SUCESS;
				} else {
					send_abort++;
				}
			} else {
				send_rec++;
			}
			break;
		case OPENFC_SRB_IN_DATA_TRANS:
		case OPENFC_SRB_WAIT_FOR_STATUS:
			/*
			 * if we are waiting for next packet
			 * to arrive for more then 10sec then
			 * we will assume that pkt got dropped
			 * so abort the I/O.
			 */
			if ((jiffies - fsp->last_pkt_time) >=
			    (OPENFC_SCSI_ER_TIMEOUT) / 2) {
				if ((disk->tgtp->flags &
				     OPENFC_TGT_REC_SUPPORTED) == 0) {
					/*
					 * if the device does not support REC
					 * ELS then abort only the read/write
					 * commands
					 */
					if ((cdb_op != READ_10) &&
					    (cdb_op != READ_6) &&
					    (cdb_op != WRITE_10) &&
					    (cdb_op != WRITE_6)) {
						/*
						 * do not do timeouts any more
						 * for this I/O
						 */
						unlock++;
						rc = OPENFC_TIMEOUT_SUCESS;
					} else {
						send_abort++;
					}
				} else {
					send_rec++;
				}
			} else {
				unlock++;
				if (fsp->timer_fn)
					fsp->timer_fn(fsp);
			}	
			break;
		}
		if (send_rec) {
			openfc_scsi_rec(fsp);
			spin_unlock(&openfcp->outstandingcmd[idx].
				    scsi_pkt_lock);
			rc = OPENFC_TIMEOUT_SUCESS;
		}
		if (send_abort) {
			/*
			 *
			 */
			openfcp->outstandingcmd[idx].ptr = NULL;
			spin_unlock(&openfcp->outstandingcmd[idx].
				    scsi_pkt_lock);
			openfc_timeout_error(fsp);
			rc = OPENFC_TIMEOUT_SUCESS;
		}
		if (unlock) {
			spin_unlock(&openfcp->outstandingcmd[idx].
				    scsi_pkt_lock);
		}
	}
	return rc;
}

/*
 * Send a REC ELS request
 */
void openfc_scsi_rec(struct fc_scsi_pkt *fsp)
{
	struct fc_scsi_pkt *nfsp = openfc_alloc_scsi_pkt(fsp->openfcp);
	struct fc_seq *sp;

	if (!nfsp)
		return;

	openfc_scsi_pkt_hold(fsp);
	nfsp->old_state = fsp->state;
	nfsp->last_pkt_time = fsp->last_pkt_time;
	nfsp->old_pkt = fsp;
	nfsp->disk = fsp->disk;
	nfsp->state = OPENFC_SRB_REC;
	sp = fc_seq_rec_req(fsp->seq_ptr, openfc_scsi_rec_rcv,
			    openfc_scsi_error, nfsp);
	if (!sp)
		return;

	/* wait for 2 sec */
	fc_exch_timer_set(fc_seq_exch(sp), (2 * HZ));
}

/*
 * Receive handler for REC ELS frame
 * if it is a reject then let the scsi layer to handle
 * the timeout. if it is a LS_ACC then if the io was not completed
 * then set the timeout and return otherwise complete the exchange
 * and tell the scsi layer to restart the I/O.
 */
void openfc_scsi_rec_rcv(struct fc_seq *sp, struct fc_frame *fp, void *arg)
{
	struct fc_scsi_pkt *nfsp = (struct fc_scsi_pkt *)arg;
	struct fc_scsi_pkt *ofsp = nfsp->old_pkt;
	struct openfc_softc *openfcp = nfsp->openfcp;
	int idx = ofsp->idx;
	int need_unlock = 0;
	struct os_lun *disk;
	struct fc_els_rec_acc *recp;
	struct fc_els_ls_rjt *rjt;
	uint32_t e_stat;
	uint8_t opcode;
	uint8_t reason;

	opcode = fc_frame_payload_op(fp);

	if (opcode == ELS_LS_RJT) {
		rjt = fc_frame_payload_get(fp, sizeof(*rjt));
		reason = net8_get(&rjt->er_reason);
		switch (reason) {
		case ELS_RJT_LOGIC:
			/*
			 * if fsp state is in CMD_SENT then
			 * cmd frame got dropped, restart it
			 */
			spin_lock(&openfcp->outstandingcmd[ofsp->idx].
				  scsi_pkt_lock);
			openfcp->outstandingcmd[idx].ptr = NULL;
			spin_unlock(&openfcp->outstandingcmd[idx].
				    scsi_pkt_lock);
			openfc_timeout_error(ofsp);
			break;
		case ELS_RJT_UNSUP:
			OFC_DBG("device does not support REC\n");
		default:
			disk = ofsp->disk;
			disk->tgtp->flags &= OPENFC_TGT_REC_NOT_SUPPORTED;
			if (ofsp->timer_fn)
				ofsp->timer_fn(ofsp);
			break;
		}
	} else if (opcode == ELS_LS_ACC) {

		recp = fc_frame_payload_get(fp, sizeof(*recp));
		spin_lock(&openfcp->outstandingcmd[ofsp->idx].scsi_pkt_lock);
		e_stat = net32_get(&recp->reca_e_stat);
		if (e_stat & ESB_ST_COMPLETE) {
			/*
			 * in this case drive thinks exchange has completed
			 * that means we have lost response pkt. retry the io
			 */
			if (openfcp->outstandingcmd[ofsp->idx].ptr == ofsp) {
				/*
				 * make sure I/O did not complete
				 */
				if ((ofsp->state != OPENFC_SRB_FREE) &&
				    (ofsp->state != OPENFC_SRB_ABORTED) &&
				    (nfsp->old_state == ofsp->state)) {
					openfcp->outstandingcmd[idx].ptr = NULL;
					spin_unlock(&openfcp->
						    outstandingcmd[idx].
						    scsi_pkt_lock);
					openfc_timeout_error(ofsp);
				} else {
					need_unlock++;
				}
			} else {
				need_unlock++;
			}
		} else {

			/*
			 * here exchange is open therefore we need to look at
			 * Seq INIT bit.if the initiator has the seq. initiative
			 * then most likely XFER ready or Data pkt got dropped.
			 * in this case we need to restart the io.
			 * else exchange is alive and no pkt has dropped
			 */
			if ((e_stat & ESB_ST_SEQ_INIT) == 0) {
				if (openfcp->outstandingcmd[ofsp->idx].ptr ==
				    ofsp) {
					/*
					 * above check is for paranoia
					 * restart I/O
					 */
					if ((ofsp->state != OPENFC_SRB_FREE) &&
					    (ofsp->state != OPENFC_SRB_ABORTED)
					    && (nfsp->old_state == ofsp->state)) {
						openfcp->outstandingcmd[idx].
						    ptr = NULL;
						spin_unlock(&openfcp->
							    outstandingcmd[idx].
							    scsi_pkt_lock);
						openfc_timeout_error(ofsp);
					} else {
						need_unlock++;
					}
				} else {
					need_unlock++;
				}
			} else {
				/*
				 * drive has the seq intiative
				 * wait for completion
				 */
				need_unlock++;
				if (ofsp->timer_fn)
					ofsp->timer_fn(ofsp);
			}
		}
		if (need_unlock)
			spin_unlock(&openfcp->outstandingcmd[idx].
				    scsi_pkt_lock);
	}
	nfsp->state = OPENFC_SRB_FREE;
	openfc_scsi_pkt_release(ofsp);
	openfc_free_scsi_pkt(nfsp);
	fc_frame_free(fp);
}

/*
 * Time out error routine:
 * abort's the I/O close the exchange and
 * send completion notification to scsi layer
 */
static void openfc_timeout_error(struct fc_scsi_pkt *fsp)
{
	int idx;
	struct os_lun *disk;

	idx = fsp->idx;
	fsp->state = OPENFC_SRB_ABORT_PENDING;
	if (fsp->seq_ptr) {
		fc_seq_abort_exch(fsp->seq_ptr);
		fc_seq_exch_complete(fsp->seq_ptr);
	}
	fsp->seq_ptr = NULL;
	fsp->status_code = OPENFC_CMD_TIME_OUT;
	fsp->cdb_status = 0;
	fsp->io_status = 0;
	disk = fsp->disk;
	disk->error_cnt++;
	/*
	 * if we get 5 consecutive errors
	 * for the same command then offline the device.
	 */
	if (disk->error_cnt >= OPENFC_MAX_ERROR_CNT) {
		disk->state = OPENFC_LUN_ERR;
		fsp->status_code = OPENFC_HRD_ERROR;
	} else {
		fsp->io_status = (SUGGEST_RETRY << 24);
	}
	if (fsp->done)
		(*fsp->done) (fsp);
}
